"use strict";
var self = Adapter;
module.exports = self;
var parseJsonBody = require('./parseJsonBody');
const { SERVICES_NAME_TO_TYPE_MAP } = require('./constants');
const { TRACE_ID_HEADER, ThreadLocal } = require('@jfrog/nodejs-commons');
const util = require('util');
const { ConfigFactory } = require('@jfrog-pipelines/configuration');
const { AmqpFactory, AmqpPublisher, publisherOptions } = require('@jfrog-pipelines/amqp-client');
const async = require('async');
const request = require('request');
function Adapter(jFrogPipelinesSessionId, token) {
    logger.debug(util.format('Initializing %s', self.name));
    this.token = token;
    this.baseUrl = ConfigFactory.config.apiUrl;
    this.headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': 'apiToken '.concat(token),
        'jfrog-pipelines-session-id': jFrogPipelinesSessionId,
        [TRACE_ID_HEADER]: ThreadLocal.traceId
    };
    if (global.msName in SERVICES_NAME_TO_TYPE_MAP) {
        this.headers['user-agent'] =
            `JFrog-${SERVICES_NAME_TO_TYPE_MAP[global.msName]}`;
    }
    else {
        logger.warn(`no service type found for service : ${global.msName}`);
    }
    this.jFrogPipelinesSessionId = jFrogPipelinesSessionId;
    this.publisher = new AmqpPublisher(AmqpFactory.getClient(), ConfigFactory.config.amqpExchange, publisherOptions, this.logger);
}
Adapter.prototype.getBuildPlaneImages =
    function (query, callback) {
        this.get('/buildPlaneImages?' + query, callback);
    };
Adapter.prototype.getNodePools =
    function (query, callback) {
        this.get(util.format('/nodePools?%s', query), callback);
    };
Adapter.prototype.getNodes =
    function (query, callback) {
        this.get(util.format('/nodes?%s', query), callback);
    };
Adapter.prototype.getNodeById =
    function (nodeId, callback) {
        this.get(util.format('/nodes/%s', nodeId), callback);
    };
Adapter.prototype.getTemplateSourceById =
    function (templateSourceId, callback) {
        this.get(util.format('/templateSources/%s', templateSourceId), callback);
    };
Adapter.prototype.putTemplateSourceById =
    function (templateSourceId, body, callback) {
        this.put(util.format('/templateSources/%s', templateSourceId), body, callback);
    };
Adapter.prototype.triggerTemplateSync =
    function (templateSourceId, body, callback) {
        var url = '/templateSources/' + templateSourceId + '/triggerSync';
        this.post(url, { payload: body }, callback);
    };
Adapter.prototype.getExtensionSourceById =
    function (extensionSourceId, callback) {
        this.get(util.format('/extensionSources/%s', extensionSourceId), callback);
    };
Adapter.prototype.putExtensionSourceById =
    function (extensionSourceId, body, callback) {
        this.put(util.format('/extensionSources/%s', extensionSourceId), body, callback);
    };
Adapter.prototype.getTemplates =
    function (query, callback) {
        this.get(util.format('/templates?%s', query), callback);
    };
Adapter.prototype.getTemplateById =
    function (templateId, callback) {
        this.get(util.format('/templates/%s', templateId), callback);
    };
Adapter.prototype.putTemplateById =
    function (templateId, body, callback) {
        this.put(util.format('/templates/%s', templateId), body, callback);
    };
Adapter.prototype.postTemplates =
    function (json, callback) {
        this.post(util.format('/templates'), json, callback);
    };
Adapter.prototype.getTemplateArtifactUrlById =
    function (templateId, query, callback) {
        this.get(util.format('/templates/%s/artifactUrl?%s', templateId, query), callback);
    };
Adapter.prototype.getMasterResources =
    function (query, callback) {
        this.get(util.format('/masterResources?%s', query), callback);
    };
Adapter.prototype.getMasterResourceById =
    function (masterResourceId, callback) {
        this.get('/masterResources/' + masterResourceId, callback);
    };
Adapter.prototype.putMasterResourceById =
    function (masterResourceId, body, callback) {
        this.put(util.format('/masterResources/%s', masterResourceId), body, callback);
    };
Adapter.prototype.postMasterResources =
    function (json, callback) {
        this.post(util.format('/masterResources'), json, callback);
    };
Adapter.prototype.getBuildPlaneMessages =
    function (query, callback) {
        var url = '/buildPlaneMessages?' + query;
        this.get(url, callback);
    };
Adapter.prototype.getBuildPlaneMessagesCount =
    function (query, callback) {
        const url = `/buildPlaneMessages/count?${query}`;
        this.get(url, callback);
    };
Adapter.prototype.postBuildPlaneMessage =
    function (json, callback) {
        var url = '/buildPlaneMessages';
        this.post(url, json, callback);
    };
Adapter.prototype.queueToNodePool =
    function (json, callback) {
        var url = '/buildPlaneMessages/queueToNodePool';
        this.post(url, json, callback);
    };
Adapter.prototype.getBuildPlaneMessageById =
    function (messageId, callback) {
        var url = '/buildPlaneMessages/' + messageId;
        this.get(url, callback);
    };
Adapter.prototype.deleteBuildPlaneMessagesByRunId =
    function (runId, callback) {
        var url = `/runs/${runId}/buildPlaneMessages`;
        this.delete(url, callback);
    };
Adapter.prototype.putBuildPlaneMessageById =
    function (messageId, json, callback) {
        var url = '/buildPlaneMessages/' + messageId;
        this.put(url, json, callback);
    };
Adapter.prototype.pollNextMessage =
    function (nodeId, json, callback) {
        var url = '/nodes/' + nodeId + '/nextMessage';
        this.post(url, json, callback);
    };
Adapter.prototype.deleteBuildPlaneMessageById =
    function (messageId, callback) {
        var url = '/buildPlaneMessages/' + messageId;
        this.delete(url, callback);
    };
Adapter.prototype.getDeployKeys =
    function (query, callback) {
        this.get(util.format('/deployKeys?%s', query), callback);
    };
Adapter.prototype.getMasterResourceArtifactUrlById =
    function (masterResourceId, query, callback) {
        this.get(util.format('/masterResources/%s/artifactUrl?%s', masterResourceId, query), callback);
    };
Adapter.prototype.getNodeStats =
    function (query, callback) {
        this.get(util.format('/nodeStats?%s', query), callback);
    };
Adapter.prototype.getSystemConfigByName =
    function (configName, callback) {
        this.get(util.format('/systemConfigs/%s', configName), callback);
    };
Adapter.prototype.putNodePoolById =
    function (nodePoolId, json, callback) {
        this.put(util.format('/nodePools/%s', nodePoolId), json, callback);
    };
Adapter.prototype.postNodeStats =
    function (json, callback) {
        this.post(util.format('/nodeStats'), json, callback);
    };
Adapter.prototype.postNodeStatistics =
    function (callback) {
        this.post(util.format('/nodes/updateStatistics'), {}, callback);
    };
Adapter.prototype.postLock =
    function (json, callback) {
        this.post(util.format('/locks'), json, callback);
    };
Adapter.prototype.getHooks =
    function (query, callback) {
        this.get(util.format('/hooks?%s', query), callback);
    };
Adapter.prototype.getHookById =
    function (hookId, callback) {
        this.get(util.format('/hooks/%s', hookId), callback);
    };
Adapter.prototype.postHook =
    function (body, callback) {
        var url = '/hooks';
        this.post(url, body, callback);
    };
Adapter.prototype.putHookById =
    function (hookId, body, callback) {
        var url = '/hooks/' + hookId;
        this.put(url, body, callback);
    };
Adapter.prototype.deleteHookById =
    function (hookId, callback) {
        var url = '/hooks/' + hookId;
        this.delete(url, callback);
    };
Adapter.prototype.postHookWebhook =
    function (hookId, body, callback) {
        this.post('/hooks/' + hookId + '/hook', body, callback);
    };
Adapter.prototype.getExtensionSources =
    function (query, callback) {
        this.get(util.format('/extensionSources?%s', query), callback);
    };
Adapter.prototype.triggerExtensionSync =
    function (extensionSourceId, body, callback) {
        var url = '/extensionSources/' + extensionSourceId + '/triggerSync';
        this.post(url, { payload: body }, callback);
    };
Adapter.prototype.getTemplateSources =
    function (query, callback) {
        this.get('/templateSources?' + query, callback);
    };
Adapter.prototype.getIdentities =
    function (query, callback) {
        this.get(util.format('/identities?%s', query), callback);
    };
Adapter.prototype.getIdentityById =
    function (identityId, callback) {
        this.get(util.format('/identities/%s', identityId), callback);
    };
Adapter.prototype.getIdentityTokens =
    function (query, callback) {
        this.get(util.format('/identityTokens?%s', query), callback);
    };
Adapter.prototype.postIdentityToken =
    function (json, callback) {
        this.post(util.format('/identityTokens'), json, callback);
    };
Adapter.prototype.deleteIdentityTokenById =
    function (tokenId, callback) {
        this.delete(util.format('/identityTokens/%s', tokenId), callback);
    };
Adapter.prototype.deleteIdentityTokenByIdWithQuery =
    function (tokenId, query, callback) {
        this.delete(util.format('/identityTokens/%s?%s', tokenId, query), callback);
    };
Adapter.prototype.getMasterIntegrations =
    function (query, callback) {
        this.get(util.format('/masterIntegrations?%s', query), callback);
    };
Adapter.prototype.getProviderById =
    function (id, callback) {
        this.get(util.format('/providers/%s', id), callback);
    };
Adapter.prototype.getProviders =
    function (query, callback) {
        this.get(util.format('/providers?%s', query), callback);
    };
Adapter.prototype.getPipelineObjects =
    function (query, callback) {
        this.get(util.format('/pipelineObjects?%s', query), callback);
    };
Adapter.prototype.postPipelineObjects =
    function (json, callback) {
        this.post('/pipelineObjects', json, callback);
    };
Adapter.prototype.deletePipelineObjectById =
    function (pipelineObjectId, callback) {
        this.delete(util.format('/pipelineObjects/%s', pipelineObjectId), callback);
    };
Adapter.prototype.getPipelineSourceById =
    function (pipelineSourceId, callback) {
        this.get(util.format('/pipelineSources/%s', pipelineSourceId), callback);
    };
Adapter.prototype.getPipelineSources =
    function (query, callback) {
        this.get(util.format('/pipelineSources?%s', query), callback);
    };
Adapter.prototype.putPipelineSourceById =
    function (pipelineSourceId, json, callback) {
        this.put(util.format('/pipelineSources/%s', pipelineSourceId), json, callback);
    };
Adapter.prototype.getPipelineSourceObjectPermissions =
    function (query, callback) {
        this.get(util.format('/pipelineSourceObjectPermissions?%s', query), callback);
    };
Adapter.prototype.getMicroPipelines =
    function (query, callback) {
        this.get(`/microPipelines?${query}`, callback);
    };
Adapter.prototype.putMicroPipelineById =
    function (microPipelineId, json, callback) {
        this.put(`/microPipelines/${microPipelineId}`, json, callback);
    };
Adapter.prototype.postPipelineSyncStatus =
    function (json, callback) {
        this.post('/pipelineSyncStatuses', json, callback);
    };
Adapter.prototype.putPipelineSyncStatusById =
    function (pipelineSyncStatusId, json, callback) {
        this.put(util.format('/pipelineSyncStatuses/%s', pipelineSyncStatusId), json, callback);
    };
Adapter.prototype.getMicroPipelineEvents =
    function (query, callback) {
        this.get(`/microPipelineEvents?${query}`, callback);
    };
Adapter.prototype.getMicroPipelineById =
    function (microPipelineId, json, callback) {
        this.get(`/microPipelines/${microPipelineId}`, callback);
    };
Adapter.prototype.putMicroPipelineEventById =
    function (microPipelineEventId, json, callback) {
        this.put(`/microPipelineEvents/${microPipelineEventId}`, json, callback);
    };
Adapter.prototype.postMicroPipelineEvents =
    function (json, callback) {
        this.post(`/microPipelineEvents`, json, callback);
    };
Adapter.prototype.deleteMicroPipelineEventById =
    function (microPipelineEventId, callback) {
        this.delete(`/microPipelineEvents/${microPipelineEventId}`, callback);
    };
Adapter.prototype.putLockById =
    function (lockId, json, callback) {
        this.put(util.format('/locks/%s', lockId), json, callback);
    };
Adapter.prototype.postPipelineStep =
    function (json, callback) {
        this.post('/pipelineSteps', json, callback);
    };
Adapter.prototype.triggerPipelineStepById =
    function (pipelineStepId, json, callback) {
        return this.post(util.format('/pipelineSteps/%s/trigger', pipelineStepId), json, callback);
    };
Adapter.prototype.postPipelineSteplet =
    function (json, callback) {
        this.post('/pipelineSteplets', json, callback);
    };
Adapter.prototype.putPipelineStepById =
    function (pipelineStepId, json, callback) {
        this.put(util.format('/pipelineSteps/%s', pipelineStepId), json, callback);
    };
Adapter.prototype.putPipelineStepletById =
    function (pipelineStepletId, json, callback) {
        this.put(util.format('/pipelineSteplets/%s', pipelineStepletId), json, callback);
    };
Adapter.prototype.deletePipelineStepById =
    function (pipelineStepId, query, callback) {
        this.delete(util.format('/pipelineSteps/%s?%s', pipelineStepId, query), callback);
    };
Adapter.prototype.deletePipelineStepletById =
    function (pipelineStepletId, query, callback) {
        this.delete(util.format('/pipelineSteplets/%s?%s', pipelineStepletId, query), callback);
    };
Adapter.prototype.deleteLockById =
    function (lockId, query, callback) {
        this.delete(util.format('/locks/%s?%s', lockId, query), callback);
    };
Adapter.prototype.getResourceVersionById =
    function (resourceVersionId, callback) {
        this.get(util.format('/resourceVersions/%s', resourceVersionId), callback);
    };
Adapter.prototype.getPipelines =
    function (query, callback) {
        this.get(util.format('/pipelines?%s', query), callback);
    };
Adapter.prototype.getPipelineById =
    function (pipelineId, callback) {
        this.get(util.format('/pipelines/%s', pipelineId), callback);
    };
Adapter.prototype.getDisabledPipelineEntities =
    function (query, callback) {
        this.get(util.format('/disabledEntities?%s', query), callback);
    };
Adapter.prototype.postPipeline =
    function (json, callback) {
        this.post('/pipelines', json, callback);
    };
Adapter.prototype.cancelRunById =
    function (runId, callback) {
        this.post('/runs/' + runId + '/cancel', {}, callback);
    };
Adapter.prototype.deleteRunById =
    function (runId, callback) {
        this.delete('/runs/' + runId, callback);
    };
Adapter.prototype.postRun =
    function (json, callback) {
        this.post('/runs', json, callback);
    };
Adapter.prototype.addToMerkleTree =
    function (runId, hash, callback) {
        this.put(util.format('/runs/%s/merkle', runId), { hash }, callback);
    };
Adapter.prototype.getRunKey =
    function (runId, callback) {
        this.get(util.format('/runs/%s/key', runId), callback);
    };
Adapter.prototype.deleteRunKey =
    function (runId, callback) {
        this.delete(util.format('/runs/%s/key', runId), callback);
    };
Adapter.prototype.getRuns =
    function (query, callback) {
        this.get(util.format('/runs?%s', query), callback);
    };
Adapter.prototype.getRunById =
    function (runId, callback) {
        this.get(util.format('/runs/%s', runId), callback);
    };
Adapter.prototype.getProjects =
    function (query, callback) {
        this.get(util.format('/projects?%s', query), callback);
    };
Adapter.prototype.getProjectById =
    function (projectId, callback) {
        this.get(util.format('/projects/%s', projectId), callback);
    };
Adapter.prototype.getRunStepConnections =
    function (query, callback) {
        this.get(util.format('/runStepConnections?%s', query), callback);
    };
Adapter.prototype.getStatistics =
    function (query, callback) {
        this.get(util.format('/statistics?%s', query), callback);
    };
Adapter.prototype.postStatistic =
    function (json, callback) {
        this.post('/statistics', json, callback);
    };
Adapter.prototype.collectMetrics =
    function (query, callback) {
        this.get(`/metrics?${query}`, callback);
    };
Adapter.prototype.timeoutStuckSteps =
    function (callback) {
        this.post('/steps/management/timeoutSteps', {}, callback);
    };
Adapter.prototype.deleteStatisticById =
    function (statisticId, callback) {
        this.delete(util.format('/statistics/%s', statisticId), callback);
    };
Adapter.prototype.deleteStatistics =
    function (query, callback) {
        this.delete(util.format('/statistics?%s', query), callback);
    };
Adapter.prototype.getArtifactUrlByStepId =
    function (stepId, query, callback) {
        this.get(util.format('/steps/%s/artifactUrl?%s', stepId, query), callback);
    };
Adapter.prototype.getArtifactUrlByRunId =
    function (runId, query, callback) {
        this.get(util.format('/runs/%s/artifactUrl?%s', runId, query), callback);
    };
Adapter.prototype.getStepExecutionEligibility =
    function (stepId, callback) {
        this.get(util.format('/steps/%s/executionEligibility', stepId), callback);
    };
Adapter.prototype.updateStepStatusByStepId =
    function (stepId, json, callback) {
        this.put(util.format('/steps/%s/status', stepId), json, callback);
    };
Adapter.prototype.requestStepApproval =
    function (stepId, callback) {
        this.post(util.format('/steps/%s/requestApproval', stepId), {}, callback);
    };
Adapter.prototype.getStepConsolesByStepId =
    function (stepId, query, callback) {
        this.get(util.format('/steps/%s/consoles?%s', stepId, query), callback);
    };
Adapter.prototype.getStepletConsolesByStepletId =
    function (stepletId, query, callback) {
        this.get(util.format('/steplets/%s/consoles?%s', stepletId, query), callback);
    };
Adapter.prototype.getSteps =
    function (query, callback) {
        this.get(util.format('/steps?%s', query), callback);
    };
Adapter.prototype.getSteplets =
    function (query, callback) {
        this.get(util.format('/steplets?%s', query), callback);
    };
Adapter.prototype.getStepletById =
    function (stepletId, callback) {
        this.get(util.format('/steplets/%s', stepletId), callback);
    };
Adapter.prototype.getStepletArtifactUrl =
    function (stepletId, query, callback) {
        this.get(util.format('/steplets/%s/runArtifactUrl?%s', stepletId, query), callback);
    };
Adapter.prototype.getRunArtifactUrlsForStep =
    function (stepId, callback) {
        this.get(util.format('/steps/%s/runArtifactUrl', stepId), callback);
    };
Adapter.prototype.putStepById =
    function (stepId, json, callback) {
        this.put(util.format('/steps/%s', stepId), json, callback);
    };
Adapter.prototype.putStepletById =
    function (stepletId, json, callback) {
        this.put(util.format('/steplets/%s', stepletId), json, callback);
    };
Adapter.prototype.getStepVariables =
    function (query, callback) {
        this.get(util.format('/stepVariables?%s', query), callback);
    };
Adapter.prototype.putRunById =
    function (runId, json, callback) {
        this.put(util.format('/runs/%s', runId), json, callback);
    };
Adapter.prototype.postRunResourceVersion =
    function (json, callback) {
        this.post('/runResourceVersions', json, callback);
    };
Adapter.prototype.putRunResourceVersionById =
    function (id, json, callback) {
        var url = '/runResourceVersions/' + id;
        this.put(url, json, callback);
    };
Adapter.prototype.getRunResourceVersions =
    function (query, callback) {
        this.get(util.format('/runResourceVersions?%s', query), callback);
    };
Adapter.prototype.getRunResourceVersionsLight =
    function (query, callback) {
        this.get(util.format('/runResourceVersions/light?%s', query), callback);
    };
Adapter.prototype.getResourceVersions =
    function (query, callback) {
        this.get(util.format('/resourceVersions?%s', query), callback);
    };
Adapter.prototype.getResourceVersionsLight =
    function (query, callback) {
        this.get(util.format('/resourceVersions/light?%s', query), callback);
    };
Adapter.prototype.postRunStepConnection =
    function (json, callback) {
        this.post('/runStepConnections', json, callback);
    };
Adapter.prototype.postStep =
    function (json, callback) {
        this.post('/steps', json, callback);
    };
Adapter.prototype.postSteplet =
    function (json, callback) {
        this.post('/steplets', json, callback);
    };
Adapter.prototype.putPipelineById =
    function (pipelineId, json, callback) {
        this.put(util.format('/pipelines/%s', pipelineId), json, callback);
    };
Adapter.prototype.deletePipelineById =
    function (pipelineId, query, callback) {
        this.delete(util.format('/pipelines/%s?%s', pipelineId, query), callback);
    };
Adapter.prototype.getPipelineSteps =
    function (query, callback) {
        this.get(util.format('/pipelineSteps?%s', query), callback);
    };
Adapter.prototype.getPipelineStepById =
    function (pipelineStepId, callback) {
        this.get(util.format('/pipelineSteps/%s', pipelineStepId), callback);
    };
Adapter.prototype.getPipelineSteplets =
    function (query, callback) {
        this.get(util.format('/pipelineSteplets/?%s', query), callback);
    };
Adapter.prototype.getPipelineStepletById =
    function (pipelineStepletId, callback) {
        this.get(util.format('/pipelineSteplets/%s', pipelineStepletId), callback);
    };
Adapter.prototype.getStepById =
    function (id, query, callback) {
        this.get(util.format('/steps/%s?%s', id, query), callback);
    };
Adapter.prototype.getNextSteps =
    function (runId, json, callback) {
        this.post(util.format('/runs/%s/nextSteps', runId), json, callback);
    };
Adapter.prototype.postPipelineStepConnection =
    function (json, callback) {
        this.post('/pipelineStepConnections', json, callback);
    };
Adapter.prototype.postPipelineStepConnectionBulk =
    function (json, callback) {
        this.post('/pipelineStepConnections/bulk', json, callback);
    };
Adapter.prototype.getPipelineStepConnections =
    function (query, callback) {
        this.get(util.format('/pipelineStepConnections?%s', query), callback);
    };
Adapter.prototype.deletePipelineStepConnectionById =
    function (pipelineStepConnectionId, callback) {
        this.delete(util.format('/pipelineStepConnections/%s', pipelineStepConnectionId), callback);
    };
Adapter.prototype.getProjectIntegrations =
    function (query, callback) {
        this.get(util.format('/projectIntegrations?%s', query), callback);
    };
Adapter.prototype.getProjectIntegrationById =
    function (projectIntegrationId, callback) {
        this.get(util.format('/projectIntegrations/%s', projectIntegrationId), callback);
    };
Adapter.prototype.getResources =
    function (query, callback) {
        this.get(util.format('/resources?%s', query), callback);
    };
Adapter.prototype.getResourceById =
    function (resourceId, callback) {
        this.get(util.format('/resources/%s', resourceId), callback);
    };
Adapter.prototype.deleteScmHookAndKeyByResourceId =
    function (resourceId, callback) {
        this.delete(util.format('/resources/%s/scm', resourceId), callback);
    };
Adapter.prototype.getSystemCodes =
    function (query, callback) {
        this.get('/systemCodes?' + query, callback);
    };
Adapter.prototype.getSystemServices =
    function (query, callback) {
        this.get('/systemServices?' + query, callback);
    };
Adapter.prototype.getSystemServiceInstances =
    function (query, callback) {
        this.get('/systemServiceInstances?' + query, callback);
    };
Adapter.prototype.getVersionById =
    function (versionId, callback) {
        this.get(util.format('/versions/%s', versionId), callback);
    };
Adapter.prototype.getVersions =
    function (query, callback) {
        this.get(util.format('/versions?%s', query), callback);
    };
Adapter.prototype.postResourceVersion =
    function (json, callback) {
        this.post('/resourceVersions', json, callback);
    };
Adapter.prototype.putVersionById =
    function (versionId, json, callback) {
        this.put(util.format('/versions/%s', versionId), json, callback);
    };
Adapter.prototype.getWorkflowControllers =
    function (query, callback) {
        this.get(util.format('/workflowControllers?%s', query), callback);
    };
Adapter.prototype.postWorkflowControllers =
    function (body, callback) {
        var url = '/workflowControllers';
        this.post(url, body, callback);
    };
Adapter.prototype.getWebhookById =
    function (webhookId, callback) {
        this.get(util.format('/webhooks/%s', webhookId), callback);
    };
Adapter.prototype.getWebhooks =
    function (query, callback) {
        this.get(util.format('/webhooks?%s', query), callback);
    };
Adapter.prototype.putWebhookById =
    function (webhookId, json, callback) {
        this.put(util.format('/webhooks/%s', webhookId), json, callback);
    };
Adapter.prototype.deleteWebhooksByHookId =
    function (hookId, query, callback) {
        this.delete(util.format('/hooks/%s/webhooks?%s', hookId, query), callback);
    };
Adapter.prototype.deleteNodeById =
    function (nodeId, callback) {
        this.delete(util.format('/nodes/%s', nodeId), callback);
    };
Adapter.prototype.deleteNodeConsolesByNodeId =
    function (nodeId, callback) {
        this.delete(util.format('/nodes/%s/nodeConsoles', nodeId), callback);
    };
Adapter.prototype.deleteNodeStatsByNodeId =
    function (nodeId, query, callback) {
        this.delete(util.format('/nodes/%s/nodeStats?%s', nodeId, query), callback);
    };
Adapter.prototype.deleteResourceById =
    function (resourceId, query, callback) {
        this.delete(util.format('/resources/%s?%s', resourceId, query), callback);
    };
Adapter.prototype.deleteResourceVersionsByResourceId =
    function (resourceId, callback) {
        var url = '/resources/' + resourceId + '/resourceVersions';
        this.delete(url, callback);
    };
Adapter.prototype.triggerRun =
    function (json, callback) {
        this.post('/runs/triggerRun', { payload: json }, callback);
    };
Adapter.prototype.triggerPipelineSync =
    function (pipelineSourceId, body, callback) {
        this.post('/pipelineSources/' + pipelineSourceId + '/triggerSync', { payload: body }, callback);
    };
Adapter.prototype.getPipelineSyncStatuses =
    function (query, callback) {
        var url = '/pipelineSyncStatuses?' + query;
        this.get(url, callback);
    };
Adapter.prototype.deletePipelineSyncStatusById =
    function (statusId, callback) {
        var url = '/pipelineSyncStatuses/' + statusId;
        this.delete(url, callback);
    };
Adapter.prototype.getBranchesByIntegrationId =
    function (projectIntegrationId, query, callback) {
        this.get(util.format('/passthrough/projectIntegrations/%s/branches?%s', projectIntegrationId, query), callback);
    };
Adapter.prototype.deleteStepConsolesByStepId =
    function (stepId, callback) {
        this.delete(util.format('/steps/%s/stepConsoles', stepId), callback);
    };
Adapter.prototype.deleteStepletConsolesByStepletId =
    function (stepletId, callback) {
        this.delete(util.format('/steplets/%s/stepletConsoles', stepletId), callback);
    };
Adapter.prototype.deleteVersionById =
    function (versionId, callback) {
        this.delete(util.format('/versions/%s', versionId), callback);
    };
Adapter.prototype.postStepConsoles =
    function (json, callback) {
        this.post(util.format('/stepConsoles'), json, callback);
    };
Adapter.prototype.postStepletConsoles =
    function (json, callback) {
        this.post(util.format('/stepletConsoles'), json, callback);
    };
Adapter.prototype.postResource =
    function (json, callback) {
        this.post('/resources', json, callback);
    };
Adapter.prototype.postSystemServiceInstance =
    function (json, callback) {
        this.post('/systemServiceInstances', json, callback);
    };
Adapter.prototype.postToNode =
    function (node, callback) {
        this.post('/nodes', node, callback);
    };
Adapter.prototype.postToNodeConsoles =
    function (json, callback) {
        this.post('/nodeConsoles', json, callback);
    };
Adapter.prototype.postToVortex =
    function (message, callback) {
        this.getSystemConfigByName('experimentalProperties', (err, { configValue }) => {
            if (err) {
                logger.warn(`Failed fetching experimentalProperties for queueName changes, passed queue name is ${message.where}`, inspect(err));
            }
            if (!_.isEmpty(configValue?.serviceQueueMapping)) {
                message.where = configValue?.serviceQueueMapping[message.where] || message.where;
            }
            this.publisher.publishMessage(message.where, message.payload)
                .then((res) => callback(null, res))
                .catch((err) => callback(err));
        });
    };
Adapter.prototype.postToSUVortex =
    function (message, callback) {
        this.post('/vortexSU', message, callback);
    };
Adapter.prototype.postToVortexUrl =
    function (message, callback) {
        this.post('/vortex', message, callback);
    };
Adapter.prototype.triggerNewBuildByResourceId =
    function (resourceId, json, callback) {
        this.post(util.format('/resources/%s/triggerNewBuildRequest', resourceId), json, callback);
    };
Adapter.prototype.putNodeById =
    function (nodeId, node, callback) {
        this.put(util.format('/nodes/%s', nodeId), node, callback);
    };
Adapter.prototype.putResourceById =
    function (id, json, callback) {
        this.put(util.format('/resources/%s', id), json, callback);
    };
Adapter.prototype.getExecutionNodes =
    function (callback) {
        this.get(util.format('/passthrough/missionControl/executionNodes'), callback);
    };
Adapter.prototype.getSHAByIntegrationId =
    function (projectIntegrationId, query, callback) {
        this.get(util.format('/passthrough/projectIntegrations/%s/sha?%s', projectIntegrationId, query), callback);
    };
Adapter.prototype.putSystemServiceInstanceById =
    function (id, json, callback) {
        this.put(util.format('/systemServiceInstances/%s', id), json, callback);
    };
Adapter.prototype.validateNodeById =
    function (nodeId, callback) {
        this.get(util.format('/nodes/%s/validate', nodeId), callback);
    };
Adapter.prototype.validateSystemServiceInstanceById =
    function (systemServiceInstanceId, callback) {
        this.get(util.format('/systemServiceInstances/%s/validate', systemServiceInstanceId), callback);
    };
Adapter.prototype.validateRepositoryOwnerToken =
    function (id, query, callback) {
        this.get(util.format('/projectIntegrations/%s/validateRepositoryOwnerToken?%s', id, query), callback);
    };
Adapter.prototype.getProviderById =
    function (providerId, callback) {
        this.get(util.format('/providers/%s', providerId), callback);
    };
Adapter.prototype.getStepletTestReports =
    function (query, callback) {
        this.get(util.format('/stepletTestReports?%s', query), callback);
    };
Adapter.prototype.postStepTestReport =
    function (json, callback) {
        this.post(util.format('/stepTestReports'), json, callback);
    };
Adapter.prototype.getEnvironmentVariables =
    function (query, callback) {
        var url = '/environmentVariables?' + query;
        this.get(url, callback);
    };
Adapter.prototype.postEnvironmentVariables =
    function (json, callback) {
        var url = '/environmentVariables';
        this.post(url, json, callback);
    };
Adapter.prototype.getEnvironmentVariableById =
    function (evId, callback) {
        var url = '/environmentVariables/' + evId;
        this.get(url, callback);
    };
Adapter.prototype.deleteEnvironmentVariableById =
    function (evId, callback) {
        var url = `environmentVariables/${evId}`;
        this.delete(url, callback);
    };
Adapter.prototype.putEnvironmentVariableById =
    function (evId, json, callback) {
        var url = '/environmentVariables/' + evId;
        this.put(url, json, callback);
    };
Adapter.prototype.deletePipelinesEntities =
    function (json, callback) {
        this.post('/pipelines/deleteEntities', json, callback);
    };
Adapter.prototype.postRunAdjacencies =
    function (json, callback) {
        var url = '/runAdjacencies/bulk';
        this.post(url, json, callback);
    };
Adapter.prototype.getUnifiedUiAndWwwUrl =
    function (callback) {
        const url = '/unifiedUiUrl';
        this.get(url, callback);
    };
Adapter.prototype.deleteAccessSecretsRecords =
    function (query, callback) {
        const url = `/accessSecretsRecords?${query}`;
        this.delete(url, callback);
    };
Adapter.prototype.getAccessSecretsRecords =
    function (query, callback) {
        const url = `/accessSecretsRecords?${query}`;
        this.get(url, callback);
    };
Adapter.prototype.get =
    function (relativeUrl, callback) {
        var bag = {};
        bag.opts = {
            method: 'GET',
            url: this.baseUrl.concat(relativeUrl),
            headers: this.headers,
            timeout: 180000
        };
        bag.who = util.format('%s call to %s for jFrogPipelinesSessionId %s', bag.opts.method, bag.opts.url, this.jFrogPipelinesSessionId);
        logger.trace(util.format('Starting %s', bag.who));
        async.series([
            _performCall.bind(null, bag),
            _parseBody.bind(null, bag)
        ], function () {
            callback(bag.err, bag.parsedBody, bag.res);
        });
    };
Adapter.prototype.post =
    function (relativeUrl, json, callback) {
        var bag = {};
        bag.opts = {
            method: 'POST',
            url: this.baseUrl.concat(relativeUrl),
            headers: this.headers,
            json: json,
            timeout: 180000
        };
        bag.who = util.format('%s call to %s for jFrogPipelinesSessionId %s', bag.opts.method, bag.opts.url, this.jFrogPipelinesSessionId);
        logger.trace(util.format('Starting %s', bag.who));
        async.series([
            _performCall.bind(null, bag),
            _parseBody.bind(null, bag)
        ], function () {
            callback(bag.err, bag.parsedBody, bag.res);
        });
    };
Adapter.prototype.put =
    function (relativeUrl, json, callback) {
        var bag = {};
        bag.opts = {
            method: 'PUT',
            url: this.baseUrl.concat(relativeUrl),
            headers: this.headers,
            json: json,
            timeout: 180000
        };
        bag.who = util.format('%s call to %s for jFrogPipelinesSessionId %s', bag.opts.method, bag.opts.url, this.jFrogPipelinesSessionId);
        logger.trace(util.format('Starting %s', bag.who));
        async.series([
            _performCall.bind(null, bag),
            _parseBody.bind(null, bag)
        ], function () {
            callback(bag.err, bag.parsedBody, bag.res);
        });
    };
Adapter.prototype.delete =
    function (relativeUrl, callback) {
        var bag = {};
        bag.opts = {
            method: 'DELETE',
            url: this.baseUrl.concat(relativeUrl),
            headers: this.headers,
            timeout: 180000
        };
        bag.who = util.format('%s call to %s for jFrogPipelinesSessionId %s', bag.opts.method, bag.opts.url, this.jFrogPipelinesSessionId);
        logger.trace(util.format('Starting %s', bag.who));
        async.series([
            _performCall.bind(null, bag),
            _parseBody.bind(null, bag)
        ], function () {
            callback(bag.err, bag.parsedBody, bag.res);
        });
    };
function _performCall(bag, next) {
    var who = bag.who + '|' + _performCall.name;
    logger.trace(who, 'Inside');
    bag.startedAt = Date.now();
    bag.timeoutLength = 1;
    bag.timeoutLimit = 180;
    __attempt(bag, next);
    function __attempt(bag, callback) {
        request(bag.opts, function (err, res, body) {
            var interval = Date.now() - bag.startedAt;
            var connectionError = false;
            if (res) {
                logger.trace(util.format('%s took %s & returned status %s', bag.who, interval, res.statusCode));
            }
            else {
                connectionError = true;
            }
            let errorCode = 'connectionError';
            let retryLimitExceeded = false;
            const exceedRetryLimit = global.msName === 'hookHandler';
            if (typeof res !== 'undefined') {
                errorCode = res.statusCode;
            }
            if ((res && res.statusCode > 299) || err) {
                err = err || res.statusCode;
                const isInvalidRequest = isInvalidRouterRequest(body);
                if ((res && (isInvalidRequest || res.statusCode >= 500)) || connectionError) {
                    bag.timeoutLength *= 2;
                    if (bag.timeoutLength > bag.timeoutLimit) {
                        bag.timeoutLength = 1;
                        retryLimitExceeded = true;
                    }
                    if (exceedRetryLimit && retryLimitExceeded && connectionError) {
                        logger.error(`${bag.who} Retry limit exceeded for the error ${errorCode}`);
                        bag.err = err;
                    }
                    else {
                        logger.error(`${bag.who} returned error, status ${errorCode}. ` +
                            `Retrying in ${bag.timeoutLength} seconds`);
                        setTimeout(function () {
                            __attempt(bag, callback);
                        }, bag.timeoutLength * 1000);
                        return;
                    }
                }
                else {
                    if (errorCode === 409 && bag?.opts?.url.includes('/locks') &&
                        bag?.opts?.method.includes('POST')) {
                        logger.debug(util.format('%s returned status %s with error %s', bag.who, errorCode, err));
                    }
                    else {
                        logger.warn(util.format('%s returned status %s with error %s', bag.who, errorCode, err));
                    }
                    bag.err = err;
                }
            }
            bag.res = res;
            bag.body = body;
            callback();
        });
    }
}
function _parseBody(bag, next) {
    var who = bag.who + '|' + _parseBody.name;
    logger.trace(who, 'Inside');
    if (bag.body) {
        if ((bag?.res?.headers['content-type']?.startsWith('text/plain'))
            || typeof bag.body === 'object') {
            bag.parsedBody = bag.body;
        }
        else {
            bag.parsedBody = parseJsonBody(bag);
        }
    }
    return next();
}
function isInvalidRouterRequest(body) {
    if (typeof body === 'string' && body.startsWith('404 page not found')) {
        return true;
    }
    return false;
}
//# sourceMappingURL=Adapter.js.map