'use strict';
const self = isExceededBuildMinutesQuota;
module.exports = self;
const util = require('util');
const async = require('async');
const _ = require('underscore');
const { CacheConfig, TypedCache } = require('@jfrog/nodejs-commons');
const createArtifactoryToken = require('./artifactory/createArtifactoryToken');
const ArtifactoryAdapter = require('./artifactory/Adapter');
const ARTIFACTORY_TOKEN_EXPIRATION_SEC = 60;
const IS_EXCEEDED_BUILD_TIME_FLAG = 'IS_EXCEEDED_BUILD_TIME_FLAG';
const CACHE_TTL = 1000 * 60 * 10;
const CHECK_CACHE_TTL = 1000 * 30;
const cacheConfig = new CacheConfig(CACHE_TTL, CHECK_CACHE_TTL);
const cache = new TypedCache(cacheConfig);
function isExceededBuildMinutesQuota(externalBag, done) {
    const who = isExceededBuildMinutesQuota.name;
    logger.trace(who, 'Inside');
    const bag = {
        jFrogPipelinesSessionId: externalBag.jFrogPipelinesSessionId,
        isExceededBuildTime: false,
        artifactoryAdapter: null,
        isCached: false
    };
    async.series([
        _getBlockActionFromCache.bind(null, bag),
        _getArtifactoryAdapter.bind(null, bag),
        _isBuildTimeBlocked.bind(null, bag),
        _cacheBlockAction.bind(null, bag)
    ], (err) => {
        if (err) {
            return done(err);
        }
        return done(null, bag.isExceededBuildTime);
    });
}
async function _getBlockActionFromCache(bag) {
    const who = self.name + '|' + _getBlockActionFromCache.name;
    const isExceededBuildTime = await cache.get(IS_EXCEEDED_BUILD_TIME_FLAG);
    if (!_.isNull(isExceededBuildTime) && !_.isUndefined(isExceededBuildTime)) {
        logger.trace(who, `${IS_EXCEEDED_BUILD_TIME_FLAG} is cached with status: ${isExceededBuildTime}`);
        bag.isExceededBuildTime = isExceededBuildTime;
        bag.isCached = true;
    }
    return bag;
}
function _getArtifactoryAdapter(bag, next) {
    if (bag.isCached)
        return next();
    const who = self.name + '|' + _getArtifactoryAdapter.name;
    logger.debug(who, 'Inside');
    if (!global.accessClient || !global.accessClientContext) {
        logger.error(who, 'Access client is required');
        return next();
    }
    const params = {
        who: who,
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId,
        accessClient: global.accessClient,
        accessClientContext: global.accessClientContext,
        tokenExpirationIntervalSecs: ARTIFACTORY_TOKEN_EXPIRATION_SEC
    };
    createArtifactoryToken(params, (err, rtAccessToken) => {
        if (err) {
            logger.error(who, `Could not generate Artifactory token with error: ${util.inspect(err)}`);
            return next();
        }
        bag.artifactoryAdapter = new ArtifactoryAdapter(bag.jFrogPipelinesSessionId, rtAccessToken);
        return next();
    });
}
function _isBuildTimeBlocked(bag, next) {
    if (!bag.artifactoryAdapter)
        return next();
    if (bag.isCached)
        return next();
    const who = self.name + '|' + _isBuildTimeBlocked.name;
    logger.trace(who, 'Inside');
    bag.artifactoryAdapter.getBlockUsages((err, response) => {
        if (err === 400) {
            logger.trace(who, `Block usage is not available for this environment: ${util.inspect(response)}`);
            return next();
        }
        if (err) {
            logger.error(who, `Retrieving blocking usages failed with status code: ${util.inspect(err)} ` +
                `and error: ${util.inspect(response)}`);
            return next();
        }
        const isExceededBuildTime = response?.buildTimeBlocked;
        if (!_.isBoolean(isExceededBuildTime)) {
            logger.error(who, `Expected 'isExceededBuildTime' type to be boolean and received type ${typeof isExceededBuildTime} ` +
                `in the response: ${util.inspect(response)}`);
            return next();
        }
        bag.isExceededBuildTime = isExceededBuildTime;
        return next();
    });
}
async function _cacheBlockAction(bag) {
    if (bag.isCached)
        return bag;
    const who = self.name + '|' + _cacheBlockAction.name;
    await cache.set(IS_EXCEEDED_BUILD_TIME_FLAG, bag.isExceededBuildTime);
    logger.trace(who, `Setting ${IS_EXCEEDED_BUILD_TIME_FLAG} status to be: ${bag.isExceededBuildTime}`);
    return bag;
}
//# sourceMappingURL=IsExceededBuildMinutesQuota.js.map