'use strict';
var self = checkExecutionCapacity;
module.exports = self;
var _ = require('underscore');
var async = require('async');
function checkExecutionCapacity(externalBag, callback) {
    var who = externalBag.who + '|' + checkExecutionCapacity.name;
    logger.debug(who, 'Inside');
    var bag = {
        who: who,
        suAdapter: externalBag.suAdapter,
        stepTriggerLicenseCheckHueristicsEnabled: externalBag.stepTriggerLicenseCheckHueristicsEnabled,
        statusCodes: global.systemCodesByGroup.status,
        nodePools: [],
        nodes: [],
        dynamicNodePools: [],
        staticNodePools: [],
        availableNodePools: [],
        allowedLicenseCapacity: null,
        hasRemainingCapacity: false,
        resBody: {}
    };
    async.series([
        _checkInputParams.bind(null, bag),
        _checkLicenseCapacity.bind(null, bag),
        _getStepTriggerSystemService.bind(null, bag),
        _getStepTriggerSystemServiceInstances.bind(null, bag),
        _getNodePools.bind(null, bag),
        _getNodes.bind(null, bag),
        _filterAvailableNodePools.bind(null, bag),
        _setResBody.bind(null, bag)
    ], function (err) {
        if (err)
            return callback(err);
        return callback(null, bag.resBody);
    });
}
function _checkInputParams(bag, next) {
    var who = bag.who + '|' + _checkInputParams.name;
    logger.debug(who, 'Inside');
    if (!bag.suAdapter)
        return next(new Error(':suAdapter is missing'));
    return next();
}
function _checkLicenseCapacity(bag, next) {
    var who = bag.who + '|' + _checkLicenseCapacity.name;
    logger.debug(who, 'Inside');
    bag.suAdapter.getExecutionNodes(function (err, results) {
        if (err) {
            logger.warn('failed to check remaining capacity:', err.message);
            return next(err);
        }
        else if (results.available > 0) {
            bag.hasRemainingCapacity = true;
            bag.availableCapacity = results.available;
        }
        return next();
    });
}
function _getStepTriggerSystemService(bag, next) {
    if (bag.stepTriggerLicenseCheckHueristicsEnabled !== true)
        return next();
    const who = bag.who + '|' + _checkLicenseCapacity.name;
    logger.debug(who, 'Inside');
    const query = 'names=' + global.msName.toLowerCase();
    bag.suAdapter.getSystemServices(query, function (err, systemServices) {
        if (err) {
            logger.error('Failed to getSystemServices for names=' + global.msName +
                ' with error: ' + err.message);
            return next(err);
        }
        if (_.isEmpty(systemServices)) {
            logger.error('getSystemServices for names=' + global.msName +
                'is empty');
            return next(err);
        }
        bag.systemService = systemServices[0];
        return next();
    });
}
function _getStepTriggerSystemServiceInstances(bag, next) {
    if (bag.stepTriggerLicenseCheckHueristicsEnabled !== true)
        return next();
    const who = bag.who + '|' + _checkLicenseCapacity.name;
    logger.debug(who, 'Inside');
    const query = `systemServiceIds=${bag.systemService.id}`;
    bag.suAdapter.getSystemServiceInstances(query, function (err, systemServiceInstances) {
        if (err) {
            logger.warn('failed to get getSystemServiceInstances err:', err.message);
            return next(err);
        }
        bag.systemServiceInstancesCount = _.size(systemServiceInstances);
        if (bag.availableCapacity >= bag.systemServiceInstancesCount) {
            bag.allowedLicenseCapacity = Math.floor(bag.availableCapacity / bag.systemServiceInstancesCount);
        }
        else {
            bag.allowedLicenseCapacity = Math.min(1, bag.availableCapacity);
        }
        bag.hasRemainingCapacity = bag.allowedLicenseCapacity > 0;
        return next();
    });
}
function _getNodePools(bag, next) {
    var who = bag.who + '|' + _getNodePools.name;
    logger.debug(who, 'Inside');
    bag.suAdapter.getNodePools('', function (err, nodePools) {
        if (err) {
            logger.warn(who, util.format('Failed to getNodePools with error: %s', err));
            return next(true);
        }
        bag.nodePools = nodePools;
        bag.nodePoolsGroupedByIsOnDemand = _.groupBy(bag.nodePools, 'isOnDemand');
        bag.dynamicNodePools = bag.nodePoolsGroupedByIsOnDemand['true'];
        bag.staticNodePools = bag.nodePoolsGroupedByIsOnDemand['false'];
        return next();
    });
}
function _getNodes(bag, next) {
    if (_.isEmpty(bag.nodePools))
        return next();
    var who = bag.who + '|' + _getNodes.name;
    logger.debug(who, 'Inside');
    var query = 'nodePoolIds=' + _.pluck(bag.nodePools, 'id').join(',');
    bag.suAdapter.getNodes(query, function (err, nodes) {
        if (err) {
            logger.warn(who, util.format('Failed to getNodes for query: %s with error: %s', query, err));
            bag.error = true;
            return next();
        }
        bag.nodes = nodes;
        bag.nodesByNodePoolId = _.groupBy(nodes, 'nodePoolId');
        return next();
    });
}
function _filterAvailableNodePools(bag, next) {
    if (_.isEmpty(bag.nodePools))
        return next();
    var who = bag.who + '|' + _filterAvailableNodePools.name;
    logger.debug(who, 'Inside');
    _.each(bag.staticNodePools, function (nodePool) {
        var staticNodes = bag.nodesByNodePoolId[nodePool.id] || [];
        var areNodesAvailable = !_.isEmpty(staticNodes);
        var areSuccessNodes = _.any(staticNodes, function (node) {
            return node.statusCode === bag.statusCodes.success.code;
        });
        var areFailedNodes = _.all(staticNodes, function (node) {
            return node.statusCode === bag.statusCodes.failure.code;
        });
        if (areSuccessNodes || !areNodesAvailable || areFailedNodes)
            bag.availableNodePools.push(nodePool);
    });
    _.each(bag.dynamicNodePools, function (nodePool) {
        var maxNodesAllowed = nodePool.numberOfNodes;
        var dynamicNodes = bag.nodesByNodePoolId[nodePool.id] || [];
        var processingNodes = _.filter(dynamicNodes, function (node) {
            return node.statusCode === bag.statusCodes.processing.code;
        });
        if (maxNodesAllowed > 0 && maxNodesAllowed > processingNodes.length)
            bag.availableNodePools.push(nodePool);
    });
    return next();
}
function _setResBody(bag, next) {
    var who = bag.who + '|' + _setResBody.name;
    logger.debug(who, 'Inside');
    bag.resBody.hasRemainingCapacity = bag.hasRemainingCapacity;
    bag.resBody.availableNodePools = bag.availableNodePools;
    bag.resBody.stepTriggerHueristics = {
        allowedCapacity: bag.allowedLicenseCapacity,
        enabled: bag.stepTriggerLicenseCheckHueristicsEnabled,
        hasRemainingCapacity: bag.hasRemainingCapacity
    };
    bag.resBody.allNodePools = bag.nodePools;
    return next();
}
//# sourceMappingURL=checkExecutionCapacity.js.map