'use strict';
var self = checkGroupExecutionEligibility;
module.exports = self;
var _ = require('underscore');
var async = require('async');
const { CALL_OBJECT_LIMIT_FOR_BATCH, CALL_OBJECT_LIMIT } = require('../constants.js');
var getInputStepsForStep = require('../getInputStepsForStep.js');
function checkGroupExecutionEligibility(externalBag, callback) {
    var who = externalBag.who + '|' + checkGroupExecutionEligibility.name;
    logger.debug(who, 'Inside');
    var bag = {
        who: who,
        suAdapter: externalBag.suAdapter,
        statusCodes: global.systemCodesByGroup.status,
        group: externalBag.group,
        runStepConnections: externalBag.runStepConnections,
        runResourceVersions: externalBag.runResourceVersions,
        actualRunStepConnections: [],
        actualRunResourceVersions: [],
        actualRunResourceVersionsById: {},
        areMissingConnections: true,
        stepIds: [],
        stepIdsSet: new Set([]),
        canQueue: false
    };
    async.series([
        _checkInputParams.bind(null, bag),
        _getRunStepConnectionsForGroup.bind(null, bag),
        _getRunResourceVersionsForGroup.bind(null, bag),
        _checkMissingConnections.bind(null, bag),
        _getInputDependencies.bind(null, bag),
        _checkInputDependencies.bind(null, bag)
    ], function (err) {
        return callback(err, bag.canQueue);
    });
}
function _checkInputParams(bag, next) {
    var who = bag.who + '|' + _checkInputParams.name;
    logger.debug(who, 'Inside');
    if (!bag.suAdapter)
        return next(new Error(':suAdapter is missing'));
    if (!bag.group)
        return next(new Error(':group is missing'));
    if (!bag.runStepConnections)
        return next(new Error(':runStepConnections is missing'));
    if (!bag.runResourceVersions)
        return next(new Error(':runResourceVersions is missing'));
    bag.stepIds = _.pluck(bag.group.groupStepIds, 'stepId');
    bag.stepIdsSet = new Set(bag.stepIds);
    bag.runId = bag.group.runId;
    return next();
}
function _getRunStepConnectionsForGroup(bag, next) {
    if (_.isEmpty(bag.stepIds))
        return next();
    var who = bag.who + '|' + _getRunStepConnectionsForGroup.name;
    logger.debug(who, 'Inside');
    var query = util.format('runIds=%s&sortOrder=1&sortBy=id', bag.runId);
    var fetchedAllRunStepConnections = false;
    var skip = 0;
    bag.actualRunStepConnections = [];
    async.whilst(function () {
        return !fetchedAllRunStepConnections;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        bag.suAdapter.getRunStepConnections(currentQuery, function (err, runStepConnections) {
            if (err) {
                logger.warn(who, util.format('Failed to getRunStepConnections for query: %s with err: %s', query, err));
                return done(true);
            }
            if (runStepConnections.length === CALL_OBJECT_LIMIT) {
                skip = skip + CALL_OBJECT_LIMIT;
            }
            else {
                fetchedAllRunStepConnections = true;
            }
            bag.actualRunStepConnections =
                bag.actualRunStepConnections.concat(runStepConnections);
            return done();
        });
    }, function (err) {
        return next(err);
    });
}
function _getRunResourceVersionsForGroup(bag, next) {
    if (_.isEmpty(bag.actualRunStepConnections))
        return next();
    var who = bag.who + '|' + _getRunResourceVersionsForGroup.name;
    logger.debug(who, 'Inside');
    var runResourceVersionIds = _.uniq(_.compact(_.pluck(bag.actualRunStepConnections, 'operationRunResourceVersionId')));
    if (_.isEmpty(runResourceVersionIds))
        return next();
    bag.actualRunResourceVersions = [];
    bag.actualRunResourceVersionsById = {};
    var runResourceVersionIdGroups = [];
    var i = 0;
    while (i < runResourceVersionIds.length) {
        runResourceVersionIdGroups.push(runResourceVersionIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var runResourceVersions = [];
    async.eachLimit(runResourceVersionIdGroups, 10, function (runResourceVersionIdGroup, done) {
        var query = util.format('runResourceVersionIds=%s&limit=%s', runResourceVersionIdGroup.join(','), CALL_OBJECT_LIMIT_FOR_BATCH);
        bag.suAdapter.getRunResourceVersionsLight(query, function (err, response) {
            if (err) {
                logger.warn(who, 'getRunResourceVersions for query:', query, 'returned error:', err);
                return done(true);
            }
            runResourceVersions = runResourceVersions.concat(response);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        bag.actualRunResourceVersions = runResourceVersions;
        bag.actualRunResourceVersionsById = _.indexBy(runResourceVersions, 'id');
        return next();
    });
}
function _checkMissingConnections(bag, next) {
    var who = bag.who + '|' + _checkMissingConnections.name;
    logger.debug(who, 'Inside');
    var runStepConnectionIds = _.pluck(bag.runStepConnections, 'id');
    var actualRunStepConnectionsId = _.pluck(bag.actualRunStepConnections, 'id');
    var missingConnectionIds = _.difference(actualRunStepConnectionsId, runStepConnectionIds);
    var msg = '';
    if (!_.isEmpty(missingConnectionIds)) {
        msg = util.format('Step trigger tried to queued to group %s ' +
            'with missing connections = %s', missingConnectionIds.join(','));
        logger.warn(who, msg);
        return next();
    }
    var runResourceVersionIds = _.pluck(bag.runResourceVersions, 'id');
    var actualRunResourceVersionIds = _.pluck(bag.actualRunResourceVersions, 'id');
    var missingRunResourceVersionIds = _.difference(actualRunResourceVersionIds, runResourceVersionIds);
    if (!_.isEmpty(missingRunResourceVersionIds)) {
        msg = util.format('Step trigger tried to queued to group %s ' +
            'with missing runResourceVersionIds = %s', missingRunResourceVersionIds.join(','));
        logger.warn(who, msg);
        return next();
    }
    bag.areMissingConnections = false;
    return next();
}
function _getInputDependencies(bag, next) {
    if (bag.areMissingConnections)
        return next();
    var who = bag.who + '|' + _getInputDependencies.name;
    logger.debug(who, 'Inside');
    bag.groupInputStepIds = [];
    async.eachLimit(bag.stepIds, 10, function (stepId, nextStepId) {
        var _bag = {
            who: who,
            suAdapter: bag.suAdapter,
            step: {
                id: stepId
            },
            runSteps: [],
            runStepConnections: bag.actualRunStepConnections,
            runResourceVersionsById: bag.actualRunResourceVersionsById,
            jFrogPipelinesSessionId: ''
        };
        getInputStepsForStep(_bag, function (err, result) {
            if (err)
                return nextStepId(err);
            bag.groupInputStepIds.push(...result.inputStepIds);
            return nextStepId();
        });
    }, function (err) {
        if (err)
            return next(err);
        bag.groupInputStepIds = _.uniq(bag.groupInputStepIds);
        bag.groupInputStepIds = _.reject(bag.groupInputStepIds, function (stepId) {
            return bag.stepIdsSet.has(stepId);
        });
        return next();
    });
}
function _checkInputDependencies(bag, next) {
    if (bag.areMissingConnections)
        return next();
    var who = bag.who + '|' + _checkInputDependencies.name;
    logger.debug(who, 'Inside');
    if (_.isEmpty(bag.groupInputStepIds)) {
        logger.debug(who, 'Group doesn\'t have any input dependencies');
        bag.canQueue = true;
        return next();
    }
    var incompleteStatusCodes = [
        bag.statusCodes.waiting.code,
        bag.statusCodes.processing.code,
        bag.statusCodes.queued.code,
        bag.statusCodes.pendingApproval.code,
        bag.statusCodes.cancelling.code,
        bag.statusCodes.timingOut.code
    ];
    var i = 0;
    var stepIdBatches = [];
    while (i < bag.groupInputStepIds.length) {
        stepIdBatches.push(bag.groupInputStepIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var incompleteSteps = [];
    async.eachLimit(stepIdBatches, 10, function (stepIdBatch, done) {
        var query = util.format('statusCodes=%s&stepIds=%s', incompleteStatusCodes.join(','), stepIdBatch.join(','));
        bag.suAdapter.getSteps(query, function (err, steps) {
            if (err) {
                logger.warn(who, 'getSteps for query:', query, 'returned error:', err);
                return next(new ActErr(who, err.id, 'Failed to get steps for query: ' + query));
            }
            incompleteSteps = incompleteSteps.concat(steps);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        if (!_.isEmpty(incompleteSteps)) {
            incompleteSteps = _.map(incompleteSteps, function (step) {
                return util.format('%s/%s', step.id, step.name);
            });
            var msg = util.format('Step trigger tried to queued to group %s ' +
                'with unmet input dependencies = %s', bag.stepIds.join(','), incompleteSteps.join(','));
            logger.warn(who, msg);
            return next();
        }
        bag.canQueue = true;
        return next();
    });
}
//# sourceMappingURL=checkGroupExecutionEligibility.js.map