'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
var self = prepData;
module.exports = self;
var _ = require('underscore');
var async = require('async');
var { getElapsedTimeSinceStartTime } = require('./utils');
var getDownstreamStepIds = require('./getDownStreamStepIds.js');
const { waitingConditions } = require('./constants');
var { CALL_OBJECT_LIMIT, CALL_OBJECT_LIMIT_FOR_BATCH } = require('../constants.js');
const isExceededBuildMinutesQuota = require('../IsExceededBuildMinutesQuota');
const BatchStepletConsoleAdapter = require('../BatchStepletConsoleAdapter');
const { DataAdapterFactory } = require('../modules/dataAdapter/DataAdapterFactory');
const { AffinityGroupManager } = require('../modules/affinityGroup/AffinityGroupManager');
const logger_1 = require("@jfrog-pipelines/logger");
function prepData(externalBag, callback) {
    var who = externalBag.who + '|' + prepData.name;
    logger.debug(who, 'Inside');
    var bag = {
        who: who,
        suAdapter: externalBag.suAdapter,
        statusCodes: global.systemCodesByGroup.status,
        hasRemainingCapacity: externalBag.hasRemainingCapacity,
        availableNodePools: externalBag.availableNodePools,
        allNodePools: externalBag.allNodePools,
        defaultPriority: externalBag.defaultPriority,
        jFrogPipelinesSessionId: externalBag.jFrogPipelinesSessionId,
        skipByBuildLimitCrossed: false,
        externalStepTypes: ['Jenkins'],
        rejectedGroups: externalBag.rejectedGroups,
        incompleteSteplets: [],
        incompleteSteps: [],
        waitingGroups: [],
        processingGroups: [],
        pipelines: [],
        runs: [],
        runStepConnections: [],
        runResourceVersionsByRunId: {},
        result: {},
        useRunAdjacency: externalBag.useRunAdjacency,
    };
    let startTime = Date.now();
    let tasks = [];
    if (bag.useRunAdjacency === true) {
        tasks = [
            _checkInputParams.bind(null, bag),
            _mapAffinityGroupsFromRunAdjacency.bind(null, bag),
            _isExceededBuildMinutesQuota.bind(null, bag),
            _enforceBuildMinutesLimit.bind(null, bag),
            _getNodePools.bind(null, bag),
            _rejectWaitingGroupsByNodePools.bind(null, bag),
            _prioritizeWaitingGroups.bind(null, bag),
            _getGroupsProjects.bind(null, bag),
            _getProjectsAllowedNodePools.bind(null, bag),
            _getOrphanedNodes.bind(null, bag),
            _setResult.bind(null, bag)
        ];
        ;
    }
    else {
        tasks = [
            _checkInputParams.bind(null, bag),
            _getIncompleteSteplets.bind(null, bag),
            _getIncompleteSteps.bind(null, bag),
            _mapAffinityGroups.bind(null, bag),
            _isExceededBuildMinutesQuota.bind(null, bag),
            _enforceBuildMinutesLimit.bind(null, bag),
            _getPipelines.bind(null, bag),
            _getWaitingGroupRuns.bind(null, bag),
            _flagChronologicalRuns.bind(null, bag),
            _findDependencyModeResourceIds.bind(null, bag),
            _findDependencyModeConnections.bind(null, bag),
            _getRunsForDependencyModes.bind(null, bag),
            _getLatestCompleteStatusForWaitOnParentSuccess.bind(null, bag),
            _getRunStepConnections.bind(null, bag),
            _getRunResourceVersions.bind(null, bag),
            _flagChronologicalGroupsByRun.bind(null, bag),
            _rejectDownstreamGroups.bind(null, bag),
            _getNodePools.bind(null, bag),
            _rejectWaitingGroupsByNodePools.bind(null, bag),
            _prioritizeWaitingGroups.bind(null, bag),
            _flagChronologicalGroups.bind(null, bag),
            _rejectGroupsWithIntermediateStatuses.bind(null, bag),
            _getGroupsProjects.bind(null, bag),
            _getProjectsAllowedNodePools.bind(null, bag),
            _getOrphanedNodes.bind(null, bag),
            _setResult.bind(null, bag)
        ];
    }
    async.series(tasks, function (err) {
        let prepDataStatus = 'success';
        if (err)
            prepDataStatus = 'failed';
        var stopTime = Date.now();
        var timeDiff = new Date(stopTime - startTime);
        var finalTime = util.format("durationMins=%s - %s minutes, %s seconds, and " +
            "%s milliseconds", (timeDiff.getTime() / 1000 / 60), timeDiff.getMinutes(), timeDiff.getSeconds(), timeDiff.getMilliseconds());
        logger.debug(bag.who, 'Finished prepData -', finalTime);
        const eventData = {
            timestamp: new Date(stopTime).toISOString(),
            domain: global.msName,
            eventType: 'prepdata',
            status: prepDataStatus,
            durationMillis: timeDiff.valueOf(),
            pipelinesSessionId: bag.jFrogPipelinesSessionId
        };
        logger_1.LoggerFactory.logEvent(eventData);
        return callback(err, bag.result);
    });
}
function _checkInputParams(bag, next) {
    var who = bag.who + '|' + _checkInputParams.name;
    logger.debug(who, 'Inside');
    if (!bag.suAdapter)
        return next(new Error(':suAdapter is missing'));
    if (!bag.hasRemainingCapacity && _.isEmpty(bag.availableNodePools))
        bag.skip = true;
    return next();
}
function _getIncompleteSteplets(bag, next) {
    if (bag.skip)
        return next();
    var who = bag.who + '|' + _getIncompleteSteplets.name;
    logger.debug(who, 'Inside');
    var incompleteStatusCodes = [
        bag.statusCodes.waiting.code,
        bag.statusCodes.processing.code,
        bag.statusCodes.queued.code
    ];
    var query = util.format('statusCodes=%s&sortOrder=1&sortBy=triggeredAt', incompleteStatusCodes.join(','));
    var fetchedAllIncompleteSteplets = false;
    var skip = 0;
    async.whilst(function () {
        return !fetchedAllIncompleteSteplets;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        bag.suAdapter.getSteplets(currentQuery, function (err, steplets) {
            if (err) {
                logger.warn(who, util.format('Failed to getSteplets for query: %s with err: %s', currentQuery, err));
                return done(true);
            }
            if (steplets.length === CALL_OBJECT_LIMIT) {
                skip = skip + CALL_OBJECT_LIMIT;
            }
            else {
                fetchedAllIncompleteSteplets = true;
            }
            bag.incompleteSteplets = bag.incompleteSteplets.concat(steplets);
            return done();
        });
    }, function (err) {
        bag.incompleteStepletsById = _.indexBy(bag.incompleteSteplets, 'id');
        bag.incompleteStepletsByStepId = _.groupBy(bag.incompleteSteplets, 'stepId');
        return next(err);
    });
}
function _getIncompleteSteps(bag, next) {
    if (bag.skip)
        return next();
    var who = bag.who + '|' + _getIncompleteSteps.name;
    logger.debug(who, 'Inside');
    var incompleteStepIds = _.uniq(_.keys(bag.incompleteStepletsByStepId));
    if (_.isEmpty(incompleteStepIds))
        return next();
    var incompleteStatusCodes = [
        bag.statusCodes.waiting.code,
        bag.statusCodes.processing.code,
        bag.statusCodes.queued.code,
        bag.statusCodes.pendingApproval.code,
        bag.statusCodes.cancelling.code,
        bag.statusCodes.timingOut.code
    ];
    var query = util.format('statusCodes=%s&sortOrder=1&sortBy=triggeredAt', incompleteStatusCodes.join(','));
    var fetchedAllIncompleteSteps = false;
    var skip = 0;
    async.whilst(function () {
        return !fetchedAllIncompleteSteps;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        bag.suAdapter.getSteps(currentQuery, function (err, steps) {
            if (err) {
                logger.warn(who, util.format('Failed to getSteps for query: %s with err: %s', query, err));
                return done(true);
            }
            if (steps.length === CALL_OBJECT_LIMIT) {
                skip = skip + CALL_OBJECT_LIMIT;
            }
            else {
                fetchedAllIncompleteSteps = true;
            }
            bag.incompleteSteps = bag.incompleteSteps.concat(steps);
            return done();
        });
    }, function (err) {
        bag.incompleteStepsByRunId = _.groupBy(bag.incompleteSteps, 'runId');
        bag.incompleteStepsById = _.indexBy(bag.incompleteSteps, 'id');
        return next(err);
    });
}
function _mapAffinityGroupsFromRunAdjacency(bag, next) {
    var who = bag.who + '|' + _mapAffinityGroupsFromRunAdjacency.name;
    logger.debug(who, 'Inside');
    const affinityGroupManager = new AffinityGroupManager(DataAdapterFactory.getApiDataAdapter());
    affinityGroupManager.getAffinityGroups().then((affinityGroups) => {
        bag.waitingGroups = affinityGroups;
        return next();
    }).catch(err => {
        return next(err);
    });
}
function _mapAffinityGroups(bag, next) {
    if (_.isEmpty(bag.incompleteSteps))
        return next();
    var who = bag.who + '|' + _mapAffinityGroups.name;
    logger.debug(who, 'Inside');
    var stepsByRunId = bag.incompleteStepsByRunId;
    var runIdList = _.uniq(_.pluck(bag.incompleteSteps, 'runId'));
    var affinityMap = {};
    _.each(runIdList, function (runId) {
        affinityMap[runId] = {};
        _.each(stepsByRunId[runId], function (step) {
            var stepPriority = step.configPropertyBag &&
                step.configPropertyBag.priority || bag.defaultPriority;
            var stepTriggeredAt = (new Date(step.triggeredAt)).getTime();
            var groupStartedAt;
            if (step.groupStartedAt)
                groupStartedAt = (new Date(step.groupStartedAt)).getTime();
            var stepAG = step.affinityGroup;
            var stepChrono = step.configPropertyBag &&
                step.configPropertyBag.chronological || false;
            var isExternal = _.contains(bag.externalStepTypes, global.systemCodesByCode[step.typeCode].name);
            var isMultiNodeStep = step.configPropertyBag &&
                step.configPropertyBag.multiNode;
            if (_.isEmpty(affinityMap[runId][stepAG])) {
                affinityMap[runId][stepAG] = {
                    affinityGroup: stepAG,
                    priority: stepPriority,
                    triggeredAt: stepTriggeredAt,
                    projectId: step.projectId,
                    pipelineId: step.pipelineId,
                    pipelineSourceId: step.pipelineSourceId,
                    groupInProgress: step.groupInProgress,
                    groupStartedAt: groupStartedAt,
                    requiresApproval: step.requiresApproval,
                    approvalRequestedAt: step.approvalRequestedAt,
                    runId: step.runId,
                    chronological: stepChrono,
                    groupStepIds: [{
                            stepId: step.id,
                            triggeredAt: stepTriggeredAt,
                            readyAt: step.readyAt,
                            priority: stepPriority
                        }],
                    isExternal: isExternal,
                    isMultiNodeStep: isMultiNodeStep,
                    nodePoolName: step.configPropertyBag &&
                        step.configPropertyBag.nodePool,
                    rejectionContext: {}
                };
                if (isMultiNodeStep)
                    __mapMultiNodeSteps(affinityMap[runId][stepAG], affinityMap, runId, bag);
            }
            else {
                if (stepPriority < affinityMap[runId][stepAG].priority)
                    affinityMap[runId][stepAG].priority = stepPriority;
                if (stepTriggeredAt < affinityMap[runId][stepAG].triggeredAt)
                    affinityMap[runId][stepAG].triggeredAt = stepTriggeredAt;
                affinityMap[runId][stepAG].groupInProgress =
                    affinityMap[runId][stepAG].groupInProgress ||
                        step.groupInProgress;
                affinityMap[runId][stepAG].chronological =
                    affinityMap[runId][stepAG].chronological || stepChrono;
                affinityMap[runId][stepAG].requiresApproval =
                    affinityMap[runId][stepAG].requiresApproval ||
                        step.requiresApproval;
                affinityMap[runId][stepAG].approvalRequestedAt =
                    affinityMap[runId][stepAG].approvalRequestedAt ||
                        step.approvalRequestedAt;
                affinityMap[runId][stepAG].groupStepIds.push({
                    stepId: step.id,
                    triggeredAt: stepTriggeredAt,
                    readyAt: step.readyAt,
                    priority: stepPriority
                });
            }
            if (!isMultiNodeStep) {
                let steplets = bag.incompleteStepletsByStepId[step.id] || [];
                if (steplets) {
                    _.each(steplets, function (steplet) {
                        __mapWaitingConditionsToGroup(affinityMap[runId][stepAG], steplet);
                    });
                }
            }
        });
        _.each(affinityMap[runId], function (group) {
            group.groupStepIds = _.sortBy(group.groupStepIds, function (stepObj) {
                return stepObj.triggeredAt;
            });
            if (group.groupInProgress)
                bag.processingGroups.push(group);
            else
                bag.waitingGroups.push(group);
        });
    });
    bag.waitingGroups = _.sortBy(bag.waitingGroups, function (group) {
        return group.triggeredAt;
    });
    const eventData = {
        timestamp: new Date().toISOString(),
        domain: global.msName,
        eventType: 'inflowsteps',
        count: bag.waitingGroups.length,
        pipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    logger_1.LoggerFactory.logEvent(eventData);
    return next();
}
function __mapWaitingConditionsToGroup(group, steplet) {
    let runtimePropertyBag = steplet.runtimePropertyBag || {};
    group.steplets = (group.steplets || []).concat(steplet);
    _.each(waitingConditions, function (condition) {
        if (_.has(runtimePropertyBag, condition)) {
            group[condition] = true;
        }
    });
}
function __mapMultiNodeSteps(multiNodeStepGroup, affinityMap, runId, bag) {
    var affinityGroupName = multiNodeStepGroup.affinityGroup;
    var affinityMapRunId = affinityMap[runId];
    var affinityMapRunIdAg = affinityMapRunId[affinityGroupName];
    var waitingStatusCode = bag.statusCodes.waiting.code;
    if (_.isEmpty(affinityMapRunIdAg))
        return;
    var skip = false;
    _.each(multiNodeStepGroup.groupStepIds, function (groupStepId) {
        if (skip)
            return;
        var stepId = groupStepId.stepId;
        var steplets = bag.incompleteStepletsByStepId[stepId] || [];
        _.each(steplets, function (steplet) {
            var agName = affinityGroupName + '_' + steplet.stepletNumber;
            affinityMapRunId[agName] = _.clone(affinityMapRunIdAg);
            if (steplet.statusCode === waitingStatusCode &&
                steplet.configPropertyBag &&
                !steplet.configPropertyBag.isQueuedToNodePool) {
                affinityMapRunId[agName].groupInProgress = false;
            }
            else
                affinityMapRunId[agName].groupInProgress = true;
            var oldAgObj = affinityMapRunId[agName].groupStepIds[0];
            affinityMapRunId[agName].groupStepIds = [{
                    stepId: oldAgObj.stepId,
                    triggeredAt: oldAgObj.triggeredAt,
                    readyAt: oldAgObj.readyAt,
                    priority: oldAgObj.priority,
                    stepletIds: [steplet.id]
                }];
            affinityMapRunId[agName].affinityGroup = agName;
            if (steplet.configPropertyBag && steplet.configPropertyBag.nodePool) {
                affinityMapRunId[agName].nodePoolName = steplet.configPropertyBag.nodePool;
            }
            __mapWaitingConditionsToGroup(affinityMapRunId[agName], steplet);
        });
    });
    if (skip)
        return;
    delete affinityMapRunId[affinityGroupName];
}
function _isExceededBuildMinutesQuota(bag, next) {
    const who = bag.who + '|' + _isExceededBuildMinutesQuota.name;
    logger.trace(who, 'Inside');
    const internalBag = {
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    isExceededBuildMinutesQuota(internalBag, function (err, result) {
        if (err) {
            logger.debug(who, `Failed to verify whether build minutes quota was exceeded with error: ${err.message}`);
            return next();
        }
        bag.skipByBuildLimitCrossed = result;
        return next();
    });
}
function _enforceBuildMinutesLimit(bag, next) {
    if (!bag.skipByBuildLimitCrossed)
        return next();
    const who = bag.who + '|' + _enforceBuildMinutesLimit.name;
    logger.trace(who, 'Inside');
    async.series([
        __postStepletCancellationMessages.bind(null, bag),
        __cancelIncompleteRuns.bind(null, bag)
    ], function (err) {
        if (err) {
            logger.warn(who, `Enforcing build minutes quota failed with error: ${err.message}`);
        }
        bag.waitingGroups = [];
        return next();
    });
}
function __postStepletCancellationMessages(bag, next) {
    const who = bag.who + '|' + __postStepletCancellationMessages.name;
    logger.trace(who, 'Inside');
    const incompleteStepletsByPipelineId = Object.entries(_.groupBy(bag.incompleteSteplets, 'pipelineId'));
    async.eachLimit(incompleteStepletsByPipelineId, 5, function (incompleteStepletsForPipelineId, done) {
        ___postStepCancelledConsoles({
            who: bag.who,
            pipelineId: incompleteStepletsForPipelineId[0],
            stepletIds: _.pluck(incompleteStepletsForPipelineId[1], 'id'),
            message: 'Cancelling step as build minutes quota was exceeded',
            jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
        });
        return done();
    }, function (err) {
        if (err) {
            logger.warn(who, `Failed to post step cancelled console message: ${err.message}`);
        }
        return next();
    });
}
function __cancelIncompleteRuns(bag, next) {
    const who = bag.who + '|' + __cancelIncompleteRuns.name;
    logger.trace(who, 'Inside');
    const incompleteGroups = bag.processingGroups.concat(bag.waitingGroups);
    const runIds = [...new Set(incompleteGroups.map(incompleteGroup => incompleteGroup.runId))];
    async.eachLimit(runIds, 5, async function (runId, done) {
        const cancelRunByIdAsync = util.promisify(bag.suAdapter.cancelRunById.bind(bag.suAdapter));
        try {
            await cancelRunByIdAsync(runId);
        }
        catch (err) {
            return done(err);
        }
        return done();
    }, function (err) {
        if (err) {
            logger.warn(who, 'Failed to cancel run', err);
            return next(err);
        }
        return next();
    });
}
function ___postStepCancelledConsoles(bag) {
    const who = bag.who + '|' + ___postStepCancelledConsoles.name;
    logger.debug(who, 'Inside');
    const consoleAdapter = new BatchStepletConsoleAdapter(bag.jFrogPipelinesSessionId, global.config.apiToken, bag.stepletIds, bag.pipelineId, global.systemConfig.jobConsoleBatchSize, global.systemConfig.jobConsoleBufferTimeInterval);
    consoleAdapter.openGrp('Step cancelled');
    consoleAdapter.openCmd('Info');
    consoleAdapter.publishMsg(bag.message);
    consoleAdapter.closeCmd(false);
    consoleAdapter.closeGrp(false);
}
function _getPipelines(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _getPipelines.name;
    logger.debug(who, 'Inside');
    var pipelineIds = _.uniq(_.pluck(bag.waitingGroups, 'pipelineId'));
    var query = util.format('pipelineIds=%s&sortOrder=1&sortBy=id', pipelineIds.join(','));
    var fetchedAllPipelines = false;
    var skip = 0;
    async.whilst(function () {
        return !fetchedAllPipelines;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        bag.suAdapter.getPipelines(currentQuery, function (err, pipelines) {
            if (err) {
                logger.warn(who, util.format('Failed to pipelines for query: %s with err: %s', query, err));
                return done(true);
            }
            if (pipelines.length === CALL_OBJECT_LIMIT)
                skip = skip + CALL_OBJECT_LIMIT;
            else
                fetchedAllPipelines = true;
            bag.pipelines = bag.pipelines.concat(pipelines);
            return done();
        });
    }, function (err) {
        bag.pipelinesById = _.indexBy(bag.pipelines, 'id');
        return next(err);
    });
}
function _getWaitingGroupRuns(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _getWaitingGroupRuns.name;
    logger.debug(who, 'Inside');
    var pipelineIds = _.uniq(_.pluck(bag.waitingGroups, 'pipelineId'));
    var statusCodes = [bag.statusCodes.ready.code,
        bag.statusCodes.processing.code];
    var query = util.format('pipelineIds=%s&statusCodes=%s&sortOrder=1&sortBy=id', pipelineIds.join(','), statusCodes.join(','));
    var fetchedAllRuns = false;
    var skip = 0;
    async.whilst(function () {
        return !fetchedAllRuns;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        bag.suAdapter.getRuns(currentQuery, function (err, runs) {
            if (err) {
                logger.warn(who, util.format('Failed to getRuns for query: %s with error: %s', query, err));
                return done(true);
            }
            if (runs.length === CALL_OBJECT_LIMIT)
                skip = skip + CALL_OBJECT_LIMIT;
            else
                fetchedAllRuns = true;
            bag.runs = bag.runs.concat(runs);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        bag.runsById = _.indexBy(bag.runs, 'id');
        bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
            return !bag.runsById[group.runId];
        });
        return next();
    });
}
function _flagChronologicalRuns(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _flagChronologicalRuns.name;
    logger.debug(who, 'Inside');
    var runsByPipelineId = _.groupBy(bag.runs, 'pipelineId');
    _.each(bag.pipelines, function (pipeline) {
        if (pipeline.ymlConfigPropertyBag &&
            pipeline.ymlConfigPropertyBag.chronological) {
            var runs = runsByPipelineId[pipeline.id];
            var lowestRun;
            _.each(runs, function (run) {
                if (!lowestRun) {
                    lowestRun = run;
                    return;
                }
                var waitingRun;
                if (run.runNumber < lowestRun.runNumber) {
                    waitingRun = lowestRun;
                    lowestRun = run;
                }
                else {
                    waitingRun = run;
                }
                bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
                    let shouldReject = (group.pipelineId === waitingRun.pipelineId &&
                        group.runId === waitingRun.id);
                    if (shouldReject) {
                        if (!group.waitingForPreviousRunToFinish) {
                            group.rejectionContext.waitingForPreviousRunToFinish = {
                                pipelineName: pipeline.name,
                                previousRunNumber: lowestRun.runNumber,
                                timestamp: new Date()
                            };
                            bag.rejectedGroups.waitingForPreviousRunToFinish.push(group);
                        }
                    }
                    return shouldReject;
                });
            });
        }
    });
    return next();
}
function _findDependencyModeResourceIds(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _findDependencyModeResourceIds.name;
    logger.debug(who, 'Inside');
    var pipelineIds = [];
    _.each(bag.runs, function (run) {
        var pipeline = bag.pipelinesById[run.pipelineId];
        if (!pipeline) {
            logger.warn(who, util.format('Pipeline %s not found for run %s', run.pipelineId, run.id));
            return;
        }
        if (!pipeline.ymlConfigPropertyBag ||
            !pipeline.ymlConfigPropertyBag.dependencyMode)
            return;
        var dependencyMode = pipeline.ymlConfigPropertyBag.dependencyMode;
        if (dependencyMode.waitOnParentComplete ||
            dependencyMode.waitOnParentSuccess ||
            dependencyMode.waitOnChildComplete)
            pipelineIds.push(pipeline.id);
    });
    pipelineIds = _.uniq(pipelineIds);
    if (_.isEmpty(pipelineIds))
        return next();
    var pipelineIdGroups = [];
    var i = 0;
    while (i < pipelineIds.length) {
        pipelineIdGroups.push(pipelineIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    bag.potentialConnectingResourceIds = [];
    async.eachLimit(pipelineIdGroups, 10, function (pipelineIdGroup, done) {
        var innerBag = {
            who: bag.who,
            suAdapter: bag.suAdapter,
            pipelineIds: pipelineIdGroup,
            pipelinesById: bag.pipelinesById,
            inResourceConnections: bag.inResourceConnections,
            potentialConnectingResourceIds: []
        };
        async.series([
            __getPipelineStepConnectionsForPipelineIds.bind(null, innerBag)
        ], function (err) {
            if (err)
                return done(err);
            bag.potentialConnectingResourceIds = _.uniq(bag.potentialConnectingResourceIds.concat(innerBag.potentialConnectingResourceIds));
            return done();
        });
    }, function (err) {
        return next(err);
    });
}
function __getPipelineStepConnectionsForPipelineIds(innerBag, next) {
    var who = innerBag.who + '|' +
        __getPipelineStepConnectionsForPipelineIds.name;
    logger.trace(who, 'Inside');
    var query = util.format('pipelineIds=%s&sortOrder=1&sortBy=id', innerBag.pipelineIds.join(','));
    var fetchedAllConnections = false;
    var skip = 0;
    var foundConnections = [];
    async.whilst(function () {
        return !fetchedAllConnections;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        innerBag.suAdapter.getPipelineStepConnections(currentQuery, function (err, connections) {
            if (err) {
                logger.warn(who, util.format('Failed to getPipelineStepConnections for query: %s ' +
                    'with error: %s', query, err));
                return done(true);
            }
            if (connections.length === CALL_OBJECT_LIMIT)
                skip = skip + CALL_OBJECT_LIMIT;
            else
                fetchedAllConnections = true;
            foundConnections = foundConnections.concat(connections);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        var potentialInputs = _.filter(foundConnections, function (connection) {
            return connection.operation === 'IN' && connection.isTrigger &&
                connection.operationResourceId;
        });
        var potentialOutputs = _.filter(foundConnections, function (connection) {
            return connection.operation === 'OUT' &&
                connection.operationResourceId;
        });
        _.each(innerBag.pipelineIds, function (pipelineId) {
            var pipeline = innerBag.pipelinesById[pipelineId];
            var dependencyMode = pipeline.ymlConfigPropertyBag.dependencyMode;
            if (dependencyMode.waitOnParentComplete ||
                dependencyMode.waitOnParentSuccess) {
                innerBag.potentialConnectingResourceIds =
                    innerBag.potentialConnectingResourceIds.concat(_.pluck(_.where(potentialInputs, { pipelineId: pipelineId }), 'operationResourceId'));
            }
            if (dependencyMode.waitOnChildComplete) {
                innerBag.potentialConnectingResourceIds =
                    innerBag.potentialConnectingResourceIds.concat(_.pluck(_.where(potentialOutputs, { pipelineId: pipelineId }), 'operationResourceId'));
            }
        });
        return next();
    });
}
function _findDependencyModeConnections(bag, next) {
    if (_.isEmpty(bag.potentialConnectingResourceIds))
        return next();
    var who = bag.who + '|' + _findDependencyModeConnections.name;
    logger.debug(who, 'Inside');
    var resourceIdGroups = [];
    var i = 0;
    while (i < bag.potentialConnectingResourceIds.length) {
        resourceIdGroups.push(bag.potentialConnectingResourceIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var inputConnections = [];
    var outputConnections = [];
    async.eachLimit(resourceIdGroups, 10, function (resourceIdGroup, done) {
        var innerBag = {
            who: bag.who,
            suAdapter: bag.suAdapter,
            resourceIds: resourceIdGroup,
            pipelinesById: bag.pipelinesById,
            outputConnections: [],
            inputConnections: []
        };
        async.series([
            __getPipelineStepConnectionsForResourceIds.bind(null, innerBag)
        ], function (err) {
            if (err)
                return done(err);
            inputConnections = inputConnections.concat(innerBag.inputConnections);
            outputConnections =
                outputConnections.concat(innerBag.outputConnections);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        var inputConnectionsByPipelineId = _.groupBy(inputConnections, 'pipelineId');
        var inputConnectionsByResourceId = _.groupBy(inputConnections, 'operationResourceId');
        var outputConnectionsByPipelineId = _.groupBy(outputConnections, 'pipelineId');
        var outputConnectionsByResourceId = _.groupBy(outputConnections, 'operationResourceId');
        bag.waitOnParentCompleteParentPipelineIdsByPipelineId = {};
        bag.waitOnParentSuccessParentPipelineIdsByPipelineId = {};
        bag.waitOnChildCompleteChildPipelineIdsByPipelineId = {};
        _.each(bag.pipelines, function (pipeline) {
            if (!pipeline.ymlConfigPropertyBag ||
                !pipeline.ymlConfigPropertyBag.dependencyMode)
                return;
            var dependencyMode = pipeline.ymlConfigPropertyBag.dependencyMode;
            if (dependencyMode.waitOnParentComplete ||
                dependencyMode.waitOnParentSuccess) {
                var inputResourceIds = _.pluck(inputConnectionsByPipelineId[pipeline.id], 'operationResourceId');
                var parentOutputPipelineIds = [];
                _.each(inputResourceIds, function (inputResourceId) {
                    parentOutputPipelineIds = _.union(parentOutputPipelineIds, _.pluck(outputConnectionsByResourceId[inputResourceId], 'pipelineId'));
                });
                parentOutputPipelineIds = _.without(parentOutputPipelineIds, pipeline.id);
                if (dependencyMode.waitOnParentComplete)
                    bag.waitOnParentCompleteParentPipelineIdsByPipelineId[pipeline.id] = parentOutputPipelineIds;
                if (dependencyMode.waitOnParentSuccess)
                    bag.waitOnParentSuccessParentPipelineIdsByPipelineId[pipeline.id] = parentOutputPipelineIds;
            }
            if (dependencyMode.waitOnChildComplete) {
                var outputResourceIds = _.pluck(outputConnectionsByPipelineId[pipeline.id], 'operationResourceId');
                var childInputPipelineIds = [];
                _.each(outputResourceIds, function (outputResourceId) {
                    childInputPipelineIds = _.union(childInputPipelineIds, _.pluck(inputConnectionsByResourceId[outputResourceId], 'pipelineId'));
                });
                bag.waitOnChildCompleteChildPipelineIdsByPipelineId[pipeline.id] =
                    _.without(childInputPipelineIds, pipeline.id);
            }
        });
        return next();
    });
}
function __getPipelineStepConnectionsForResourceIds(innerBag, next) {
    var who = innerBag.who + '|' +
        __getPipelineStepConnectionsForResourceIds.name;
    logger.trace(who, 'Inside');
    var query = util.format('operationResourceIds=%s&sortOrder=1&sortBy=id', innerBag.resourceIds.join(','));
    var fetchedAllConnections = false;
    var skip = 0;
    var foundConnections = [];
    async.whilst(function () {
        return !fetchedAllConnections;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        innerBag.suAdapter.getPipelineStepConnections(currentQuery, function (err, connections) {
            if (err) {
                logger.warn(who, util.format('Failed to getPipelineStepConnections for query: %s ' +
                    'with error: %s', query, err));
                return done(true);
            }
            if (connections.length === CALL_OBJECT_LIMIT)
                skip = skip + CALL_OBJECT_LIMIT;
            else
                fetchedAllConnections = true;
            foundConnections = foundConnections.concat(connections);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        var potentialInputsByResourceId = _.groupBy(_.filter(foundConnections, function (connection) {
            return connection.operation === 'IN' && connection.isTrigger &&
                connection.operationResourceId;
        }), 'operationResourceId');
        var potentialOutputsByResourceId = _.groupBy(_.filter(foundConnections, function (connection) {
            return connection.operation === 'OUT' &&
                connection.operationResourceId;
        }), 'operationResourceId');
        _.each(innerBag.resourceIds, function (resourceId) {
            var inputPipelineIds = _.uniq(_.pluck(potentialInputsByResourceId[resourceId], 'pipelineId'));
            var outputPipelineIds = _.uniq(_.pluck(potentialOutputsByResourceId[resourceId], 'pipelineId'));
            if (_.isEmpty(inputPipelineIds) || _.isEmpty(outputPipelineIds)) {
                potentialInputsByResourceId[resourceId] = [];
                potentialOutputsByResourceId[resourceId] = [];
            }
            else if (inputPipelineIds.length === outputPipelineIds.length) {
                var commonPipelineIds = _.intersection(inputPipelineIds, outputPipelineIds);
                if (inputPipelineIds.length === commonPipelineIds.length) {
                    potentialInputsByResourceId[resourceId] = [];
                    potentialOutputsByResourceId[resourceId] = [];
                }
            }
        });
        innerBag.inputConnections = _.flatten(_.values(potentialInputsByResourceId), true);
        innerBag.outputConnections = _.flatten(_.values(potentialOutputsByResourceId), true);
        return next();
    });
}
function _getRunsForDependencyModes(bag, next) {
    if (bag.skip)
        return next();
    if (_.isEmpty(bag.waitOnParentCompleteParentPipelineIdsByPipelineId) &&
        _.isEmpty(bag.waitOnParentSuccessParentPipelineIdsByPipelineId) &&
        _.isEmpty(bag.waitOnChildCompleteChildPipelineIdsByPipelineId))
        return next();
    var who = bag.who + '|' + _getRunsForDependencyModes.name;
    logger.trace(who, 'Inside');
    var pipelineIds = _.union(_.flatten(_.values(bag.waitOnParentCompleteParentPipelineIdsByPipelineId)), _.flatten(_.values(bag.waitOnParentSuccessParentPipelineIdsByPipelineId)), _.flatten(_.values(bag.waitOnChildCompleteChildPipelineIdsByPipelineId)));
    pipelineIds = _.uniq(pipelineIds);
    if (_.isEmpty(pipelineIds))
        return next();
    var pipelineIdGroups = [];
    var i = 0;
    while (i < pipelineIds.length) {
        pipelineIdGroups.push(pipelineIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var incompleteRuns = [];
    async.eachLimit(pipelineIdGroups, 10, function (pipelineIdGroup, done) {
        var innerBag = {
            who: bag.who,
            suAdapter: bag.suAdapter,
            pipelineIds: pipelineIdGroup,
            statusCodes: [global.systemCodesByName.creating.code,
                global.systemCodesByName.ready.code,
                global.systemCodesByName.waiting.code,
                global.systemCodesByName.processing.code],
            runs: []
        };
        async.series([
            __getRunsForPipelineIds.bind(null, innerBag)
        ], function (err) {
            if (err)
                return done(err);
            incompleteRuns = incompleteRuns.concat(innerBag.runs);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        var incompleteRunsByPipelineId = _.groupBy(incompleteRuns, 'pipelineId');
        var pipelineIdsWaiting = [];
        var waitingOnIncompleteParent = {};
        _.each(bag.waitOnParentCompleteParentPipelineIdsByPipelineId, function (parentPipelineIds, pipelineId) {
            var incompleteParent = _.some(parentPipelineIds, function (parentPipelineId) {
                let result = !_.isEmpty(incompleteRunsByPipelineId[parentPipelineId]);
                if (result) {
                    let parentPipeline = bag.pipelinesById[parentPipelineId];
                    if (parentPipeline)
                        waitingOnIncompleteParent[pipelineId] =
                            (waitingOnIncompleteParent[pipelineId] || []).concat(parentPipeline.name);
                }
                return result;
            });
            if (incompleteParent)
                pipelineIdsWaiting.push(parseInt(pipelineId, 10));
        });
        var waitingOnSuccessParent = {};
        _.each(bag.waitOnParentSuccessParentPipelineIdsByPipelineId, function (parentPipelineIds, pipelineId) {
            var processingParent = _.some(parentPipelineIds, function (parentPipelineId) {
                let result = _.findWhere(incompleteRunsByPipelineId[parentPipelineId], { statusCode: global.systemCodesByName.processing.code });
                if (result) {
                    let parentPipeline = bag.pipelinesById[parentPipelineId];
                    if (parentPipeline)
                        waitingOnSuccessParent[pipelineId] =
                            (waitingOnSuccessParent[pipelineId] || []).concat(parentPipeline.name);
                }
                return result;
            });
            if (processingParent)
                pipelineIdsWaiting.push(parseInt(pipelineId, 10));
        });
        var waitingOnCompleteChild = {};
        _.each(bag.waitOnChildCompleteChildPipelineIdsByPipelineId, function (childPipelineIds, pipelineId) {
            pipelineId = parseInt(pipelineId, 10);
            var incompleteChild = _.some(childPipelineIds, function (childPipelineId) {
                let result = false;
                let incompleteChildRuns = incompleteRunsByPipelineId[childPipelineId];
                if (_.isEmpty(incompleteRunsByPipelineId[childPipelineId]))
                    return false;
                if (incompleteChildRuns) {
                    let childPipeline = bag.pipelinesById[childPipelineId];
                    if (childPipeline) {
                        waitingOnCompleteChild[pipelineId] =
                            (waitingOnCompleteChild[pipelineId] || []).concat(childPipeline.name);
                    }
                    result = true;
                }
                if (_.contains(bag.waitOnParentCompleteParentPipelineIdsByPipelineId[childPipelineId], pipelineId) || _.contains(bag.waitOnParentSuccessParentPipelineIdsByPipelineId[childPipelineId], pipelineId))
                    return false;
                return result;
            });
            if (incompleteChild) {
                pipelineIdsWaiting.push(pipelineId);
            }
        });
        var waitingGroupRunsById = _.indexBy(bag.runs, 'id');
        bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
            if (waitingGroupRunsById[group.runId].statusCode ===
                global.systemCodesByName.processing.code)
                return false;
            let shouldReject = _.contains(pipelineIdsWaiting, group.pipelineId);
            if (shouldReject) {
                addRejectionContextForWaitingOnIncompleteParentGroups(_.uniq(waitingOnIncompleteParent), group);
                addRejectionContextForWaitingOnSuccessParentGroups(_.uniq(waitingOnSuccessParent), group);
                addRejectionContextForWaitingOnCompleteChildGroups(_.uniq(waitingOnCompleteChild), group, bag);
            }
            return shouldReject;
        });
        return next();
    });
    function addRejectionContextForWaitingOnSuccessParentGroups(waitingOnSuccessParent, group) {
        if (!_.isEmpty(waitingOnSuccessParent[group.pipelineId]) && !group.waitingForParentPipelineToSuccess) {
            group.rejectionContext.waitingForParentPipelineToSuccess = {
                pipelineNames: waitingOnSuccessParent[group.pipelineId],
                timestamp: new Date()
            };
            bag.rejectedGroups.waitingForParentPipelineToSuccess.push(group);
        }
    }
    function addRejectionContextForWaitingOnIncompleteParentGroups(waitingOnIncompleteParent, group) {
        if (!_.isEmpty(waitingOnIncompleteParent[group.pipelineId]) && !group.waitingForParentPipelineToComplete) {
            group.rejectionContext.waitingOnIncompleteParent = {
                pipelineNames: waitingOnIncompleteParent[group.pipelineId],
                timestamp: new Date()
            };
            bag.rejectedGroups.waitingForParentPipelineToComplete.push(group);
        }
    }
}
function addRejectionContextForWaitingOnCompleteChildGroups(waitingOnCompleteChild, group, bag) {
    if (!_.isEmpty(waitingOnCompleteChild[group.pipelineId]) && !group.waitingForChildPipelineToComplete) {
        group.rejectionContext.waitingForChildPipelineToComplete = {
            pipelineNames: waitingOnCompleteChild[group.pipelineId],
            timestamp: new Date()
        };
        bag.rejectedGroups.waitingForChildPipelineToComplete.push(group);
    }
}
function __getRunsForPipelineIds(innerBag, next) {
    var who = innerBag.who + '|' + __getRunsForPipelineIds.name;
    logger.trace(who, 'Inside');
    var query = util.format('pipelineIds=%s&statusCodes=%s&sortOrder=1&sortBy=id', innerBag.pipelineIds.join(','), innerBag.statusCodes.join(','));
    var fetchedAllRuns = false;
    var skip = 0;
    async.whilst(function () {
        return !fetchedAllRuns;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT, skip);
        innerBag.suAdapter.getRuns(currentQuery, function (err, runs) {
            if (err) {
                logger.warn(who, util.format('Failed to getRuns for query: %s with error: %s', query, err));
                return done(true);
            }
            if (runs.length === CALL_OBJECT_LIMIT)
                skip = skip + CALL_OBJECT_LIMIT;
            else
                fetchedAllRuns = true;
            innerBag.runs = innerBag.runs.concat(runs);
            return done();
        });
    }, function (err) {
        return next(err);
    });
}
function _getLatestCompleteStatusForWaitOnParentSuccess(bag, next) {
    if (_.isEmpty(bag.waitOnParentSuccessParentPipelineIdsByPipelineId))
        return next();
    var who = bag.who + '|' + _getLatestCompleteStatusForWaitOnParentSuccess.name;
    logger.trace(who, 'Inside');
    var pipelineIds = _.uniq(_.flatten(_.values(bag.waitOnParentSuccessParentPipelineIdsByPipelineId)));
    var pipelineIdGroups = [];
    var i = 0;
    while (i < pipelineIds.length) {
        pipelineIdGroups.push(pipelineIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var parentPipelines = [];
    async.eachLimit(pipelineIdGroups, 10, function (pipelineIdGroup, done) {
        var query = util.format('pipelineIds=%s&isDeleted=false&limit=%s', pipelineIdGroup.join(','), CALL_OBJECT_LIMIT_FOR_BATCH);
        bag.suAdapter.getPipelines(query, function (err, pipelines) {
            if (err) {
                logger.warn(who, util.format('Failed to getPipelines for query: %s with error: %s', query, err));
                return done(true);
            }
            parentPipelines = parentPipelines.concat(pipelines);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        var parentPipelinesById = _.indexBy(parentPipelines, 'id');
        var unsuccessfulParentPipelineIds = [];
        let waitingForParentPipelineToSuccess = {};
        _.each(bag.waitOnParentSuccessParentPipelineIdsByPipelineId, function (parentPipelineIds, pipelineId) {
            var unsuccessfulParent = _.some(parentPipelineIds, function (parentPipelineId) {
                var parentPipeline = parentPipelinesById[parentPipelineId];
                let result = parentPipeline &&
                    parentPipeline.latestCompletedRunId &&
                    (parentPipeline.latestSuccessfulRunId !==
                        parentPipeline.latestCompletedRunId);
                if (result) {
                    waitingForParentPipelineToSuccess[pipelineId] =
                        (waitingForParentPipelineToSuccess[pipelineId] || []).concat(parentPipeline.name);
                }
                return result;
            });
            if (unsuccessfulParent)
                unsuccessfulParentPipelineIds.push(parseInt(pipelineId, 10));
        });
        var waitingGroupRunsById = _.indexBy(bag.runs, 'id');
        bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
            if (waitingGroupRunsById[group.runId].statusCode ===
                global.systemCodesByName.processing.code)
                return false;
            let shouldReject = _.contains(unsuccessfulParentPipelineIds, group.pipelineId);
            if (shouldReject && !group.waitingForParentPipelineToSuccess) {
                group.rejectionContext.waitingForParentPipelineToSuccess = {
                    pipelineNames: waitingForParentPipelineToSuccess[group.pipelineId],
                    timestamp: new Date()
                };
                bag.rejectedGroups.waitingForParentPipelineToSuccess.push(group);
            }
            return shouldReject;
        });
        return next();
    });
}
function _getRunStepConnections(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _getRunStepConnections.name;
    logger.debug(who, 'Inside');
    var runIds = _.uniq(_.pluck(bag.waitingGroups, 'runId'));
    var query = util.format('runIds=%s&sortOrder=1&sortBy=id', runIds.join(','));
    var fetchedAllRunStepConnections = false;
    var skip = 0;
    var CALL_OBJECT_LIMIT_FOR_RSC = 300;
    async.whilst(function () {
        return !fetchedAllRunStepConnections;
    }, function (done) {
        var currentQuery = util.format('%s&limit=%s&skip=%s', query, CALL_OBJECT_LIMIT_FOR_RSC, skip);
        bag.suAdapter.getRunStepConnections(currentQuery, function (err, runStepConnections) {
            if (err) {
                logger.warn(who, util.format('Failed to getRunStepConnections for query: %s with err: %s', query, err));
                return done(true);
            }
            if (runStepConnections.length === CALL_OBJECT_LIMIT_FOR_RSC) {
                skip = skip + CALL_OBJECT_LIMIT_FOR_RSC;
            }
            else {
                fetchedAllRunStepConnections = true;
            }
            bag.runStepConnections =
                bag.runStepConnections.concat(runStepConnections);
            return done();
        });
    }, function (err) {
        bag.runStepConnectionsByStepId = _.groupBy(bag.runStepConnections, 'stepId');
        bag.runStepConnectionsByRunId = _.groupBy(bag.runStepConnections, 'runId');
        return next(err);
    });
}
function _getRunResourceVersions(bag, next) {
    if (_.isEmpty(bag.runStepConnections))
        return next();
    var who = bag.who + '|' + _getRunResourceVersions.name;
    logger.debug(who, 'Inside');
    var runResourceVersionIds = _.uniq(_.compact(_.pluck(bag.runStepConnections, 'operationRunResourceVersionId')));
    if (_.isEmpty(runResourceVersionIds))
        return next();
    var runResourceVersionIdGroups = [];
    var i = 0;
    while (i < runResourceVersionIds.length) {
        runResourceVersionIdGroups.push(runResourceVersionIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var runResourceVersions = [];
    async.eachLimit(runResourceVersionIdGroups, 10, function (runResourceVersionIdGroup, done) {
        var query = util.format('runResourceVersionIds=%s&limit=%s', runResourceVersionIdGroup.join(','), CALL_OBJECT_LIMIT_FOR_BATCH);
        bag.suAdapter.getRunResourceVersionsLight(query, function (err, response) {
            if (err) {
                logger.warn(who, 'getRunResourceVersions for query:', query, 'returned error:', err);
                return done(true);
            }
            runResourceVersions = runResourceVersions.concat(response);
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        bag.runResourceVersionsById = _.indexBy(runResourceVersions, 'id');
        bag.runResourceVersionsByRunId = _.groupBy(runResourceVersions, 'runId');
        return next();
    });
}
function _flagChronologicalGroupsByRun(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _flagChronologicalGroupsByRun.name;
    logger.debug(who, 'Inside');
    var processingChronoGroups = _.filter(bag.processingGroups, function (group) {
        return group.chronological;
    });
    var processingChronoGroupsByPipelineIdAffinityGroupName = _.groupBy(processingChronoGroups, function (group) {
        return group.pipelineId + '_' + group.affinityGroup;
    });
    var oldestChronoGroupRunIds = {};
    bag.rejectableChronoGroups = [];
    _.each(bag.waitingGroups, function (group) {
        if (group.chronological) {
            var pipelineIdAndName = group.pipelineId + '_' + group.affinityGroup;
            if (!oldestChronoGroupRunIds[pipelineIdAndName]) {
                oldestChronoGroupRunIds[pipelineIdAndName] = group.runId;
                return;
            }
            var oldRunId = oldestChronoGroupRunIds[pipelineIdAndName];
            if (group.runId < oldRunId)
                oldestChronoGroupRunIds[pipelineIdAndName] = group.runId;
        }
    });
    _.each(bag.waitingGroups, function (group) {
        if (group.chronological) {
            var pipelineIdAndName = group.pipelineId + '_' + group.affinityGroup;
            if (!_.isEmpty(processingChronoGroupsByPipelineIdAffinityGroupName[pipelineIdAndName]) ||
                oldestChronoGroupRunIds[pipelineIdAndName] !== group.runId)
                bag.rejectableChronoGroups.push({
                    pipelineId: group.pipelineId,
                    affinityGroup: group.affinityGroup,
                    runId: group.runId
                });
        }
    });
    return next();
}
function _rejectDownstreamGroups(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _rejectDownstreamGroups.name;
    logger.debug(who, 'Inside');
    async.eachSeries(bag.waitingGroups.concat(bag.processingGroups), function (waitingGroup, nextWaitingGroup) {
        var downstreamStepIds = [];
        var runStepConnections = bag.runStepConnectionsByRunId[waitingGroup.runId];
        var runStepConnectionsByStepId = _.groupBy(runStepConnections, 'stepId');
        var runStepConnectionsByOperationStepId = _.groupBy(runStepConnections, 'operationStepId');
        var runStepConnectionsByResourceName = _.groupBy(runStepConnections, 'operationRunResourceVersionName');
        async.eachSeries(waitingGroup.groupStepIds, function (stepObj, nextStepObj) {
            var _bag = {
                who: who,
                stepId: stepObj.stepId,
                runResourceVersionsById: bag.runResourceVersionsById,
                runStepConnectionsByStepId: runStepConnectionsByStepId,
                runStepConnectionsByOperationStepId: runStepConnectionsByOperationStepId,
                runStepConnectionsByResourceName: runStepConnectionsByResourceName,
            };
            getDownstreamStepIds(_bag, function (downstreamStepIdsResult) {
                downstreamStepIds =
                    downstreamStepIds.concat(downstreamStepIdsResult);
                return process.nextTick(nextStepObj);
            });
        }, function () {
            if (_.isEmpty(downstreamStepIds))
                return process.nextTick(nextWaitingGroup);
            downstreamStepIds = _.difference(downstreamStepIds, _.pluck(waitingGroup.groupStepIds, 'stepId'));
            bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
                var overlap = _.intersection(downstreamStepIds, _.pluck(group.groupStepIds, 'stepId'));
                let shouldReject = !_.isEmpty(overlap);
                if (shouldReject && !group.waitingForInputStepsToComplete) {
                    let incompleteInputStepsByRunId = bag.incompleteStepsByRunId[group.runId] || [];
                    let incompleteStepNamesByRunId = _.indexBy(incompleteInputStepsByRunId, 'name');
                    let inputStepNamesByStepletId = {};
                    _.each(group.groupStepIds, function (groupStepObj) {
                        let step = bag.incompleteStepsById[groupStepObj.stepId] || {};
                        let steplets = bag.incompleteStepletsByStepId[step.id] || [];
                        if (steplets && incompleteInputStepsByRunId && step && step.configPropertyBag && step.configPropertyBag.inputSteps) {
                            _.each(step.configPropertyBag.inputSteps, function (inputStep) {
                                if (incompleteStepNamesByRunId[inputStep.name]) {
                                    _.each(steplets, function (steplet) {
                                        inputStepNamesByStepletId[steplet.id] =
                                            (inputStepNamesByStepletId[steplet.id] || new Set()).add(inputStep.name);
                                    });
                                }
                            });
                        }
                    });
                    if (!_.isEmpty(inputStepNamesByStepletId)) {
                        group.rejectionContext.waitingForInputStepsToComplete = {
                            inputStepNamesByStepletId: inputStepNamesByStepletId,
                            timestamp: new Date()
                        };
                        bag.rejectedGroups.waitingForInputStepsToComplete.push(group);
                    }
                }
                return shouldReject;
            });
            return process.nextTick(nextWaitingGroup);
        });
    }, function () {
        if (_.isEmpty(bag.rejectableChronoGroups))
            return next();
        bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
            if (!group.chronological)
                return false;
            let waitingForPreviousRunToFinish = {};
            var shouldReject = _.some(bag.rejectableChronoGroups, function (chronoGroup) {
                let result = (group.pipelineId === chronoGroup.pipelineId &&
                    group.affinityGroup === chronoGroup.affinityGroup &&
                    group.runId === chronoGroup.runId);
                if (result) {
                    let previousRun = bag.runsById[group.runId];
                    let pipeline = bag.pipelinesById[group.pipelineId];
                    if (previousRun && pipeline)
                        waitingForPreviousRunToFinish = {
                            previousRunNumber: previousRun.runNumber,
                            pipelineName: pipeline.name,
                            timestamp: new Date()
                        };
                }
                return result;
            });
            if (shouldReject && !group.waitingForPreviousRunToFinish) {
                group.rejectionContext.waitingForPreviousRunToFinish = waitingForPreviousRunToFinish;
                bag.rejectedGroups.waitingForPreviousRunToFinish.push(group);
            }
            return shouldReject;
        });
        return process.nextTick(next);
    });
}
function _prioritizeWaitingGroups(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _prioritizeWaitingGroups.name;
    logger.debug(who, 'Inside');
    bag.waitingGroups = _.sortBy(bag.waitingGroups, function (group) {
        return group.priority;
    });
    return next();
}
function _flagChronologicalGroups(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _flagChronologicalGroups.name;
    logger.debug(who, 'Inside');
    var waitingChronoGroupByPipelineId = {};
    bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
        if (!group.chronological)
            return false;
        let shouldReject = true;
        const existingChronoGroup = waitingChronoGroupByPipelineId[group.pipelineId];
        if (!existingChronoGroup) {
            waitingChronoGroupByPipelineId[group.pipelineId] = group;
            shouldReject = false;
        }
        else if (group.isMultiNodeStep && group.runId === existingChronoGroup.runId) {
            shouldReject = false;
        }
        if (shouldReject && !group.waitingForPreviousRunToFinish) {
            let pipeline = bag.pipelinesById[existingChronoGroup.pipelineId] || {};
            let run = bag.runsById[existingChronoGroup.runId] || {};
            group.rejectionContext.waitingForPreviousRunToFinish = {
                pipelineName: pipeline.name,
                previousRunNumber: run.runNumber,
                timestamp: new Date()
            };
            bag.rejectedGroups.waitingForPreviousRunToFinish.push(group);
        }
        return shouldReject;
    });
    return next();
}
function _rejectGroupsWithIntermediateStatuses(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _rejectGroupsWithIntermediateStatuses.name;
    logger.debug(who, 'Inside');
    bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
        return _.some(group.groupStepIds, function (stepObj) {
            var stepStatus = bag.incompleteStepsById[stepObj.stepId].statusCode;
            return stepStatus === bag.statusCodes.pendingApproval.code ||
                stepStatus === bag.statusCodes.cancelling.code ||
                stepStatus === bag.statusCodes.timingOut.code;
        });
    });
    return next();
}
function _rejectWaitingGroupsByNodePools(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _rejectWaitingGroupsByNodePools.name;
    logger.debug(who, 'Inside');
    var availableNodePoolsByName = _.indexBy(bag.availableNodePools, 'name');
    var allNodePoolsByName = _.indexBy(bag.allNodePools, 'name');
    bag.waitingGroups = _.reject(bag.waitingGroups, function (group) {
        if (group.isExternal)
            return false;
        if (group.requiresApproval && !group.approvalRequestedAt)
            return false;
        let shouldReject = _.isEmpty(availableNodePoolsByName[group.nodePoolName]);
        if (shouldReject && !group.waitingForFreeNodePoolCapacity) {
            let nodePool = allNodePoolsByName[group.nodePoolName];
            if (nodePool) {
                group.rejectionContext.waitingForFreeNodePoolCapacity = {
                    nodePoolName: group.nodePoolName,
                    numberOfNodes: nodePool.numberOfNodes,
                    timestamp: new Date()
                };
                bag.rejectedGroups.waitingForFreeNodePoolCapacity.push(group);
            }
        }
        return shouldReject;
    });
    return next();
}
function _getGroupsProjects(bag, next) {
    if (_.isEmpty(bag.waitingGroups) &&
        _.isEmpty(bag.processingGroups))
        return next();
    var who = bag.who + '|' + _getGroupsProjects.name;
    logger.debug(who, 'Inside');
    var projectIds = _.union(_.pluck(bag.processingGroups, 'projectId'), _.pluck(bag.waitingGroups, 'projectId'));
    var query = util.format('projectIds=%s', projectIds.join(','));
    bag.suAdapter.getProjects(query, function (err, projects) {
        if (err) {
            logger.warn(who, util.format('Failed to getProjects for query: %s with error: %s', query, err));
            return next(true);
        }
        bag.groupsProjects = projects;
        return next();
    });
}
function _getProjectsAllowedNodePools(bag, next) {
    if (_.isEmpty(bag.groupsProjects))
        return next();
    var who = bag.who + '|' + _getProjectsAllowedNodePools.name;
    logger.debug(who, 'Inside');
    bag.allowedNodePoolsByProject = {};
    async.eachLimit(bag.groupsProjects, 10, function (project, nextProject) {
        var query = util.format('projectKeys=%s', project.sourceId);
        bag.suAdapter.getNodePools(query, function (err, nodePools) {
            if (err) {
                logger.warn(who, util.format('Failed to getNodePools for project %s with error: %s', project.sourceId, err));
                return nextProject(err);
            }
            bag.allowedNodePoolsByProject[project.id] = nodePools;
            return nextProject();
        });
    }, function (err) {
        return next(err);
    });
}
function _getNodePools(bag, next) {
    var who = bag.who + '|' + _getNodePools.name;
    logger.debug(who, 'Inside');
    bag.suAdapter.getNodePools('', function (err, nodePools) {
        if (err) {
            logger.warn(who, util.format('Failed to getNodePools with error: %s', err));
            return next(true);
        }
        bag.nodePools = nodePools;
        return next();
    });
}
function _getOrphanedNodes(bag, next) {
    var who = bag.who + '|' + _getOrphanedNodes.name;
    logger.debug(who, 'Inside');
    bag.suAdapter.getNodes('', function (err, nodes) {
        if (err) {
            logger.warn(who, util.format('Failed to getNodes with error: %s', err));
            return next(true);
        }
        bag.orphanedOnDemandNodesByProjectId = _.groupBy(_.where(nodes, {
            isOnDemand: true,
            nodePoolId: null
        }), 'projectId');
        _.each(_.keys(bag.orphanedOnDemandNodesByProjectId), function (projectId) {
            bag.nodePools.push({
                name: 'orphaned#on-demand#nodepool',
                projectId: parseInt(projectId, 10),
                isOrphaned: true,
                isOnDemand: true
            });
        });
        return next();
    });
}
function _setResult(bag, next) {
    var who = bag.who + '|' + _setResult.name;
    logger.debug(who, 'Inside');
    bag.result = {
        incompleteStepsById: bag.incompleteStepsById,
        incompleteStepletsById: bag.incompleteStepletsById,
        incompleteStepletsByStepId: bag.incompleteStepletsByStepId,
        incompleteStepsByRunId: bag.incompleteStepsByRunId,
        waitingGroups: bag.waitingGroups,
        processingGroups: bag.processingGroups,
        pipelines: bag.pipelines,
        pipelinesById: bag.pipelinesById,
        runStepConnections: bag.runStepConnections,
        runStepConnectionsByStepId: bag.runStepConnectionsByStepId,
        runStepConnectionsByRunId: bag.runStepConnectionsByRunId,
        runResourceVersionsById: bag.runResourceVersionsById,
        runResourceVersionsByRunId: bag.runResourceVersionsByRunId,
        groupsProjects: bag.groupsProjects,
        allowedNodePoolsByProject: bag.allowedNodePoolsByProject,
        nodePools: bag.nodePools,
        orphanedOnDemandNodesByProjectId: bag.orphanedOnDemandNodesByProjectId,
        runsById: bag.runsById,
        rejectedGroups: bag.rejectedGroups
    };
    return next();
}
//# sourceMappingURL=prepData.js.map