'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
var self = queueExternalSteps;
module.exports = self;
var { getElapsedTimeSinceStartTime } = require('./utils');
var { CALL_OBJECT_LIMIT_FOR_BATCH } = require('../constants.js');
var LOCK_TIMEOUT_IN_MS = 1 * 60 * 1000;
var LOCK_TIMER_INTERVAL_IN_MS = 50 * 1000;
var _ = require('underscore');
var async = require('async');
var StepConsoleAdapter = require('../stepConsoleAdapter.js');
var getDownstreamStepIds = require('./getDownStreamStepIds.js');
var checkGroupExecutionEligibility = require('./checkGroupExecutionEligibility.js');
var Lock = require('../Lock.js');
const { DataAdapterFactory } = require('../modules/dataAdapter/DataAdapterFactory');
const logger_1 = require("@jfrog-pipelines/logger");
function queueExternalSteps(externalBag, callback) {
    var who = externalBag.who + '|' + queueExternalSteps.name;
    logger.debug(who, 'Inside');
    var bag = {
        who: who,
        suAdapter: externalBag.suAdapter,
        statusCodes: global.systemCodesByGroup.status,
        waitingGroups: externalBag.waitingGroups,
        groupsProjects: externalBag.groupsProjects,
        runStepConnections: externalBag.runStepConnections,
        runStepConnectionsByRunId: externalBag.runStepConnectionsByRunId,
        runStepConnectionsByStepId: externalBag.runStepConnectionsByStepId,
        runResourceVersionsByRunId: externalBag.runResourceVersionsByRunId,
        runResourceVersionsById: externalBag.runResourceVersionsById,
        jFrogPipelinesSessionId: externalBag.jFrogPipelinesSessionId,
        defaultPriority: externalBag.defaultPriority,
        rejectedGroups: externalBag.rejectedGroups,
        stepTriggerLicenseCheckLockRemoved: externalBag.stepTriggerLicenseCheckLockRemoved,
        stepTriggerHueristics: externalBag.stepTriggerHueristics,
        result: {},
        useRunAdjacency: externalBag.useRunAdjacency
    };
    let startTime = Date.now();
    async.series([
        _checkInputParams.bind(null, bag),
        _queueExternalSteps.bind(null, bag)
    ], function (err) {
        let finalTime = getElapsedTimeSinceStartTime(startTime);
        logger.debug(bag.who, 'Finished queueExternalSteps -', finalTime);
        return callback(err, bag.result);
    });
}
function _checkInputParams(bag, next) {
    var who = bag.who + '|' + _checkInputParams.name;
    logger.debug(who, 'Inside');
    if (!bag.suAdapter)
        return next(new Error(':suAdapter is missing'));
    return next();
}
function _queueExternalSteps(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _queueExternalSteps.name;
    logger.debug(who, 'Inside');
    let startTime = Date.now();
    var waitingExternalGroupsByProjectId = _.groupBy(_.where(bag.waitingGroups, { isExternal: true }), 'projectId');
    async.eachLimit(_.keys(waitingExternalGroupsByProjectId), 5, function (projectId, nextProject) {
        projectId = parseInt(projectId, 10);
        var innerBag = {
            who: bag.who + '|projectId:' + projectId,
            lock: {
                isValid: false,
                destroy: () => { },
                release: () => { }
            },
            suAdapter: bag.suAdapter,
            project: _.findWhere(bag.groupsProjects, { id: projectId }),
            waitingGroups: waitingExternalGroupsByProjectId[projectId] || [],
            statusCodes: bag.statusCodes,
            runStepConnections: bag.runStepConnections,
            runStepConnectionsByRunId: bag.runStepConnectionsByRunId,
            runResourceVersionsByRunId: bag.runResourceVersionsByRunId,
            runResourceVersionsById: bag.runResourceVersionsById,
            jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId,
            rejectedGroups: bag.rejectedGroups,
            stepTriggerLicenseCheckLockRemoved: bag.stepTriggerLicenseCheckLockRemoved,
            stepTriggerHueristics: bag.stepTriggerHueristics,
            error: false,
            stepsQueued: 0,
            useRunAdjacency: bag.useRunAdjacency,
        };
        const tasks = [
            __obtainExternalStepsLock.bind(null, innerBag),
        ];
        if (bag.useRunAdjacency) {
            tasks.push(__getWaitingExternalStepsByRunAdjacencies.bind(null, innerBag), __setExternalStepsReadyAtByRunAdjacencies.bind(null, innerBag));
        }
        else {
            tasks.push(__getWaitingExternalSteps.bind(null, innerBag), __setExternalStepsReadyAt.bind(null, innerBag));
        }
        tasks.push(__tryAndQueueExternalSteps.bind(null, innerBag), __releaseExternalStepsLock.bind(null, innerBag));
        async.series(tasks, function (err) {
            var stopTime = Date.now();
            var timeDiff = new Date(stopTime - startTime);
            if (err || innerBag.error)
                logger.warn(bag.who, 'Failed to process external steps for project', projectId, 'with error:', err || bag.error);
            else
                logger.debug(bag.who, 'Successfully processed external steps for project', projectId);
            const eventData = {
                timestamp: new Date(stopTime).toISOString(),
                domain: global.msName,
                eventType: "externalStepsQueued",
                queuedStepsCount: bag.stepsQueued,
                projectId: projectId,
                durationMillis: timeDiff.valueOf(),
                pipelinesSessionId: bag.jFrogPipelinesSessionId
            };
            logger_1.LoggerFactory.logEvent(eventData);
            if (bag.lock) {
                bag.lock.destroy();
                bag.lock = null;
            }
            return nextProject();
        });
    }, function () {
        return next();
    });
}
function __obtainExternalStepsLock(bag, next) {
    var who = bag.who + '|' + __obtainExternalStepsLock.name;
    logger.debug(who, 'Inside');
    var params = {
        name: global.msName + '-externalSteps-' + bag.project.id,
        lockTimeoutInMs: LOCK_TIMEOUT_IN_MS,
        lockTimerInMs: LOCK_TIMER_INTERVAL_IN_MS,
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    bag.lock = new Lock(params);
    bag.lock.obtain(function (err) {
        if (err)
            logger.debug('failed to obtain lock with name:', params.name);
        return next(err);
    });
}
function __getWaitingExternalStepsByRunAdjacencies(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __getWaitingExternalStepsByRunAdjacencies.name;
    logger.debug(who, 'Inside');
    const dataAdapter = DataAdapterFactory.getApiDataAdapter();
    async.eachLimit(bag.waitingGroups, 5, async (waitingGroup) => {
        try {
            let query = `affinityGroups=${waitingGroup.affinityGroup}&runIds=${waitingGroup.runId}`;
            query += '&include=id';
            const runAdjacencies = await dataAdapter.getRunAdjacencyByQuery(query);
            if (runAdjacencies.length === 0 || runAdjacencies.length !== waitingGroup.runAdjacencyList.length) {
                waitingGroup.rejectByRunAdjacency = true;
            }
        }
        catch (err) {
            throw err;
        }
    }, (err) => {
        if (err) {
            return next(err);
        }
        bag.waitingGroups = _.reject(bag.waitingGroups, (waitingGroup) => {
            return waitingGroup.rejectByRunAdjacency;
        });
        return next();
    });
}
function __getWaitingExternalSteps(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __getWaitingExternalSteps.name;
    logger.debug(who, 'Inside');
    var waitingGroupsByRunId = _.groupBy(bag.waitingGroups, 'runId');
    var stepIds = _.pluck(_.flatten(_.pluck(bag.waitingGroups, 'groupStepIds')), 'stepId');
    if (_.isEmpty(stepIds))
        return next();
    var incompleteStatusCodes = [
        bag.statusCodes.waiting.code,
        bag.statusCodes.processing.code,
        bag.statusCodes.queued.code
    ];
    var stepIdGroups = [];
    var i = 0;
    while (i < stepIds.length) {
        stepIdGroups.push(stepIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var currentStepRunIds = [];
    async.eachLimit(stepIdGroups, 10, function (stepIdGroup, done) {
        var query = util.format('stepIds=%s&statusCodes=%s&limit=%s', stepIdGroup.join(','), incompleteStatusCodes.join(','), CALL_OBJECT_LIMIT_FOR_BATCH);
        bag.suAdapter.getSteps(query, function (err, steps) {
            if (err) {
                bag.error = true;
                logger.warn(who, 'getSteps for query:', query, 'returned error:', err);
                return done();
            }
            var stepsInProgress = _.where(steps, { groupInProgress: true });
            _.each(stepsInProgress, function (step) {
                waitingGroupsByRunId[step.runId] = _.reject(waitingGroupsByRunId[step.runId], function (group) {
                    return group.affinityGroup === step.affinityGroup;
                });
            });
            currentStepRunIds = _.uniq(currentStepRunIds.concat(_.pluck(_.where(steps, { groupInProgress: false }), 'runId')));
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        var existingRunIds = _.keys(waitingGroupsByRunId);
        if (currentStepRunIds.length === existingRunIds.length)
            return next();
        var completedRunIds = _.difference(existingRunIds, currentStepRunIds);
        _.each(completedRunIds, function (runId) {
            delete waitingGroupsByRunId[runId];
        });
        bag.waitingGroups = _.flatten(_.values(waitingGroupsByRunId), true);
        return next();
    });
}
function __setExternalStepsReadyAtByRunAdjacencies(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __setExternalStepsReadyAtByRunAdjacencies.name;
    logger.debug(who, 'Inside');
    const dataAdapter = DataAdapterFactory.getApiDataAdapter();
    async.eachLimit(bag.waitingGroups, 5, async (waitingGroup) => {
        var alreadySet = _.some(waitingGroup.groupStepIds, function (step) {
            return step.readyAt;
        });
        if (alreadySet)
            return;
        const firstStepsList = waitingGroup.firstStepsList;
        if (firstStepsList.length === 0) {
            return;
        }
        try {
            await dataAdapter.updateStepsReadyAt(firstStepsList);
        }
        catch (err) {
            throw err;
        }
    }, (err) => {
        return next(err);
    });
}
function __setExternalStepsReadyAt(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __setExternalStepsReadyAt.name;
    logger.debug(who, 'Inside');
    async.eachLimit(bag.waitingGroups, 5, function (waitingGroup, done) {
        var alreadySet = _.some(waitingGroup.groupStepIds, function (step) {
            return step.readyAt;
        });
        if (alreadySet)
            return done();
        var groupStepIds = _.clone(waitingGroup.groupStepIds);
        var runStepConnections = bag.runStepConnectionsByRunId[waitingGroup.runId];
        var runStepConnectionsByStepId = _.groupBy(runStepConnections, 'stepId');
        var runStepConnectionsByOperationStepId = _.groupBy(runStepConnections, 'operationStepId');
        var runStepConnectionsByResourceName = _.groupBy(runStepConnections, 'operationRunResourceVersionName');
        async.eachSeries(waitingGroup.groupStepIds, function (stepObj, nextStepObj) {
            var _bag = {
                who: who,
                stepId: stepObj.stepId,
                runResourceVersionsById: bag.runResourceVersionsById,
                runStepConnectionsByStepId: runStepConnectionsByStepId,
                runStepConnectionsByOperationStepId: runStepConnectionsByOperationStepId,
                runStepConnectionsByResourceName: runStepConnectionsByResourceName,
            };
            getDownstreamStepIds(_bag, function (downstreamStepIds) {
                groupStepIds = _.reject(groupStepIds, function (groupStep) {
                    return _.contains(downstreamStepIds, groupStep.stepId);
                });
                return process.nextTick(nextStepObj);
            });
        }, function () {
            groupStepIds = _.sortBy(groupStepIds, function (groupStep) {
                return groupStep.priority;
            });
            var firstStep = _.first(groupStepIds);
            if (!firstStep)
                return done();
            var update = {
                readyAt: new Date()
            };
            bag.suAdapter.putStepById(firstStep.stepId, update, function (err) {
                if (err) {
                    logger.warn(who, 'Failed to putStepById for id:', firstStep.stepId, 'with error:', err);
                    return done(true);
                }
                return process.nextTick(done);
            });
        });
    }, function (err) {
        if (err)
            bag.error = true;
        return next();
    });
}
function __tryAndQueueExternalSteps(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __tryAndQueueExternalSteps.name;
    logger.debug(who, 'Inside');
    async.eachSeries(bag.waitingGroups, function (group, done) {
        if (!bag.lock.isValid) {
            logger.error(who, 'invalid lock, exiting');
            return next(true);
        }
        if (!group || _.isEmpty(group.groupStepIds))
            return process.nextTick(done);
        if (bag.stepTriggerHueristics.enabled && !bag.stepTriggerHueristics.hasRemainingCapacity) {
            logger.debug(who, `This stepTrigger utilised its allowed capacity`);
            return next();
        }
        let skipAquireLock = (bag.stepTriggerLicenseCheckLockRemoved === true ||
            (bag.stepTriggerHueristics.enabled === true && bag.stepTriggerHueristics.allowedCapacity >= 2));
        var seriesBag = {
            who: bag.who + '|stepId:' + group.groupStepIds[0].stepId,
            statusCodes: bag.statusCodes,
            suAdapter: bag.suAdapter,
            waitingGroups: bag.waitingGroups,
            group: group,
            hookIdsByStepId: {},
            runStepConnections: bag.runStepConnections,
            runResourceVersionsByRunId: bag.runResourceVersionsByRunId,
            project: bag.project,
            stepCheckLock: null,
            jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId,
            rejectedGroups: bag.rejectedGroups,
            skipAquireLock: skipAquireLock,
            error: false,
            useRunAdjacency: bag.useRunAdjacency
        };
        let tasks = [
            ___obtainLock.bind(null, seriesBag),
            ___checkLicenseCapacity.bind(null, seriesBag),
        ];
        if (!bag.useRunAdjacency) {
            tasks.push(___checkGroupExecutionEligibility.bind(null, seriesBag));
        }
        tasks.push(___updateExternalStep.bind(null, seriesBag), ___releaseLock.bind(null, seriesBag), ___queueExternalStep.bind(null, seriesBag), ___postExternalStepConsoles.bind(null, seriesBag));
        async.series(tasks, function (err) {
            if (!err && !seriesBag.group.reject)
                bag.stepsQueued += 1;
            logger.trace(who, 'queued group:', group.affinityGroup, 'with runId:', group.runId);
            if (bag.stepTriggerHueristics.enabled === true) {
                bag.stepTriggerHueristics.allowedCapacity -= 1;
                bag.stepTriggerHueristics.hasRemainingCapacity =
                    bag.stepTriggerHueristics.allowedCapacity > 0;
            }
            return process.nextTick(done);
        });
    }, function () {
        return next();
    });
}
function ___obtainLock(bag, next) {
    if (bag.skipAquireLock) {
        bag.stepCheckLock = {
            isValid: true,
        };
        return next();
    }
    var who = bag.who + '|' + ___obtainLock.name;
    logger.debug(who, 'Inside');
    var params = {
        name: global.msName + '-check-capacity',
        lockTimeoutInMs: 10000,
        lockTimerInMs: 8000,
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    bag.stepCheckLock = new Lock(params);
    var maxRetries = 30;
    var retries = 0;
    async.whilst(function () {
        return !bag.stepCheckLock.isValid;
    }, function (done) {
        bag.stepCheckLock.obtain(function (err) {
            if (err) {
                logger.trace('failed to obtain lock with name:', params.name);
                retries += 1;
                if (retries > maxRetries)
                    return done(err);
                else
                    return setTimeout(done, 1000);
            }
            else {
                return done();
            }
        });
    }, function (err) {
        return next(err);
    });
}
function ___checkLicenseCapacity(bag, next) {
    var who = bag.who + '|' + ___checkLicenseCapacity.name;
    logger.debug(who, 'Inside');
    bag.suAdapter.getExecutionNodes(function (err, results) {
        var rejectAll = false;
        let insufficientLicenseCapacity = false;
        if (err) {
            logger.debug('failed to check remaining capacity:', err.message);
            rejectAll = true;
        }
        else if (results.available <= 0) {
            rejectAll = true;
            insufficientLicenseCapacity = true;
        }
        if (rejectAll) {
            _.each(bag.waitingGroups, function (group) {
                group.reject = true;
                if (insufficientLicenseCapacity && !group.waitingForFreeLicenseCapacity) {
                    group.rejectionContext.waitingForFreeLicenseCapacity = {
                        maxCapacity: results.allowed,
                        timestamp: new Date()
                    };
                    bag.rejectedGroups.waitingForFreeLicenseCapacity.push(group);
                }
            });
            bag.group.reject = true;
        }
        return next();
    });
}
function ___checkGroupExecutionEligibility(bag, next) {
    if (bag.group.reject)
        return next();
    var who = bag.who + '|' + ___checkGroupExecutionEligibility.name;
    logger.debug(who, 'Inside');
    var internalBag = {
        who: who,
        suAdapter: bag.suAdapter,
        group: bag.group,
        runStepConnections: bag.runStepConnections || [],
        runResourceVersions: bag.runResourceVersionsByRunId[bag.group.runId] || [],
    };
    checkGroupExecutionEligibility(internalBag, function (err, canQueue) {
        if (err)
            logger.warn(who, err);
        if (!canQueue)
            bag.group.reject = true;
        return next();
    });
}
function ___updateExternalStep(bag, next) {
    if (bag.group.reject)
        return next();
    var who = bag.who + '|' + ___updateExternalStep.name;
    logger.debug(who, 'Inside');
    var update = {
        groupInProgress: true,
        groupStartedAt: new Date()
    };
    bag.group.groupStartedAt = update.groupStartedAt;
    async.eachLimit(bag.group.groupStepIds, 10, function (stepObj, done) {
        bag.suAdapter.putStepById(stepObj.stepId, update, function (err, step) {
            if (err) {
                logger.warn(who, 'Failed to putStepById for id:', stepObj.stepId, 'with error:', err);
                return done(true);
            }
            if (step.configPropertyBag && step.configPropertyBag.hookId)
                bag.hookIdsByStepId[step.id] = step.configPropertyBag.hookId;
            return done();
        });
    }, function (err) {
        if (err)
            bag.error = true;
        return next();
    });
}
function ___releaseLock(bag, next) {
    if (bag.skipAquireLock)
        return next();
    var who = bag.who + '|' + ___releaseLock.name;
    logger.debug(who, 'Inside');
    bag.stepCheckLock.release(function (err) {
        if (err)
            logger.debug(who, 'Failed to release lock with name: ' +
                bag.stepCheckLock.name);
        return next();
    });
}
function ___queueExternalStep(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    var who = bag.who + '|' + ___queueExternalStep.name;
    logger.debug(who, 'Inside');
    var stepId = _.first(_.pluck(bag.group.groupStepIds, 'stepId'));
    var messagePayload = {
        action: 'trigger',
        stepId: stepId
    };
    bag.suAdapter.postHookWebhook(bag.hookIdsByStepId[stepId], messagePayload, function (err) {
        if (err) {
            logger.warn(who, 'failed to postHookWebhook for hook:', bag.hookIdsByStepId[stepId], 'with err', err);
            return next(true);
        }
        return next();
    });
}
function ___postExternalStepConsoles(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    var who = bag.who + '|' + ___postExternalStepConsoles.name;
    logger.debug(who, 'Inside');
    async.eachLimit(bag.group.groupStepIds, 10, function (stepObj, done) {
        var consoleAdapter = new StepConsoleAdapter(bag.jFrogPipelinesSessionId, global.config.apiToken, stepObj.stepId, bag.group.pipelineId, global.systemConfig.jobConsoleBatchSize, global.systemConfig.jobConsoleBufferTimeInterval);
        var logMessage = 'Successfully queued step. Waiting for execution to start.';
        consoleAdapter.openGrp('Queuing step');
        consoleAdapter.openCmd('Info');
        consoleAdapter.publishMsg(logMessage);
        consoleAdapter.closeCmd(true);
        consoleAdapter.closeGrp(true);
        return done();
    }, function (err) {
        return next(err);
    });
}
function __releaseExternalStepsLock(bag, next) {
    if (!bag.lock)
        return next();
    var who = bag.who + '|' + __releaseExternalStepsLock.name;
    logger.debug(who, 'Inside');
    bag.lock.release(function (err) {
        if (err)
            logger.warn('failed to release lock');
        return next();
    });
}
//# sourceMappingURL=queueExternalSteps.js.map