'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
var self = queueStepsToNodePools;
module.exports = self;
var _ = require('underscore');
var async = require('async');
var uuid = require('node-uuid');
var { CALL_OBJECT_LIMIT_FOR_BATCH, CALL_OBJECT_LIMIT, WILDCARD_AUDIENCE } = require('../constants.js');
var LAST_JOB_QUEUED_INTERVAL_IN_MINS = 2;
var LOCK_TIMEOUT_IN_MS = 1 * 60 * 1000;
var LOCK_TIMER_INTERVAL_IN_MS = 50 * 1000;
var { getElapsedTimeSinceStartTime } = require('./utils');
var BatchStepletConsoleAdapter = require('../BatchStepletConsoleAdapter.js');
var Lock = require('../Lock.js');
var nodeActions = require('../nodeActions.js');
var getDownstreamStepIds = require('./getDownStreamStepIds.js');
var sortStepsByDependency = require('./sortStepsByDependency.js');
var checkGroupExecutionEligibility = require('./checkGroupExecutionEligibility.js');
const { DataAdapterFactory } = require('../modules/dataAdapter/DataAdapterFactory.js');
const logger_1 = require("@jfrog-pipelines/logger");
function queueStepsToNodePools(externalBag, callback) {
    var who = externalBag.who + '|' + queueStepsToNodePools.name;
    logger.debug(who, 'Inside');
    let startTime = Date.now();
    var bag = {
        who: who,
        suAdapter: externalBag.suAdapter,
        statusCodes: global.systemCodesByGroup.status,
        nodePools: externalBag.nodePools,
        processingGroups: externalBag.processingGroups,
        waitingGroups: externalBag.waitingGroups,
        groupsProjects: externalBag.groupsProjects,
        allowedNodePoolsByProject: externalBag.allowedNodePoolsByProject,
        pipelinesById: externalBag.pipelinesById,
        incompleteStepsByRunId: externalBag.incompleteStepsByRunId,
        incompleteStepsById: externalBag.incompleteStepsById,
        runStepConnections: externalBag.runStepConnections,
        runStepConnectionsByRunId: externalBag.runStepConnectionsByRunId,
        runStepConnectionsByStepId: externalBag.runStepConnectionsByStepId,
        orphanedOnDemandNodesByProjectId: externalBag.orphanedOnDemandNodesByProjectId,
        runResourceVersionsById: externalBag.runResourceVersionsById,
        runResourceVersionsByRunId: externalBag.runResourceVersionsByRunId,
        jFrogPipelinesSessionId: externalBag.jFrogPipelinesSessionId,
        defaultPriority: externalBag.defaultPriority,
        runsById: externalBag.runsById,
        rejectedGroups: externalBag.rejectedGroups,
        stepTriggerLicenseCheckLockRemoved: externalBag.stepTriggerLicenseCheckLockRemoved,
        stepTriggerHueristics: externalBag.stepTriggerHueristics,
        result: {
            waitingGroups: [],
            processingGroups: []
        },
        successfulNodePools: 0,
        failedNodePools: 0,
        useRunAdjacency: externalBag.useRunAdjacency
    };
    async.series([
        _checkInputParams.bind(null, bag),
        _processNodePools.bind(null, bag),
        _setResult.bind(null, bag)
    ], function (err) {
        let nodePoolQueueStatus = "success";
        if (err)
            nodePoolQueueStatus = "failed";
        var stopTime = Date.now();
        var timeDiff = new Date(stopTime - startTime);
        var finalTime = util.format("durationMins=%s - %s minutes, %s seconds, and " +
            "%s milliseconds", (timeDiff.getTime() / 1000 / 60), timeDiff.getMinutes(), timeDiff.getSeconds(), timeDiff.getMilliseconds());
        logger.debug(bag.who, "Finished queueStepsToNodePools -", finalTime);
        const eventData = {
            timestamp: new Date(stopTime).toISOString(),
            domain: global.msName,
            eventType: "allnodepoolsprocess",
            status: nodePoolQueueStatus,
            durationMillis: timeDiff.valueOf(),
            successfulNodePools: bag.successfulNodePools,
            failedNodePools: bag.failedNodePools,
            pipelinesSessionId: bag.jFrogPipelinesSessionId
        };
        logger_1.LoggerFactory.logEvent(eventData);
        return callback(err, bag.result);
    });
}
function _checkInputParams(bag, next) {
    var who = bag.who + '|' + _checkInputParams.name;
    logger.debug(who, 'Inside');
    if (!bag.suAdapter)
        return next(new Error(':suAdapter is missing'));
    if (bag.useRunAdjacency) {
        bag.waitingGroups = _.reject(bag.waitingGroups, (waitingGroup) => {
            return waitingGroup.requiresApproval === true;
        });
    }
    return next();
}
function _processNodePools(bag, next) {
    if (_.isEmpty(bag.nodePools))
        return next();
    var who = bag.who + '|' + _processNodePools.name;
    logger.debug(who, 'Inside');
    var processingGroupsMap = {};
    _.each(bag.processingGroups, function (processingGroup) {
        let nodePool = _.find(bag.allowedNodePoolsByProject[processingGroup.projectId], function (nodePool) {
            return nodePool.name === processingGroup.nodePoolName;
        });
        if (nodePool) {
            if (!processingGroupsMap[nodePool.id])
                processingGroupsMap[nodePool.id] = [];
            processingGroupsMap[nodePool.id].push(processingGroup);
        }
    });
    var waitingGroupsMap = {};
    _.each(bag.waitingGroups, function (waitingGroup) {
        let nodePool = _.find(bag.allowedNodePoolsByProject[waitingGroup.projectId], function (nodePool) {
            return nodePool.name === waitingGroup.nodePoolName;
        });
        if (nodePool) {
            if (!waitingGroupsMap[nodePool.id])
                waitingGroupsMap[nodePool.id] = [];
            waitingGroupsMap[nodePool.id].push(waitingGroup);
        }
    });
    async.eachLimit(bag.nodePools, 5, function (nodePool, nextNodePool) {
        var innerBag = {
            who: bag.who + '|nodePoolName:' + nodePool.name,
            lock: {
                isValid: false,
                destroy: () => { },
                release: () => { }
            },
            suAdapter: bag.suAdapter,
            nodePool: nodePool,
            nodes: [],
            project: _.findWhere(bag.groupsProjects, { id: nodePool.projectId }),
            waitingGroups: waitingGroupsMap[nodePool.id] || [],
            statusCodes: bag.statusCodes,
            pipelinesById: bag.pipelinesById,
            incompleteStepsByRunId: bag.incompleteStepsByRunId,
            incompleteStepsById: bag.incompleteStepsById,
            processingGroups: processingGroupsMap[nodePool.id] || [],
            runStepConnections: bag.runStepConnections,
            runStepConnectionsByRunId: bag.runStepConnectionsByRunId,
            runStepConnectionsByStepId: bag.runStepConnectionsByStepId,
            runResourceVersionsById: bag.runResourceVersionsById,
            runResourceVersionsByRunId: bag.runResourceVersionsByRunId,
            jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId,
            runsById: bag.runsById,
            error: false,
            defaultPriority: bag.defaultPriority,
            rejectedGroups: bag.rejectedGroups,
            stepTriggerLicenseCheckLockRemoved: bag.stepTriggerLicenseCheckLockRemoved,
            stepTriggerHueristics: bag.stepTriggerHueristics,
            stepsQueued: 0,
            useRunAdjacency: bag.useRunAdjacency
        };
        if (!nodePool.isOnDemand && _.isEmpty(innerBag.waitingGroups))
            return process.nextTick(nextNodePool);
        if (nodePool.isOrphaned)
            innerBag.nodes =
                bag.orphanedOnDemandNodesByProjectId[nodePool.projectId];
        let startTime = new Date();
        let tasks = [
            __obtainLock.bind(null, innerBag),
            __getNodePool.bind(null, innerBag),
            __getNodes.bind(null, innerBag),
            __filterAvailableNodes.bind(null, innerBag),
            __getWaitingSteps.bind(null, innerBag),
            __setReadyAt.bind(null, innerBag),
            __rejectGroupByNodes.bind(null, innerBag),
            __rejectGroupByNonRootEnforcement.bind(null, innerBag),
            __checkEnvironments.bind(null, innerBag),
            __checkNodePoolPermissions.bind(null, innerBag),
            __findAvailableNodeCount.bind(null, innerBag),
            __tryAndQueueSteps.bind(null, innerBag)
        ];
        if (bag.useRunAdjacency) {
            tasks = [
                __obtainLock.bind(null, innerBag),
                __getNodePool.bind(null, innerBag),
                __getNodes.bind(null, innerBag),
                __filterAvailableNodes.bind(null, innerBag),
                __getBuildPlaneMessagesCount.bind(null, innerBag),
                __getWaitingStepsByRunAdjacencies.bind(null, innerBag),
                __setReadyAtByRunAdjacencies.bind(null, innerBag),
                __rejectGroupByNodes.bind(null, innerBag),
                __rejectGroupByNonRootEnforcement.bind(null, innerBag),
                __checkEnvironments.bind(null, innerBag),
                __checkNodePoolPermissions.bind(null, innerBag),
                __findAvailableNodeCount.bind(null, innerBag),
                __tryAndQueueSteps.bind(null, innerBag)
            ];
        }
        tasks = tasks.concat([
            __releaseLock.bind(null, innerBag),
            __publishToNPM.bind(null, innerBag)
        ]);
        async.series(tasks, function (err) {
            let nodePoolQueueStatus = "success";
            var stopTime = Date.now();
            var timeDiff = new Date(stopTime - startTime);
            var finalTime = util.format("durationMins=%s - %s minutes, %s seconds, and " +
                "%s milliseconds", (timeDiff.getTime() / 1000 / 60), timeDiff.getMinutes(), timeDiff.getSeconds(), timeDiff.getMilliseconds());
            logger.debug(bag.who, `Finished processing nodePool - ${nodePool.name} and it took ${finalTime}`);
            if (err || bag.error) {
                logger.debug(bag.who, 'Failed to process nodePool:', nodePool.name, 'with error:', err || bag.error);
                nodePoolQueueStatus = 'failed';
                bag.failedNodePools++;
            }
            else {
                logger.debug(bag.who, 'Successfully processed nodePool:', nodePool.name);
                bag.successfulNodePools++;
            }
            const eventData = {
                timestamp: new Date(stopTime).toISOString(),
                domain: global.msName,
                eventType: "nodepoolprocess",
                status: nodePoolQueueStatus,
                queuedStepsCount: innerBag.stepsQueued,
                durationMillis: timeDiff.valueOf(),
                nodepoolName: nodePool.name,
                pipelinesSessionId: bag.jFrogPipelinesSessionId
            };
            logger_1.LoggerFactory.logEvent(eventData);
            if (bag.lock) {
                bag.lock.destroy();
                bag.lock = null;
            }
            return nextNodePool();
        });
    }, function () {
        return next();
    });
}
function __obtainLock(bag, next) {
    var who = bag.who + '|' + __obtainLock.name;
    logger.debug(who, 'Inside');
    var params = {
        name: global.msName + '-nodePools-' + bag.nodePool.id,
        lockTimeoutInMs: LOCK_TIMEOUT_IN_MS,
        lockTimerInMs: LOCK_TIMER_INTERVAL_IN_MS,
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    bag.lock = new Lock(params);
    bag.lock.obtain(function (err) {
        if (err)
            logger.debug('failed to obtain lock with name:', params.name);
        return next(err);
    });
}
function __getNodePool(bag, next) {
    if (!bag.lock.isValid)
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __getNodePool.name;
    logger.debug(who, 'Inside');
    var query = 'nodePoolIds=' + bag.nodePool.id;
    bag.suAdapter.getNodePools(query, function (err, nodePools) {
        if (err) {
            logger.debug(who, 'Failed to getNodePools for query:', query, 'with error:', err);
            bag.error = true;
            return next();
        }
        if (_.isEmpty(nodePools)) {
            logger.debug(who, 'nodePools empty for query', query);
            bag.error = true;
            return next();
        }
        bag.nodePool = nodePools[0];
        return next();
    });
}
function __getNodes(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __getNodes.name;
    logger.debug(who, 'Inside');
    var query = 'nodePoolIds=' + bag.nodePool.id;
    bag.suAdapter.getNodes(query, function (err, nodes) {
        if (err) {
            logger.debug(who, util.format('Failed to getNodes for query: %s with error: %s', query, err));
            bag.error = true;
            return next();
        }
        bag.nodes = nodes;
        return next();
    });
}
function __filterAvailableNodes(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.nodes))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __filterAvailableNodes.name;
    logger.debug(who, 'Inside');
    bag.nodes = _.reject(bag.nodes, function (node) {
        if (node.statusCode !== bag.statusCodes.waiting.code &&
            node.statusCode !== bag.statusCodes.queued.code &&
            node.statusCode !== bag.statusCodes.processing.code &&
            node.statusCode !== bag.statusCodes.success.code &&
            node.statusCode !== bag.statusCodes.cached.code)
            return node;
    });
    return next();
}
function __getWaitingStepsByRunAdjacencies(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __getWaitingStepsByRunAdjacencies.name;
    logger.debug(who, 'Inside');
    const dataAdapter = DataAdapterFactory.getApiDataAdapter();
    async.eachLimit(bag.waitingGroups, 5, async (waitingGroup) => {
        try {
            let query = `affinityGroups=${waitingGroup.affinityGroup}&runIds=${waitingGroup.runId}`;
            query += '&include=id';
            const runAdjacencies = await dataAdapter.getRunAdjacencyByQuery(query);
            if (runAdjacencies.length === 0 || runAdjacencies.length !== waitingGroup.runAdjacencyList.length) {
                waitingGroup.rejectByRunAdjacency = true;
            }
        }
        catch (err) {
            logger.debug(who, err);
            throw err;
        }
    }, (err) => {
        if (err) {
            return next(err);
        }
        bag.waitingGroups = _.reject(bag.waitingGroups, (waitingGroup) => {
            return waitingGroup.rejectByRunAdjacency;
        });
        return next();
    });
}
function __getWaitingSteps(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __getWaitingSteps.name;
    logger.debug(who, 'Inside');
    var waitingGroupsByRunId = _.groupBy(bag.waitingGroups, 'runId');
    var stepIds = _.uniq(_.pluck(_.flatten(_.pluck(bag.waitingGroups, 'groupStepIds')), 'stepId'));
    if (_.isEmpty(stepIds))
        return next();
    var incompleteStatusCodes = [
        bag.statusCodes.waiting.code,
        bag.statusCodes.processing.code,
        bag.statusCodes.queued.code
    ];
    var stepIdGroups = [];
    var i = 0;
    while (i < stepIds.length) {
        stepIdGroups.push(stepIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    var currentStepRunIds = [];
    async.eachLimit(stepIdGroups, 10, function (stepIdGroup, done) {
        var query = util.format('stepIds=%s&statusCodes=%s&limit=%s', stepIdGroup.join(','), incompleteStatusCodes.join(','), CALL_OBJECT_LIMIT_FOR_BATCH);
        bag.suAdapter.getSteps(query, function (err, steps) {
            if (err) {
                bag.error = true;
                logger.debug(who, 'getSteps for query:', query, 'returned error:', err);
                return done();
            }
            var stepsInProgress = _.where(steps, { groupInProgress: true });
            _.each(stepsInProgress, function (step) {
                waitingGroupsByRunId[step.runId] = _.reject(waitingGroupsByRunId[step.runId], function (group) {
                    if (group.affinityGroup === step.affinityGroup) {
                        bag.processingGroups.push(group);
                        return true;
                    }
                });
            });
            currentStepRunIds = _.uniq(currentStepRunIds.concat(_.pluck(_.where(steps, { groupInProgress: false }), 'runId')));
            return done();
        });
    }, function (err) {
        if (err)
            return next(err);
        var existingRunIds = _.keys(waitingGroupsByRunId);
        if (currentStepRunIds.length === existingRunIds.length)
            return next();
        var completedRunIds = _.difference(existingRunIds, currentStepRunIds);
        _.each(completedRunIds, function (runId) {
            delete waitingGroupsByRunId[runId];
        });
        bag.waitingGroups = _.flatten(_.values(waitingGroupsByRunId), true);
        return next();
    });
}
function __setReadyAtByRunAdjacencies(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __setReadyAtByRunAdjacencies.name;
    logger.debug(who, 'Inside');
    const dataAdapter = DataAdapterFactory.getApiDataAdapter();
    async.eachLimit(bag.waitingGroups, 5, async (waitingGroup) => {
        var alreadySet = _.some(waitingGroup.groupStepIds, function (step) {
            return step.readyAt;
        });
        if (alreadySet)
            return;
        const firstStepsList = waitingGroup.firstStepsList;
        if (firstStepsList.length === 0) {
            return;
        }
        try {
            const updatedRecords = await dataAdapter.updateStepsReadyAt(firstStepsList);
        }
        catch (err) {
            logger.debug(who, err);
            throw err;
        }
    }, (err) => {
        return next(err);
    });
}
function __setReadyAt(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __setReadyAt.name;
    logger.debug(who, 'Inside');
    async.eachLimit(bag.waitingGroups, 5, function (waitingGroup, done) {
        var alreadySet = _.some(waitingGroup.groupStepIds, function (step) {
            return step.readyAt;
        });
        if (alreadySet)
            return process.nextTick(done);
        var groupStepIds = _.clone(waitingGroup.groupStepIds);
        var runStepConnections = bag.runStepConnectionsByRunId[waitingGroup.runId];
        var runStepConnectionsByStepId = _.groupBy(runStepConnections, 'stepId');
        var runStepConnectionsByOperationStepId = _.groupBy(runStepConnections, 'operationStepId');
        var runStepConnectionsByResourceName = _.groupBy(runStepConnections, 'operationRunResourceVersionName');
        async.eachSeries(waitingGroup.groupStepIds, function (stepObj, nextStepObj) {
            var _bag = {
                who: who,
                stepId: stepObj.stepId,
                runResourceVersionsById: bag.runResourceVersionsById,
                runStepConnectionsByStepId: runStepConnectionsByStepId,
                runStepConnectionsByOperationStepId: runStepConnectionsByOperationStepId,
                runStepConnectionsByResourceName: runStepConnectionsByResourceName,
            };
            getDownstreamStepIds(_bag, function (downstreamStepIds) {
                groupStepIds = _.reject(groupStepIds, function (groupStep) {
                    return _.contains(downstreamStepIds, groupStep.stepId);
                });
                return process.nextTick(nextStepObj);
            });
        }, function () {
            groupStepIds = _.sortBy(groupStepIds, function (groupStep) {
                return groupStep.priority;
            });
            var firstStep = _.first(groupStepIds);
            if (!firstStep)
                return process.nextTick(done);
            var update = {
                readyAt: new Date()
            };
            bag.suAdapter.putStepById(firstStep.stepId, update, function (err) {
                if (err) {
                    logger.debug(who, 'Failed to putStepById for id:', firstStep.stepId, 'with error:', err);
                    return done(true);
                }
                return process.nextTick(done);
            });
        });
    }, function (err) {
        if (err)
            bag.error = true;
        return next();
    });
}
function __rejectGroupByNodes(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __rejectGroupByNodes.name;
    logger.debug(who, 'Inside');
    var skipByNoNode = false;
    if (bag.nodePool.isOnDemand) {
        if (bag.nodePool.numberOfNodes)
            return next();
        else
            skipByNoNode = true;
    }
    else {
        if (_.isEmpty(bag.nodes))
            skipByNoNode = true;
    }
    if (skipByNoNode)
        _.each(bag.waitingGroups, function (group) {
            group.skipByNoNode = true;
        });
    return next();
}
function __rejectGroupByNonRootEnforcement(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __rejectGroupByNonRootEnforcement.name;
    logger.debug(who, 'Inside');
    let skipByNonRootEnforcement = false;
    if (global.config.enforceNonRootNodes) {
        if (!bag.nodePool.initAsNonRoot) {
            skipByNonRootEnforcement = true;
        }
    }
    if (skipByNonRootEnforcement)
        _.each(bag.waitingGroups, function (group) {
            group.skipByNonRootEnforcement = true;
        });
    return next();
}
function __checkEnvironments(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    if (!bag.project || bag.project.name === 'default')
        return next();
    var who = bag.who + '|' + __checkEnvironments.name;
    logger.debug(who, 'Inside');
    if (_.isEmpty(bag.nodePool.environments)) {
        _.each(bag.waitingGroups, function (group) {
            group.skipByNodePoolPermissions = true;
        });
        return next();
    }
    var pipelineSourceIds = _.compact(_.pluck(bag.waitingGroups, 'pipelineSourceId'));
    if (_.isEmpty(pipelineSourceIds))
        return next();
    var query = util.format('pipelineSourceIds=%s', pipelineSourceIds.join(','));
    bag.suAdapter.getPipelineSources(query, function (err, pipelineSources) {
        if (err) {
            bag.error = true;
            logger.debug(who, 'getPipelineSources for query:', query, 'returned error:', err);
            return next();
        }
        var pipelineSourcesById = _.indexBy(pipelineSources, 'id');
        var missingPipelineSources = [];
        _.each(bag.waitingGroups, function (group) {
            var pipelineSource = pipelineSourcesById[group.pipelineSourceId];
            if (!pipelineSource) {
                missingPipelineSources.push(group.pipelineSourceId);
                return;
            }
            if (_.isEmpty(_.intersection(bag.nodePool.environments, pipelineSource.environments)))
                group.skipByNodePoolPermissions = true;
        });
        if (!_.isEmpty(missingPipelineSources)) {
            bag.error = true;
            logger.debug(who, 'pipelineSources not found for ids:', missingPipelineSources);
        }
        return next();
    });
}
function __checkNodePoolPermissions(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    if (!bag.project || bag.project.name !== 'default')
        return next();
    var who = bag.who + '|' + __checkNodePoolPermissions.name;
    logger.debug(who, 'Inside');
    var pipelineSourceIds = _.compact(_.pluck(bag.waitingGroups, 'pipelineSourceId'));
    if (_.isEmpty(pipelineSourceIds))
        return next();
    var query = util.format('objectTypes=nodePool&objectIds=%s&pipelineSourceIds=%s', bag.nodePool.id, pipelineSourceIds.join(','));
    bag.suAdapter.getPipelineSourceObjectPermissions(query, function (err, pipelineSourceObjectPermissions) {
        if (err) {
            bag.error = true;
            logger.debug(who, 'getPipelineSourceObjectPermissions for query:', query, 'returned error:', err);
            return next();
        }
        var permissions = _.groupBy(pipelineSourceObjectPermissions, 'pipelineSourceId');
        _.each(bag.waitingGroups, function (group) {
            if (group.projectId === bag.project.id &&
                _.isEmpty(permissions[group.pipelineSourceId]))
                group.skipByNodePoolPermissions = true;
        });
        return next();
    });
}
function __getBuildPlaneMessagesCount(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __getBuildPlaneMessagesCount.name;
    logger.debug(who, 'Inside');
    const statusCodes = [bag.statusCodes.ready.code, bag.statusCodes.processing.code, bag.statusCodes.queued.code];
    const query = `nodePoolIds=${bag.nodePool.id}&statusCodes=${statusCodes.join(',')}`;
    bag.suAdapter.getBuildPlaneMessagesCount(query, function (err, count) {
        if (err) {
            return next(err);
        }
        bag.buildPlaneMessagesCount = count;
        return next();
    });
}
function __findAvailableNodeCount(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups) &&
        _.isEmpty(bag.processingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __findAvailableNodeCount.name;
    logger.debug(who, 'Inside');
    bag.nodePool.totalGroupsThatCanBeQueued = 0;
    if (bag.nodePool.isOnDemand) {
        var processingGroupsCount = _.countBy(bag.processingGroups, function (group) {
            return group.isExternal ? 'external' : 'notExternal';
        }).notExternal || 0;
        if (bag.useRunAdjacency) {
            processingGroupsCount = bag.buildPlaneMessagesCount;
        }
        bag.nodePool.totalGroupsThatCanBeQueued = bag.nodePool.numberOfNodes -
            processingGroupsCount;
        var activeStatusCodes = [bag.statusCodes.waiting.code,
            bag.statusCodes.queued.code,
            bag.statusCodes.processing.code, bag.statusCodes.success.code];
        var activeNodes = _.filter(bag.nodes, function (node) {
            return (_.contains(activeStatusCodes, node.statusCode));
        });
        activeNodes = _.reject(activeNodes, function (activeNode) {
            return activeNode.stopRequested === true;
        });
        var cachedNodes = _.filter(bag.nodes, function (node) {
            return bag.nodePool.isCacheEnabled &&
                node.statusCode === bag.statusCodes.cached.code &&
                node.stopRequested !== true;
        });
        bag.nodePool.cachedNodes = _.sortBy(cachedNodes, function (cachedNode) {
            return new Date(cachedNode.statusLastUpdatedAt);
        });
        bag.nodePool.activeNodesCount = activeNodes.length;
        bag.nodePool.cachedNodesCount = cachedNodes.length;
    }
    else {
        var processingNodesCount = _.filter(bag.nodes, function (node) {
            return node.statusCode === bag.statusCodes.processing.code &&
                node.stopRequested !== true;
        });
        var totalNodeCount = bag.nodePool.numberOfNodes === null ?
            bag.nodes.length : Math.min(bag.nodes.length, bag.nodePool.numberOfNodes);
        bag.nodePool.totalGroupsThatCanBeQueued = totalNodeCount -
            processingNodesCount.length;
    }
    return next();
}
function __tryAndQueueSteps(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    if (bag.nodePool.isOrphaned)
        return next();
    var who = bag.who + '|' + __tryAndQueueSteps.name;
    logger.debug(who, 'Inside');
    async.eachSeries(bag.waitingGroups, function (group, done) {
        if (!bag.lock.isValid) {
            logger.error(who, 'invalid lock, exiting');
            return next(true);
        }
        if (!group || _.isEmpty(group.groupStepIds))
            return process.nextTick(done);
        if (bag.stepTriggerHueristics.enabled && !bag.stepTriggerHueristics.hasRemainingCapacity) {
            logger.debug(who, `This stepTrigger utilised its allowed capacity`);
            return next();
        }
        let skipAquireLock = (bag.stepTriggerLicenseCheckLockRemoved === true ||
            (bag.stepTriggerHueristics.enabled === true && bag.stepTriggerHueristics.allowedCapacity >= 2));
        let startTime = new Date();
        var seriesBag = {
            who: bag.who + '|stepId:' + group.groupStepIds[0].stepId,
            statusCodes: bag.statusCodes,
            suAdapter: bag.suAdapter,
            waitingGroups: bag.waitingGroups,
            group: group,
            project: bag.project,
            nodePool: bag.nodePool,
            stepsByRunId: bag.incompleteStepsByRunId,
            incompleteStepsById: bag.incompleteStepsById,
            runStepConnections: bag.runStepConnections,
            runStepConnectionsByRunId: bag.runStepConnectionsByRunId,
            runResourceVersionsById: bag.runResourceVersionsById,
            runResourceVersionsByRunId: bag.runResourceVersionsByRunId,
            pipelinesById: bag.pipelinesById,
            runStepConnectionsByStepId: bag.runStepConnectionsByStepId,
            jf_pipelines_inputs: {},
            stepCheckLock: {
                isValid: false,
                release: () => { },
                destroy: () => { }
            },
            jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId,
            error: false,
            defaultPriority: bag.defaultPriority,
            runsById: bag.runsById,
            rejectedGroups: bag.rejectedGroups,
            stepTriggerLicenseCheckLockRemoved: bag.stepTriggerLicenseCheckLockRemoved,
            stepTriggerHueristics: bag.stepTriggerHueristics,
            skipAquireLock: skipAquireLock,
            useRunAdjacency: bag.useRunAdjacency
        };
        const tasks = [
            ___obtainLock.bind(null, seriesBag),
            ___checkNodeAvailability.bind(null, seriesBag),
        ];
        if (!bag.useRunAdjacency) {
            tasks.push(___checkGroupExecutionEligibility.bind(null, seriesBag));
        }
        tasks.push(___createRunSetupToken.bind(null, seriesBag), ___checkRunExecutionTokenAlreadyExists.bind(null, seriesBag), ___createRunExecutionToken.bind(null, seriesBag), ___addRunExecutionTokenToRun.bind(null, seriesBag), ___updateStep.bind(null, seriesBag), ___getSteplets.bind(null, seriesBag), ___updateSteplet.bind(null, seriesBag), ___postStepSkippedConsoles.bind(null, seriesBag), ___queueStep.bind(null, seriesBag), ___releaseLock.bind(null, seriesBag), ___postStepInfoConsoles.bind(null, seriesBag));
        async.series(tasks, function (err) {
            let finalTime = getElapsedTimeSinceStartTime(startTime);
            let waitingGroupQueueStatus = "failed";
            var stopTime = Date.now();
            var timeDiff = new Date(stopTime - startTime);
            logger.debug(seriesBag.who, `Finished tryAndQueueSteps - ${finalTime}`);
            if (!err && !seriesBag.group.reject) {
                logger.trace(who, 'queued group:', group.affinityGroup, 'with runId:', group.runId);
                bag.stepsQueued = bag.stepsQueued + _.size(seriesBag.group.groupStepIds);
                if (_.isEmpty(bag.processingGroups))
                    bag.processingGroups = [group];
                else
                    bag.processingGroups.push(group);
            }
            if (bag.stepTriggerHueristics.enabled === true) {
                bag.stepTriggerHueristics.allowedCapacity -= 1;
                bag.stepTriggerHueristics.hasRemainingCapacity =
                    bag.stepTriggerHueristics.allowedCapacity > 0;
            }
            const eventData = {
                timestamp: new Date(stopTime).toISOString(),
                domain: global.msName,
                eventType: "queueWaitingGroup",
                status: waitingGroupQueueStatus,
                queuedStepsCount: bag.stepsQueued,
                durationMillis: timeDiff.valueOf(),
                affinityGroup: bag.waitingGroups[0]?.affinityGroup,
                nodePoolName: bag.waitingGroups[0]?.nodePoolName,
                pipelinesSessionId: bag.jFrogPipelinesSessionId
            };
            logger_1.LoggerFactory.logEvent(eventData);
            if (bag.stepCheckLock) {
                bag.stepCheckLock.destroy();
                bag.stepCheckLock = null;
            }
            return process.nextTick(done);
        });
    }, function () {
        return next();
    });
}
function ___obtainLock(bag, next) {
    if (bag.skipAquireLock) {
        bag.stepCheckLock = {
            ...bag.stepCheckLock,
            isValid: true,
        };
        return next();
    }
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    var who = bag.who + '|' + ___obtainLock.name;
    logger.debug(who, 'Inside');
    var params = {
        name: global.msName + '-check-capacity',
        lockTimeoutInMs: 10000,
        lockTimerInMs: 8000,
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    bag.stepCheckLock = new Lock(params);
    var maxRetries = 30;
    var retries = 0;
    async.whilst(function () {
        return !bag.stepCheckLock.isValid;
    }, function (done) {
        bag.stepCheckLock.obtain(function (err) {
            if (err) {
                logger.trace('failed to obtain lock with name:', params.name);
                retries += 1;
                if (retries > maxRetries)
                    return done(err);
                else
                    return setTimeout(done, 1000);
            }
            else {
                return done();
            }
        });
    }, function (err) {
        return next(err);
    });
}
function ___checkNodeAvailability(bag, next) {
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    var who = bag.who + '|' + ___checkNodeAvailability.name;
    logger.debug(who, 'Inside');
    if (bag.nodePool.totalGroupsThatCanBeQueued <= 0) {
        bag.group.reject = true;
        if (!bag.group.waitingForFreeNodePoolCapacity) {
            bag.group.rejectionContext.waitingForFreeNodePoolCapacity = {
                nodePoolName: bag.nodePool.name,
                numberOfNodes: bag.nodePool.numberOfNodes,
                timestamp: new Date()
            };
            bag.rejectedGroups.waitingForFreeNodePoolCapacity.push(bag.group);
        }
        return next();
    }
    bag.suAdapter.getExecutionNodes(function (err, results) {
        var rejectAll = false;
        let insufficientLicenseCapacity = false;
        if (err) {
            logger.debug('failed to check remaining capacity:', err.message);
            rejectAll = true;
        }
        else if (results.available <= 0) {
            rejectAll = true;
            insufficientLicenseCapacity = true;
        }
        if (rejectAll) {
            _.each(bag.waitingGroups, function (group) {
                group.reject = true;
                if (insufficientLicenseCapacity && !group.waitingForFreeLicenseCapacity) {
                    group.rejectionContext.waitingForFreeLicenseCapacity = {
                        maxCapacity: results.allowed,
                        timestamp: new Date()
                    };
                    bag.rejectedGroups.waitingForFreeLicenseCapacity.push(group);
                }
            });
            bag.group.reject = true;
        }
        return next();
    });
}
function ___checkGroupExecutionEligibility(bag, next) {
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    if (bag.group.reject)
        return next();
    var who = bag.who + '|' + ___checkGroupExecutionEligibility.name;
    logger.debug(who, 'Inside');
    var _bag = {
        who: who,
        suAdapter: bag.suAdapter,
        group: bag.group,
        runStepConnections: bag.runStepConnections || [],
        runResourceVersions: bag.runResourceVersionsByRunId[bag.group.runId] || [],
    };
    checkGroupExecutionEligibility(_bag, function (err, canQueue) {
        if (err)
            logger.debug(who, err);
        if (!canQueue)
            bag.group.reject = true;
        return next();
    });
}
function ___createRunSetupToken(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    if (_.isEmpty(global.accessClient) || _.isEmpty(global.accessClientContext))
        return next();
    var who = bag.who + '|' + ___createRunSetupToken.name;
    logger.debug(who, 'Inside');
    var oneDayInSeconds = 24 * 60 * 60;
    var tokenRequest = {
        username: `run-setup-${bag.group.runId}@jfpip`,
        scope: 'internal:pipelines/project/' +
            bag.group.projectId + ':r,w,d,m',
        refreshable: false,
        audience: WILDCARD_AUDIENCE,
        expiresIn: oneDayInSeconds,
        grantType: 'client_credentials'
    };
    global.accessClient.createToken(global.accessClientContext, tokenRequest).then(function (tokenResponse) {
        bag.runSetupToken = tokenResponse.accessToken;
        return next();
    }).catch(function (err) {
        var logMessage = 'Failed to create Access Token with error: ' +
            err;
        logger.debug(who, logMessage);
        bag.error = true;
        bag.group.reject = true;
        return next();
    });
}
function ___checkRunExecutionTokenAlreadyExists(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    var who = bag.who + '|' + ___checkRunExecutionTokenAlreadyExists.name;
    logger.debug(who, 'Inside');
    bag.suAdapter.getRunById(bag.group.runId, function (err, run) {
        if (err) {
            var logMessage = 'Failed to get run with error: ' +
                err;
            logger.debug(who, logMessage);
            bag.error = true;
            bag.group.reject = true;
            return next();
        }
        bag.run = run;
        if (!_.isEmpty(run.systemPropertyBag?.token)) {
            bag.runToken = run.systemPropertyBag.token;
            bag.skipRunExecutionToken = true;
        }
        return next();
    });
}
function ___createRunExecutionToken(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    if (_.isEmpty(global.accessClient) || _.isEmpty(global.accessClientContext))
        return next();
    if (bag.skipRunExecutionToken)
        return next();
    var who = bag.who + '|' + ___createRunExecutionToken.name;
    logger.debug(who, 'Inside');
    var oneDayInSeconds = 24 * 60 * 60;
    var tokenRequest = {
        username: `run-${bag.group.runId}@jfpip`,
        scope: 'internal:pipelines/project/' +
            bag.group.projectId + ':r,w,d',
        refreshable: false,
        audience: WILDCARD_AUDIENCE,
        expiresIn: oneDayInSeconds,
        grantType: 'client_credentials'
    };
    global.accessClient.createToken(global.accessClientContext, tokenRequest).then(function (tokenResponse) {
        bag.runToken = tokenResponse.accessToken;
        return next();
    }).catch(function (err) {
        var logMessage = 'Failed to create Access Token with error: ' +
            err;
        logger.debug(who, logMessage);
        bag.error = true;
        bag.group.reject = true;
        return next();
    });
}
function ___addRunExecutionTokenToRun(bag, next) {
    if (bag.error)
        return next();
    if (!bag.runToken)
        return next();
    if (bag.skipRunExecutionToken)
        return next();
    var who = bag.who + '|' + ___addRunExecutionTokenToRun.name;
    logger.debug(who, 'Inside');
    if (!bag.run) {
        var logMessage = `Failed to find run ${bag.group.runId}.`;
        logger.debug(who, logMessage);
        bag.error = true;
        bag.group.reject = true;
        return next();
    }
    var systemPropertyBag = bag.run.systemPropertyBag || {};
    systemPropertyBag.token = bag.runToken;
    var update = {
        systemPropertyBag: systemPropertyBag
    };
    bag.suAdapter.putRunById(bag.group.runId, update, function (err, run) {
        if (err) {
            var logMessage = 'Failed to update run with error: ' +
                err;
            logger.debug(who, logMessage);
            bag.error = true;
            bag.group.reject = true;
            return next();
        }
        bag.run = run;
        return next();
    });
}
function ___updateStep(bag, next) {
    if (!bag.group.skipByNoNode &&
        !bag.group.skipByNodePoolPermissions &&
        !bag.group.skipByNonRootEnforcement) {
        return next();
    }
    if (bag.group.reject)
        return next();
    var who = bag.who + '|' + ___updateStep.name;
    logger.debug(who, 'Inside');
    var update = {};
    update.statusCode = bag.statusCodes.skipped.code;
    update.pendingLogsComplete = true;
    update.endedAt = new Date();
    async.eachLimit(bag.group.groupStepIds, 10, function (stepObj, done) {
        bag.suAdapter.putStepById(stepObj.stepId, update, function (err) {
            if (err) {
                logger.debug(who, 'Failed to putStepById for id:', stepObj.stepId, 'with error:', err);
                return done(true);
            }
            return done();
        });
    }, function (err) {
        if (err)
            bag.error = true;
        return next();
    });
}
function ___releaseLock(bag, next) {
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    if (bag.skipAquireLock)
        return next();
    var who = bag.who + '|' + ___releaseLock.name;
    logger.debug(who, 'Inside');
    bag.stepCheckLock.release(function (err) {
        if (err)
            logger.debug(who, 'Failed to release lock with name: ' +
                bag.stepCheckLock.name);
        return next();
    });
}
function ___getSteplets(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    var who = bag.who + '|' + ___getSteplets.name;
    logger.debug(who, 'Inside');
    var groupStepIdList = _.uniq(_.pluck(bag.group.groupStepIds, 'stepId'));
    var query = util.format('projectIds=%s&stepIds=%s', bag.group.projectId, groupStepIdList.join(','));
    bag.suAdapter.getSteplets(query, function (err, steplets) {
        if (err) {
            logger.debug(who, 'Failed to getSteplets for query:', query, 'with error:', err);
            return next(true);
        }
        bag.steplets = steplets;
        bag.stepletsById = _.indexBy(bag.steplets, 'id');
        bag.stepletsByStepId = _.groupBy(bag.steplets, 'stepId');
        return next();
    });
}
function ___updateSteplet(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    if (_.isEmpty(bag.steplets))
        return next();
    if (!bag.group.skipByNoNode && !bag.group.skipByNodePoolPermissions &&
        !bag.group.skipByNonRootEnforcement)
        return next();
    var who = bag.who + '|' + ___updateSteplet.name;
    logger.debug(who, 'Inside');
    var update = {
        statusCode: bag.statusCodes.skipped.code,
        endedAt: new Date()
    };
    async.eachLimit(bag.steplets, 10, function (steplet, done) {
        bag.suAdapter.putStepletById(steplet.id, update, function (err, steplet) {
            if (err) {
                logger.debug(who, 'Failed to putStepletById for id:', steplet.id, 'with error:', err);
                return done(true);
            }
            return done();
        });
    }, function (err) {
        if (err)
            bag.error = true;
        return next();
    });
}
function ___postStepSkippedConsoles(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    if (!bag.group.skipByNoNode && !bag.group.skipByNodePoolPermissions
        && !bag.group.skipByNonRootEnforcement)
        return next();
    var who = bag.who + '|' + ___postStepSkippedConsoles.name;
    logger.debug(who, 'Inside');
    var consoleAdapter = new BatchStepletConsoleAdapter(bag.jFrogPipelinesSessionId, global.config.apiToken, _.pluck(bag.steplets, 'id'), bag.group.pipelineId, global.systemConfig.jobConsoleBatchSize, global.systemConfig.jobConsoleBufferTimeInterval);
    var grpMsg;
    var logMessage;
    var closeCmd = false;
    var closeGrp = false;
    if (bag.group.skipByNoNode) {
        grpMsg = 'Step skipped';
        logMessage = 'Skipping step as no node is found. ' +
            'Add a node to run a step.';
    }
    else if (bag.group.skipByNodePoolPermissions) {
        grpMsg = 'Step skipped';
        logMessage = 'Skipping step as it is part of a pipeline source that ' +
            'does not have permission to use the specified nodePool.';
    }
    else if (bag.group.skipByNonRootEnforcement) {
        grpMsg = 'Step skipped';
        logMessage = 'Cannot initialize node pool with a root user' +
            ' as a global restriction for' +
            ' using only non-root node pools has been set (enforceNonRootNodes: true).\n' +
            'Either contact your administrator or choose a non-root' +
            ' node pool with operating' +
            ' system Ubuntu 18 or Ubuntu 20.';
    }
    else {
        return next();
    }
    consoleAdapter.openGrp(grpMsg);
    consoleAdapter.openCmd('Info');
    consoleAdapter.publishMsg(logMessage);
    consoleAdapter.closeCmd(closeCmd);
    consoleAdapter.closeGrp(closeGrp);
    return next();
}
function ___queueStep(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    if (bag.group.skipByNoNode ||
        bag.group.skipByNodePoolPermissions ||
        bag.group.skipByNonRootEnforcement)
        return next();
    var who = bag.who + '|' + ___queueStep.name;
    logger.debug(who, 'Inside');
    var innerBag = {
        who: bag.who,
        defaultPriority: bag.defaultPriority,
        suAdapter: bag.suAdapter,
        runSetupToken: bag.runSetupToken,
        runToken: bag.runToken,
        nodePool: bag.nodePool,
        stepsByRunId: bag.stepsByRunId,
        runStepConnections: bag.runStepConnections,
        runStepConnectionsByRunId: bag.runStepConnectionsByRunId,
        runResourceVersionsById: bag.runResourceVersionsById,
        group: bag.group,
        stepletsByStepId: bag.stepletsByStepId,
        stepletsById: bag.stepletsById,
        stepIdsToExecute: [],
        stepsToExecute: [],
        statusCodes: bag.statusCodes,
        useRunAdjacency: bag.useRunAdjacency
    };
    let tasks = [];
    if (innerBag.useRunAdjacency) {
        tasks = [
            ___getAffinityGroupStepsToExecuteUsingRunAdjacency.bind(null, innerBag),
            ___queueToNodePool.bind(null, innerBag)
        ];
    }
    else {
        tasks = [
            ___getAffinityGroupStepsToExecute.bind(null, innerBag),
            ___getInSteps.bind(null, innerBag),
            ___prioritizeStepsToExecute.bind(null, innerBag),
            ___queueToNodePool.bind(null, innerBag)
        ];
    }
    async.series(tasks, function (err) {
        if (!err)
            bag.nodePool.totalGroupsThatCanBeQueued--;
        return next(err);
    });
}
function ___postBuildPlaneMessage(bag, next) {
    if (_.isEmpty(bag.stepIdsToExecute))
        return next();
    var who = bag.who + '|' + ___postBuildPlaneMessage.name;
    logger.debug(who, 'Inside');
    var messagePayload = {
        affinityGroup: bag.group.affinityGroup,
        runId: bag.group.runId,
        stepIds: bag.stepIdsToExecute,
        steps: bag.stepsToExecute,
        builderApiToken: bag.runToken,
        builderSetupApiToken: bag.runSetupToken
    };
    logger.debug('Message payload - steps', JSON.stringify(messagePayload.steps, null, 2));
    var message = {
        projectId: bag.group.projectId,
        pipelineId: bag.group.pipelineId,
        runId: bag.group.runId,
        payload: messagePayload,
        nodePoolId: bag.nodePool.id,
        priority: bag.group.priority,
        triggeredAt: bag.group.triggeredAt,
        statusCode: bag.statusCodes.ready.code
    };
    bag.suAdapter.postBuildPlaneMessage(message, function (err, message) {
        if (err) {
            logger.debug(who, 'failed to postBuildPlaneMessage for nodePoolId:', message.nodePoolId, 'with err', err);
            return next(true);
        }
        logger.debug(who, 'Successfully posted the message with messageId ' +
            message.id);
        return next();
    });
}
function ___queueToNodePool(bag, next) {
    if (_.isEmpty(bag.stepIdsToExecute))
        return next();
    var who = bag.who + '|' + ___queueToNodePool.name;
    logger.debug(who, 'Inside');
    const instanceSize = bag.nodePool.providerMetadataPropertyBag &&
        bag.nodePool.providerMetadataPropertyBag.instanceSize;
    const updateStep = {
        groupInProgress: true,
        groupStartedAt: new Date(),
        configPropertyBag: {
            isOnDemand: bag.nodePool.isOnDemand,
            instanceSize
        },
        systemPropertyBag: {
            runSetupToken: bag.runSetupToken,
        }
    };
    const messagePayload = {
        affinityGroup: bag.group.affinityGroup,
        runId: bag.group.runId,
        stepIds: bag.stepIdsToExecute,
        steps: bag.stepsToExecute,
        builderApiToken: bag.runToken,
        builderSetupApiToken: bag.runSetupToken
    };
    logger.debug('Message payload with steps', JSON.stringify(messagePayload.steps, null, 2));
    const message = {
        projectId: bag.group.projectId,
        pipelineId: bag.group.pipelineId,
        runId: bag.group.runId,
        payload: messagePayload,
        nodePoolId: bag.nodePool.id,
        priority: bag.group.priority,
        triggeredAt: bag.group.triggeredAt,
        statusCode: bag.statusCodes.ready.code,
        updateStep: updateStep
    };
    bag.suAdapter.queueToNodePool(message, function (err, apiResponse) {
        if (err) {
            logger.debug(who, 'failed ___queueToNodePool for nodePoolId:', message.nodePoolId, 'with err', err);
            return next(true);
        }
        logger.debug(who, 'Succeded ___queueToNodePool the message with messageId ' +
            apiResponse.id);
        return next();
    });
}
function ___getAffinityGroupStepsToExecuteUsingRunAdjacency(bag, next) {
    if (_.isEmpty(bag.group.groupStepIds))
        return next();
    var who = bag.who + '|' + ___getAffinityGroupStepsToExecuteUsingRunAdjacency.name;
    logger.debug(who, 'Inside');
    bag.stepsToExecute = bag.group.getStepsToExecute();
    bag.stepIdsToExecute = _.pluck(bag.stepsToExecute, 'stepId');
    return next();
}
function ___getAffinityGroupStepsToExecute(bag, next) {
    if (_.isEmpty(bag.group.groupStepIds))
        return next();
    var who = bag.who + '|' + ___getAffinityGroupStepsToExecute.name;
    logger.debug(who, 'Inside');
    var steps = bag.stepsByRunId[bag.group.runId];
    var runStepsByStepId = _.indexBy(steps, 'id');
    var groupStepsById = _.indexBy(bag.group.groupStepIds, 'stepId');
    var groupStepIds = _.pluck(bag.group.groupStepIds, 'stepId');
    var _bag = {
        who: bag.who,
        stepIds: groupStepIds,
        runStepConnections: bag.runStepConnectionsByRunId[bag.group.runId],
        runResourceVersionsById: bag.runResourceVersionsById
    };
    sortStepsByDependency(_bag, function (sortedStepIds) {
        bag.group.groupStepIds = _.map(sortedStepIds, function (stepId) {
            return groupStepsById[stepId];
        });
        var statusCodes = global.systemCodesByGroup.status;
        bag.stepsToExecute = _.compact(_.map(bag.group.groupStepIds, function (groupStep) {
            var runStep = runStepsByStepId[groupStep.stepId];
            var step = {
                stepId: groupStep.stepId,
                stepName: runStep.name,
                priority: groupStep.priority,
                inputStepIds: [],
                configPropertyBag: runStep.configPropertyBag
            };
            var stepletIds = [];
            if (_.has(groupStep, 'stepletIds'))
                stepletIds = groupStep.stepletIds;
            else
                stepletIds = _.pluck(bag.stepletsByStepId[groupStep.stepId], 'id');
            step.stepletIds = _.filter(stepletIds, function (stepletId) {
                var steplet = bag.stepletsById[stepletId];
                var stepletConfigPropertyBag = steplet.configPropertyBag;
                return (steplet && steplet.statusCode ===
                    statusCodes.waiting.code && !(stepletConfigPropertyBag &&
                    stepletConfigPropertyBag.isQueuedToNodePool));
            });
            groupStep.stepletIds = step.stepletIds;
            if (!_.isEmpty(step.stepletIds))
                return step;
        }));
        return next();
    });
}
function ___getInSteps(bag, next) {
    if (_.isEmpty(bag.stepsToExecute))
        return next();
    var who = bag.who + '|' + ___getInSteps.name;
    logger.debug(who, 'Inside');
    async.each(bag.stepsToExecute, function (step, nextStep) {
        logger.debug(who + '|' + step.stepId);
        var innerBag = {
            who: bag.who,
            step: step,
            allConnections: bag.runStepConnections,
            runResourceVersionsById: bag.runResourceVersionsById,
            runId: bag.group.runId,
        };
        async.series([
            ____getDirectInConnections.bind(null, innerBag),
            ____getIndirectInConnections.bind(null, innerBag)
        ], () => {
            let inStepIds = [];
            var directINStepIds = _.compact(_.uniq(_.pluck(innerBag.inStepConnections, 'operationStepId')));
            var indirectINStepIds = _.uniq(_.pluck(innerBag.outRunResourceConnections, 'stepId'));
            if (!_.isEmpty(directINStepIds))
                inStepIds = inStepIds.concat(directINStepIds);
            if (!_.isEmpty(indirectINStepIds))
                inStepIds = inStepIds.concat(indirectINStepIds);
            innerBag.step.inputStepIds = _.uniq(inStepIds);
            return nextStep();
        });
    }, function () {
        return next();
    });
}
function ____getDirectInConnections(bag, next) {
    var who = bag.who + '|' + ____getDirectInConnections.name;
    logger.debug(who, 'Inside');
    bag.inStepConnections = _.filter(bag.allConnections, function (connection) {
        return connection.stepId === bag.step.stepId &&
            connection.operationStepId && connection.operation === 'IN';
    });
    bag.inRunResourceConnections = _.filter(bag.allConnections, function (connection) {
        return connection.stepId === bag.step.stepId &&
            connection.operation === 'IN' &&
            connection.operationRunResourceVersionId;
    });
    return next();
}
function ____getIndirectInConnections(bag, next) {
    const who = bag.who + '|' + ____getIndirectInConnections.name;
    logger.debug(who, 'Inside');
    const inputResourceNamesAndBranches = _.uniq(_.compact(_.map(bag.inRunResourceConnections, function (connection) {
        const runResourceVersion = bag.runResourceVersionsById[connection.operationRunResourceVersionId];
        if (!runResourceVersion)
            return null;
        return {
            resourceName: connection.operationRunResourceVersionName,
            pipelineSourceBranch: runResourceVersion.pipelineSourceBranch
        };
    })));
    if (_.isEmpty(inputResourceNamesAndBranches))
        return next();
    bag.outRunResourceConnections = _.filter(bag.allConnections, function (connection) {
        if (connection.operation !== 'OUT')
            return false;
        if (!connection.operationRunResourceVersionId)
            return false;
        if (connection.runId !== bag.runId)
            return false;
        const runResourceVersion = bag.runResourceVersionsById[connection.operationRunResourceVersionId];
        if (!runResourceVersion)
            return false;
        return _.findWhere(inputResourceNamesAndBranches, {
            resourceName: connection.operationRunResourceVersionName,
            pipelineSourceBranch: runResourceVersion.pipelineSourceBranch
        });
    });
    return next();
}
function ___prioritizeStepsToExecute(bag, next) {
    if (_.isEmpty(bag.stepsToExecute))
        return next();
    var who = bag.who + '|' + ___prioritizeStepsToExecute.name;
    logger.debug(who, 'Inside');
    _.each(bag.stepsToExecute, function (step) {
        step.sequence = 0;
    });
    var stepsByStepId = _.indexBy(bag.stepsToExecute, 'stepId');
    bag.stepsToExecute = _.map(bag.stepsToExecute, function (step) {
        let parentSequence = 0;
        _.each(step.inputStepIds, function (inputStepId) {
            var inputStep = stepsByStepId[inputStepId];
            if (inputStep) {
                if (parentSequence < inputStep.sequence) {
                    parentSequence = inputStep.sequence;
                }
            }
        });
        step.sequence = Number(parentSequence) + 1;
        if (!step.priority) {
            step.priority = bag.defaultPriority;
        }
        return step;
    });
    var sequenceGroupedSteps = _.groupBy(bag.stepsToExecute, 'sequence');
    var orderedSteps = _.map(sequenceGroupedSteps, function (sequenceSteps) {
        return _.sortBy(sequenceSteps, 'priority');
    });
    bag.stepsToExecute = _.flatten(_.values(orderedSteps), 1);
    bag.stepIdsToExecute = _.pluck(bag.stepsToExecute, 'stepId');
    return next();
}
function ___postStepInfoConsoles(bag, next) {
    if (bag.error)
        return next();
    if (bag.group.reject)
        return next();
    if (bag.group.skipByNoNode || bag.group.skipByNodePoolPermissions
        || bag.group.skipByNonRootEnforcement)
        return next();
    var who = bag.who + '|' + ___postStepInfoConsoles.name;
    logger.debug(who, 'Inside');
    async.eachLimit(bag.group.groupStepIds, 10, function (stepObj, done) {
        if (_.isEmpty(stepObj.stepletIds))
            return done();
        var consoleAdapter = new BatchStepletConsoleAdapter(bag.jFrogPipelinesSessionId, global.config.apiToken, stepObj.stepletIds, bag.group.pipelineId, global.systemConfig.jobConsoleBatchSize, global.systemConfig.jobConsoleBufferTimeInterval);
        var grpMsg = 'Queuing step';
        var logMessage = 'Successfully queued step to node pool ' +
            bag.nodePool.name + '. Waiting for node to start execution.';
        var closeCmd = true;
        var closeGrp = true;
        consoleAdapter.openGrp(grpMsg);
        consoleAdapter.openCmd('Info');
        consoleAdapter.publishMsg(logMessage);
        consoleAdapter.closeCmd(closeCmd);
        consoleAdapter.closeGrp(closeGrp);
        return done();
    }, function (err) {
        return next(err);
    });
}
function ___startCachedNode(innerBag, next) {
    var who = innerBag.who + '|' + ___startCachedNode.name;
    logger.debug(who, 'Inside');
    var statusBody = {
        statusCode: global.statusCodes.WAITING
    };
    innerBag.suAdapter.putNodeById(innerBag.nodeId, statusBody, function (err, node) {
        if (err) {
            logger.debug(who, 'failed to :putNode for :nodeId', innerBag.nodeId, ' with err', err);
            return next(err);
        }
        innerBag.node = node;
        return next();
    });
}
function __releaseLock(bag, next) {
    if (!bag.lock)
        return next();
    var who = bag.who + '|' + __releaseLock.name;
    logger.debug(who, 'Inside');
    bag.lock.release(function (err) {
        if (err)
            logger.debug('failed to release lock');
        return next();
    });
}
function _setResult(bag, next) {
    var who = bag.who + '|' + _setResult.name;
    logger.debug(who, 'Inside');
    bag.result = {
        waitingGroups: bag.waitingGroups,
        processingGroups: bag.processingGroups
    };
    return next();
}
function __publishToNPM(bag, next) {
    const who = bag.who + '|' + __publishToNPM.name;
    logger.debug(who, 'Inside');
    let msg = {
        where: 'core.nodePoolService',
        payload: {
            nodePool: {
                nodePoolId: bag.nodePool.id
            }
        }
    };
    logger.trace('Sending the message to rabbitmq', JSON.stringify(msg));
    bag.suAdapter.postToVortex(msg, (err) => {
        if (err) {
            logger.debug(who, 'failed to :postToVortex for :nodePoolId', bag.nodePool.id, 'with error:', err);
        }
        return next();
    });
}
//# sourceMappingURL=queueStepsToNodePools.js.map