'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
var self = sendApprovalRequests;
module.exports = self;
var { CALL_OBJECT_LIMIT_FOR_BATCH } = require('../constants.js');
var LOCK_TIMEOUT_IN_MS = 1 * 60 * 1000;
var LOCK_TIMER_INTERVAL_IN_MS = 50 * 1000;
var { getElapsedTimeSinceStartTime } = require('./utils');
var Lock = require('../Lock.js');
const { DataAdapterFactory } = require('../modules/index.js');
const logger_1 = require("@jfrog-pipelines/logger");
function sendApprovalRequests(externalBag, callback) {
    var who = externalBag.who + '|' + sendApprovalRequests.name;
    logger.debug(who, 'Inside');
    let startTime = Date.now();
    var bag = {
        who: who,
        suAdapter: externalBag.suAdapter,
        waitingGroups: externalBag.waitingGroups,
        rejectedGroups: externalBag.rejectedGroups,
        groupsRequiringApproval: [],
        jFrogPipelinesSessionId: externalBag.jFrogPipelinesSessionId,
        useRunAdjacency: externalBag.useRunAdjacency,
        result: {}
    };
    async.series([
        _checkInputParams.bind(null, bag),
        _filterWaitingGroups.bind(null, bag),
        _sendApprovalRequests.bind(null, bag)
    ], function (err) {
        let finalTime = getElapsedTimeSinceStartTime(startTime);
        logger.debug(bag.who, 'Finished sendApprovalRequests -', finalTime);
        return callback(err, bag.result);
    });
}
function _checkInputParams(bag, next) {
    var who = bag.who + '|' + _checkInputParams.name;
    logger.debug(who, 'Inside');
    if (!bag.suAdapter)
        return next(new Error(':suAdapter is missing'));
    return next();
}
function _filterWaitingGroups(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _filterWaitingGroups.name;
    logger.debug(who, 'Inside');
    var partitioned = _.partition(bag.waitingGroups, function (group) {
        return group.requiresApproval && !group.approvalRequestedAt;
    });
    bag.groupsRequiringApproval = partitioned[0];
    bag.result.waitingGroups = partitioned[1];
    _.each(bag.groupsRequiringApproval, function (group) {
        if (!group.waitingForApproval) {
            group.rejectionContext.waitingForApproval = {
                timestamp: new Date()
            };
            bag.rejectedGroups.waitingForApproval.push(group);
        }
    });
    return next();
}
function _sendApprovalRequests(bag, next) {
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + _sendApprovalRequests.name;
    logger.debug(who, 'Inside');
    var groupsRequiringApprovalByProjectId = _.groupBy(bag.groupsRequiringApproval, 'projectId');
    async.eachLimit(_.keys(groupsRequiringApprovalByProjectId), 5, function (projectId, nextProject) {
        projectId = parseInt(projectId, 10);
        var innerBag = {
            who: bag.who + '|projectId:' + projectId,
            lock: {
                isValid: false,
                release: () => { },
                destroy: () => { }
            },
            suAdapter: bag.suAdapter,
            projectId: projectId,
            waitingGroups: groupsRequiringApprovalByProjectId[projectId] || [],
            waitingSteps: [],
            jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId,
            error: false,
            totalApprovalRequests: 0
        };
        const tasks = [
            __obtainApprovalRequestLock.bind(null, innerBag),
        ];
        if (bag.useRunAdjacency) {
            tasks.push(__getWaitingStepsUsingRunAdjacency.bind(null, innerBag));
        }
        else {
            tasks.push(__getWaitingSteps.bind(null, innerBag));
        }
        tasks.push(__checkIfStepsShouldBeSkipped.bind(null, innerBag), __requestApprovals.bind(null, innerBag), __releaseApprovalRequestLock.bind(null, innerBag));
        async.series(tasks, function (err) {
            let sendRequestApprovalsStatus = '';
            if (err || innerBag.error) {
                sendRequestApprovalsStatus = 'failed';
                logger.warn(bag.who, 'Failed to request approvals for project', projectId, 'with error:', err || bag.error);
            }
            else {
                sendRequestApprovalsStatus = 'success';
                logger.debug(bag.who, 'Requested approvals for steps in project', projectId);
            }
            if (innerBag.totalApprovalRequests && innerBag.totalApprovalRequests > 0) {
                const eventData = {
                    timestamp: new Date().toISOString(),
                    domain: global.msName,
                    eventType: 'approvalrequests',
                    status: sendRequestApprovalsStatus,
                    noOfRequests: innerBag.totalApprovalRequests,
                    projectId: projectId,
                    pipelinesSessionId: bag.jFrogPipelinesSessionId
                };
                logger_1.LoggerFactory.logEvent(eventData);
            }
            if (bag.lock) {
                bag.lock.destroy();
                bag.lock = null;
            }
            return nextProject();
        });
    }, function () {
        return next();
    });
}
function __obtainApprovalRequestLock(bag, next) {
    var who = bag.who + '|' + __obtainApprovalRequestLock.name;
    logger.debug(who, 'Inside');
    var params = {
        name: global.msName + '-stepApprovalRequest-' + bag.projectId,
        lockTimeoutInMs: LOCK_TIMEOUT_IN_MS,
        lockTimerInMs: LOCK_TIMER_INTERVAL_IN_MS,
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    bag.lock = new Lock(params);
    bag.lock.obtain(function (err) {
        if (err)
            logger.debug('failed to obtain lock with name:', params.name);
        return next(err);
    });
}
function __getWaitingStepsUsingRunAdjacency(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __getWaitingStepsUsingRunAdjacency.name;
    logger.debug(who, 'Inside');
    bag.waitingSteps = [];
    const waitingStepIds = [];
    for (const waitingGroup of bag.waitingGroups) {
        waitingStepIds.push(...waitingGroup.getApprovalGatesStepIds());
    }
    if (waitingStepIds.length === 0) {
        return next();
    }
    const dataAdapter = DataAdapterFactory.getApiDataAdapter();
    dataAdapter.getStepsWaitingForApprovalRequest(waitingStepIds)
        .then((waitingSteps) => {
        bag.waitingSteps = waitingSteps;
        return next();
    })
        .catch((err) => {
        return next(err);
    });
}
function __getWaitingSteps(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __getWaitingSteps.name;
    logger.debug(who, 'Inside');
    var stepIds = _.uniq(_.pluck(_.flatten(_.pluck(bag.waitingGroups, 'groupStepIds')), 'stepId'));
    if (_.isEmpty(stepIds))
        return next();
    var stepIdGroups = [];
    var i = 0;
    while (i < stepIds.length) {
        stepIdGroups.push(stepIds.slice(i, i + CALL_OBJECT_LIMIT_FOR_BATCH));
        i += CALL_OBJECT_LIMIT_FOR_BATCH;
    }
    async.eachLimit(stepIdGroups, 10, function (stepIdGroup, done) {
        var query = util.format('stepIds=%s&statusCodes=%s&limit=%s', stepIdGroup.join(','), global.systemCodesByGroup.status.waiting.code, CALL_OBJECT_LIMIT_FOR_BATCH);
        bag.suAdapter.getSteps(query, function (err, steps) {
            if (err) {
                logger.warn(who, 'getSteps for query:', query, 'returned error:', err);
                return done(err);
            }
            bag.waitingSteps = bag.waitingSteps.concat(_.filter(steps, function (step) {
                return !step.approvalRequestedAt && step.requiresApproval;
            }));
            return done();
        });
    }, function (err) {
        if (err)
            bag.error = true;
        return next();
    });
}
function __checkIfStepsShouldBeSkipped(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingGroups))
        return next();
    var who = bag.who + '|' + __checkIfStepsShouldBeSkipped.name;
    logger.debug(who, 'Inside');
    async.eachLimit(bag.waitingSteps, 10, function (step, done) {
        bag.suAdapter.getStepExecutionEligibility(step.id, function (err, stepExecutionResult) {
            if (err) {
                logger.warn(who, util.format('Failed to check execution ' +
                    'eligibility of step %s with error: %s', step.id, err));
                return done(true);
            }
            if (!stepExecutionResult.canExecuteStep) {
                logger.debug(who, util.format('Step %s will be skipped', step.id));
                step.skipStep = true;
            }
            return done();
        });
    }, function (err) {
        if (err)
            bag.error = true;
        return next();
    });
}
function __requestApprovals(bag, next) {
    if (bag.error)
        return next();
    if (!bag.lock.isValid)
        return next();
    if (_.isEmpty(bag.waitingSteps))
        return next();
    var who = bag.who + '|' + __requestApprovals.name;
    logger.debug(who, 'Inside');
    let approvalRequests = 0;
    async.eachSeries(bag.waitingSteps, function (step, done) {
        if (!bag.lock.isValid) {
            logger.error(who, 'invalid lock, exiting');
            return done(true);
        }
        if (step.skipStep)
            return process.nextTick(done);
        bag.suAdapter.requestStepApproval(step.id, function (err) {
            if (err) {
                logger.warn(who, 'Failed to requestStepApproval for id:', step.id, 'with error:', err);
                return done(true);
            }
            approvalRequests++;
            return done();
        });
    }, function () {
        bag.totalApprovalRequests = approvalRequests;
        return next();
    });
}
function __releaseApprovalRequestLock(bag, next) {
    if (!bag.lock)
        return next();
    var who = bag.who + '|' + __releaseApprovalRequestLock.name;
    logger.debug(who, 'Inside');
    bag.lock.release(function (err) {
        if (err)
            logger.warn('failed to release lock');
        return next();
    });
}
//# sourceMappingURL=sendApprovalRequests.js.map