'use strict';
const self = sendRejectedStepletConsoles;
module.exports = self;
const _ = require('underscore');
const async = require('async');
const LOCK_TIMEOUT_IN_MS = 1 * 60 * 1000;
const LOCK_TIMER_INTERVAL_IN_MS = 50 * 1000;
const BatchStepletConsoleAdapter = require('../BatchStepletConsoleAdapter.js');
const Lock = require('../Lock.js');
const { waitingMessageComposerFactory, waitingConditions } = require('./constants.js');
function sendRejectedStepletConsoles(externalBag, callback) {
    const who = externalBag.who + '|' + sendRejectedStepletConsoles.name;
    logger.debug(who, 'Inside');
    const bag = {
        who: who,
        suAdapter: externalBag.suAdapter,
        lock: {},
        rejectedGroups: externalBag.rejectedGroups,
        jFrogPipelinesSessionId: externalBag.jFrogPipelinesSessionId
    };
    async.series([
        _checkInputParams.bind(null, bag),
        _obtainLock.bind(null, bag),
        _computeMessages.bind(null, bag),
        _updateSteplets.bind(null, bag),
        _sendStepletConsoles.bind(null, bag),
    ], function (err) {
        _releaseLock(bag, function () {
            if (err)
                return callback(err);
            return callback(null, bag.resBody);
        });
    });
}
function _checkInputParams(bag, next) {
    const who = bag.who + '|' + _checkInputParams.name;
    logger.debug(who, 'Inside');
    if (!bag.suAdapter)
        return next(new Error(':suAdapter is missing'));
    return next();
}
function _obtainLock(bag, next) {
    const who = bag.who + '|' + _obtainLock.name;
    logger.debug(who, 'Inside');
    const params = {
        name: global.msName + '-send-rejected-steps-consoles',
        lockTimeoutInMs: LOCK_TIMEOUT_IN_MS,
        lockTimerInMs: LOCK_TIMER_INTERVAL_IN_MS,
        jFrogPipelinesSessionId: bag.jFrogPipelinesSessionId
    };
    bag.lock = new Lock(params);
    bag.lock.obtain(function (err) {
        if (err)
            logger.debug('failed to obtain lock with name:', params.name);
        return next(err);
    });
}
function _computeMessages(bag, next) {
    if (!bag.lock)
        return next();
    const who = bag.who + '|' + _computeMessages.name;
    logger.debug(who, 'Inside');
    let stepletConsoles = [];
    _.each(waitingConditions, function (condition) {
        let groups = bag.rejectedGroups[condition];
        let messageComposer = waitingMessageComposerFactory[condition];
        if (_.isEmpty(groups) || _.isUndefined(messageComposer))
            return;
        _.each(groups, function (group) {
            let rejectionContext = group.rejectionContext[condition] || {};
            let consoles = messageComposer(rejectionContext);
            let consolesByStepletId = _.indexBy(consoles, 'stepletId') || {};
            if (_.isEmpty(consoles))
                return;
            _.each(group.steplets, function (steplet) {
                stepletConsoles.push(createConsole(steplet, group, condition, rejectionContext, consolesByStepletId[steplet.id] || consoles));
            });
        });
    });
    bag.stepletConsoles = stepletConsoles;
    return next();
    function createConsole(steplet, group, condition, rejectionContext, console) {
        return {
            steplet: steplet,
            pipelineId: group.pipelineId,
            condition: condition,
            timestamp: rejectionContext.timestamp || new Date(),
            console: console,
        };
    }
}
function _sendStepletConsoles(bag, next) {
    if (_.isEmpty(bag.stepletConsoles))
        return next();
    if (!bag.lock)
        return next();
    const who = bag.who + '|' + _sendStepletConsoles.name;
    logger.debug(who, 'Inside');
    async.eachLimit(bag.stepletConsoles, 10, function (stepletConsole, nextStepletConsole) {
        let consoleAdapter = new BatchStepletConsoleAdapter(bag.jFrogPipelinesSessionId, config.apiToken, [stepletConsole.steplet.id], stepletConsole.pipelineId, global.systemConfig.jobConsoleBatchSize, global.systemConfig.jobConsoleBufferTimeInterval);
        consoleAdapter.openGrp(stepletConsole.console.group);
        consoleAdapter.openCmd(stepletConsole.console.cmd);
        consoleAdapter.publishMsg(stepletConsole.console.message);
        consoleAdapter.closeCmd(true);
        consoleAdapter.closeGrp(true);
        consoleAdapter.flushConsoles(nextStepletConsole);
    }, function () {
        return next();
    });
}
function _updateSteplets(bag, next) {
    if (_.isEmpty(bag.stepletConsoles))
        return next();
    if (!bag.lock)
        return next();
    const who = bag.who + '|' + _updateSteplets.name;
    logger.debug(who, 'Inside');
    async.eachLimit(bag.stepletConsoles, 10, function ({ steplet, condition, timestamp }, nextStepletConsole) {
        let update = {
            runtimePropertyBag: {
                ...steplet.runtimePropertyBag,
                [condition]: timestamp
            }
        };
        bag.suAdapter.putStepletById(steplet.id, update, function (err) {
            if (err)
                return nextStepletConsole(err);
            return process.nextTick(nextStepletConsole);
        });
    }, function (err) {
        return next(err);
    });
}
function _releaseLock(bag, next) {
    if (!bag.lock)
        return next();
    const who = bag.who + '|' + _releaseLock.name;
    logger.debug(who, 'Inside');
    bag.lock.release(function (err) {
        if (err)
            logger.warn('failed to release lock');
        return next();
    });
}
//# sourceMappingURL=sendRejectedStepletConsoles.js.map