'use strict';
var self = sortStepsByDependency;
module.exports = self;
var _ = require('underscore');
var async = require('async');
function sortStepsByDependency(externalBag, callback) {
    var who = externalBag.who + '|' + sortStepsByDependency.name;
    logger.debug(who, 'Inside');
    var stepIds = externalBag.stepIds;
    var stepsById = _.object(stepIds, stepIds);
    var StepsCache = {};
    var runStepConnections = _.filter(externalBag.runStepConnections, function (connection) {
        return stepsById[connection.stepId];
    });
    var runResourceVersionsById = externalBag.runResourceVersionsById;
    var runStepConnectionsByOperationStepId = _.groupBy(runStepConnections, 'operationStepId');
    var runStepConnectionsByStepId = _.groupBy(runStepConnections, 'stepId');
    var runStepConnectionsByResourceName = _.groupBy(runStepConnections, 'operationRunResourceVersionName');
    var bag = {
        who: who,
        StepsCache: StepsCache,
        runStepConnectionsByOperationStepId: runStepConnectionsByOperationStepId,
        runStepConnectionsByStepId: runStepConnectionsByStepId,
        runStepConnectionsByResourceName: runStepConnectionsByResourceName,
        runResourceVersionsById: runResourceVersionsById,
        sortedSteps: [],
        stepsById: stepsById,
        stepIds: stepIds
    };
    async.series([
        _computeStepDependencies.bind(null, bag),
        _sortStepsByInputOrder.bind(null, bag)
    ], function () {
        return callback(bag.sortedSteps);
    });
}
function _computeStepDependencies(bag, next) {
    var who = bag.who + '|' + _computeStepDependencies.name;
    logger.debug(who, 'Inside');
    var computeStepDependencies = _getStepDependencyComputer(bag.stepsById, bag.runResourceVersionsById, bag.runStepConnectionsByStepId, bag.runStepConnectionsByOperationStepId, bag.runStepConnectionsByResourceName, bag.StepsCache);
    async.eachSeries(bag.stepIds, function (stepId, done) {
        if (!_.isUndefined(bag.StepsCache[stepId]))
            return process.nextTick(done);
        computeStepDependencies(stepId, function () {
            return process.nextTick(done);
        });
    }, function () {
        return process.nextTick(next);
    });
}
function _sortStepsByInputOrder(bag, next) {
    var who = bag.who + '|' + _sortStepsByInputOrder.name;
    logger.debug(who, 'Inside');
    var groupByIndegrees = _.groupBy(bag.StepsCache, 'indegrees');
    var result = [];
    var stepsQueue = groupByIndegrees[0];
    async.whilst(function () {
        return !_.isEmpty(stepsQueue);
    }, function (done) {
        var stepObject = stepsQueue.shift();
        _.each(stepObject.nextSteps, function (nextStep) {
            nextStep.indegrees--;
            if (nextStep.indegrees === 0)
                stepsQueue.push(nextStep);
        });
        result.push(stepObject.stepId);
        return process.nextTick(done);
    }, function () {
        bag.sortedSteps = _.filter(result, function (stepId) {
            return bag.stepsById[stepId];
        });
        return process.nextTick(next);
    });
}
function _getStepDependencyComputer(stepsById, runResourceVersionsById, runStepConnectionsByStepId, runStepConnectionsByOperationStepId, runStepConnectionsByResourceName, cache) {
    return function (stepId, callback) {
        var queue = [{
                stepId: stepId,
            }];
        _computeNextStep({
            currentStepId: -1,
            nextStepId: stepId,
            cache
        });
        var allRunResourceVersionNamesQueued = [];
        async.whilst(function () {
            return !_.isEmpty(queue);
        }, function (done) {
            var element = queue.shift();
            if (element.stepId) {
                _.each(runStepConnectionsByOperationStepId[element.stepId], function (connection) {
                    if (connection.operation !== 'IN')
                        return;
                    if (!_.isObject(stepsById[connection.stepId])) {
                        _computeNextStep({
                            currentStepId: element.stepId,
                            nextStepId: connection.stepId,
                            cache
                        });
                        queue.push({
                            stepId: connection.stepId
                        });
                    }
                });
                var runStepConnectionsForStep = runStepConnectionsByStepId[element.stepId];
                _.each(runStepConnectionsForStep, function (connection) {
                    if (connection.operation !== 'OUT')
                        return;
                    if (!connection.operationRunResourceVersionId)
                        return;
                    var runResourceVersion = runResourceVersionsById[connection.operationRunResourceVersionId];
                    if (!runResourceVersion)
                        return;
                    var resourceNameAndBranch = {
                        runResourceVersionName: connection.operationRunResourceVersionName,
                        pipelineSourceBranch: runResourceVersion.pipelineSourceBranch,
                        inStepId: element.stepId,
                    };
                    if (_.findWhere(allRunResourceVersionNamesQueued, resourceNameAndBranch))
                        return;
                    queue.push(resourceNameAndBranch);
                    allRunResourceVersionNamesQueued.push(resourceNameAndBranch);
                });
            }
            else if (element.runResourceVersionName) {
                var resDownstreamINConnections = _.filter(runStepConnectionsByResourceName[element.runResourceVersionName], function (connection) {
                    var runResourceVersion = runResourceVersionsById[connection.operationRunResourceVersionId];
                    if (!runResourceVersion)
                        return false;
                    return connection.operation === 'IN' &&
                        runResourceVersion.pipelineSourceBranch ===
                            element.pipelineSourceBranch;
                });
                _.each(resDownstreamINConnections, function (connection) {
                    if (!_.isObject(stepsById[connection.stepId])) {
                        _computeNextStep({
                            currentStepId: element.inStepId,
                            nextStepId: connection.stepId,
                            cache
                        });
                        queue.push({
                            stepId: connection.stepId
                        });
                    }
                });
            }
            return process.nextTick(done);
        }, function () {
            return process.nextTick(callback);
        });
    };
}
function _getStepObject(stepId, cache) {
    if (cache[stepId])
        return cache[stepId];
    var stepObject = {
        stepId,
        nextSteps: [],
        inputStepIds: [],
        indegrees: 0
    };
    cache[stepId] = stepObject;
    return cache[stepId];
}
function _computeNextStep({ currentStepId, nextStepId, cache }) {
    var currentStep = _getStepObject(currentStepId, cache);
    var nextStep = _getStepObject(nextStepId, cache);
    var nextStepIds = _.pluck(currentStep.nextSteps, 'stepId');
    if (!_.contains(nextStepIds, nextStep.stepId))
        currentStep.nextSteps.push(nextStep);
    if (!_.contains(nextStep.inputStepIds, currentStep.stepId)) {
        nextStep.inputStepIds.push(currentStepId);
        nextStep.indegrees = nextStep.inputStepIds.length;
    }
    return;
}
//# sourceMappingURL=sortStepsByDependency.js.map