'use strict';
var self = sortStepsByDependency;
module.exports = self;
var _ = require('underscore');
function sortStepsByDependency(stepIds, runStepConnections, runResourceVersionsById) {
    var stepsById = _.object(stepIds, stepIds);
    var StepsCache = {};
    var computeStepDependencies = _getStepDependencyComputer(stepsById, StepsCache);
    _.each(stepIds, function (stepId) {
        if (_.isUndefined(StepsCache[stepId]))
            computeStepDependencies({
                stepId,
                runStepConnections,
                runResourceVersionsById
            });
    });
    var sortedSteps = _sortStepsByInputOrder(StepsCache);
    return _.filter(sortedSteps, function (stepId) {
        return !_.isUndefined(stepsById[stepId]);
    });
}
function _getStepDependencyComputer(stepsById, cache) {
    return function ({ stepId, runStepConnections, runResourceVersionsById }) {
        var queue = [{
                stepId: stepId,
            }];
        _computeNextStep({
            currentStepId: -1,
            nextStepId: stepId,
            cache
        });
        var allRunResourceVersionNamesQueued = [];
        while (!_.isEmpty(queue)) {
            var element = queue.shift();
            if (element.stepId) {
                var stepDownstreamINConnections = _.where(runStepConnections, {
                    operation: 'IN',
                    operationStepId: element.stepId
                });
                _.each(stepDownstreamINConnections, function (connection) {
                    if (!_.isObject(stepsById[connection.stepId])) {
                        _computeNextStep({
                            currentStepId: element.stepId,
                            nextStepId: connection.stepId,
                            cache
                        });
                        queue.push({
                            stepId: connection.stepId
                        });
                    }
                });
                var downstreamOUTConnections = _.where(runStepConnections, {
                    operation: 'OUT',
                    stepId: element.stepId
                });
                _.each(downstreamOUTConnections, function (connection) {
                    if (!connection.operationRunResourceVersionId)
                        return;
                    var runResourceVersion = runResourceVersionsById[connection.operationRunResourceVersionId];
                    if (!runResourceVersion)
                        return;
                    var resourceNameAndBranch = {
                        runResourceVersionName: connection.operationRunResourceVersionName,
                        pipelineSourceBranch: runResourceVersion.pipelineSourceBranch,
                        inStepId: element.stepId,
                    };
                    if (_.findWhere(allRunResourceVersionNamesQueued, resourceNameAndBranch))
                        return;
                    queue.push(resourceNameAndBranch);
                    allRunResourceVersionNamesQueued.push(resourceNameAndBranch);
                });
            }
            else if (element.runResourceVersionName) {
                var resDownstreamINConnections = _.filter(runStepConnections, function (connection) {
                    var runResourceVersion = runResourceVersionsById[connection.operationRunResourceVersionId];
                    if (!runResourceVersion)
                        return false;
                    return connection.operation === 'IN' &&
                        connection.operationRunResourceVersionName ===
                            element.runResourceVersionName &&
                        runResourceVersion.pipelineSourceBranch ===
                            element.pipelineSourceBranch;
                });
                _.each(resDownstreamINConnections, function (connection) {
                    if (!_.isObject(stepsById[connection.stepId])) {
                        _computeNextStep({
                            currentStepId: element.inStepId,
                            nextStepId: connection.stepId,
                            cache
                        });
                        queue.push({
                            stepId: connection.stepId
                        });
                    }
                });
            }
        }
    };
}
function _getStepObject(stepId, cache) {
    if (cache[stepId])
        return cache[stepId];
    var stepObject = {
        stepId,
        nextSteps: [],
        inputStepIds: [],
        indegrees: 0
    };
    cache[stepId] = stepObject;
    return cache[stepId];
}
function _computeNextStep({ currentStepId, nextStepId, cache }) {
    var currentStep = _getStepObject(currentStepId, cache);
    var nextStep = _getStepObject(nextStepId, cache);
    var nextStepIds = _.pluck(currentStep.nextSteps, 'stepId');
    if (!_.contains(nextStepIds, nextStep.stepId))
        currentStep.nextSteps.push(nextStep);
    if (!_.contains(nextStep.inputStepIds, currentStep.stepId)) {
        nextStep.inputStepIds.push(currentStepId);
        nextStep.indegrees = nextStep.inputStepIds.length;
    }
    return;
}
function _sortStepsByInputOrder(input) {
    var groupByIndegrees = _.groupBy(input, 'indegrees');
    var result = [];
    var stepsQueue = groupByIndegrees[0];
    while (!_.isEmpty(stepsQueue)) {
        var stepObject = stepsQueue.shift();
        for (var nextStepObject of stepObject.nextSteps) {
            nextStepObject.indegrees--;
            if (nextStepObject.indegrees === 0)
                stepsQueue.push(nextStepObject);
        }
        result.push(stepObject.stepId);
    }
    return result;
}
//# sourceMappingURL=sortStepsByDependencySync.js.map