'use strict';
var self = Adapter;
module.exports = self;
var util = require('util');
var async = require('async');
var parseLinks = require('parse-links');
var parseJsonBody = require('../parseJsonBody');
function Adapter(jFrogPipelinesSessionId, token) {
    logger.trace(util.format('Initializing %s for jFrogPipelinesSessionId: %s', self.name, jFrogPipelinesSessionId));
    this.baseUrl = global.config.artifactoryUrl;
    this.jFrogPipelinesSessionId = jFrogPipelinesSessionId;
    if (token)
        this.authorizationHeader = 'Bearer ' + token;
}
Adapter.prototype.get = function (relativeUrl, options, callback) {
    var opts = {
        method: 'GET',
        url: relativeUrl.indexOf('http') === 0 ? relativeUrl : this.baseUrl +
            relativeUrl,
        headers: {
            'Authorization': this.authorizationHeader,
            'Content-Type': 'application/json',
            'User-Agent': 'Pipelines API'
        },
        timeout: 180000
    };
    var bag = {
        opts: opts,
        relativeUrl: relativeUrl,
        token: this.token,
        retry: options && options.retry,
        jFrogPipelinesSessionId: this.jFrogPipelinesSessionId
    };
    bag.who = util.format('common|artifactory|%s|GET|url:%s|' +
        'jFrogPipelinesSessionId:%s', self.name, relativeUrl, this.jFrogPipelinesSessionId);
    logger.trace('Starting', bag.who);
    async.series([
        _performCall.bind(null, bag),
        _parseResponse.bind(null, bag)
    ], function () {
        logger.trace('Completed', bag.who);
        callback(bag.err, bag.parsedBody, bag.headerLinks, bag.res);
    });
};
Adapter.prototype.post = function (relativeUrl, options, data, callback) {
    var opts = {
        method: 'POST',
        url: relativeUrl.indexOf('http') === 0 ? relativeUrl : this.baseUrl +
            relativeUrl,
        headers: {
            'Authorization': this.authorizationHeader,
            'Content-Type': 'text/plain',
            'User-Agent': 'Pipelines API'
        },
        body: data,
        timeout: 180000
    };
    var bag = {
        opts: opts,
        relativeUrl: relativeUrl,
        token: this.token,
        retry: options && options.retry,
        jFrogPipelinesSessionId: this.jFrogPipelinesSessionId
    };
    bag.who = util.format('common|artifactory|%s|POST|url:%s|' +
        'jFrogPipelinesSessionId:%s', self.name, relativeUrl, this.jFrogPipelinesSessionId);
    logger.trace('Starting', bag.who);
    async.series([
        _performCall.bind(null, bag),
        _parseResponse.bind(null, bag)
    ], function () {
        logger.trace('Completed', bag.who);
        callback(bag.err, bag.parsedBody, bag.headerLinks, bag.res);
    });
};
Adapter.prototype.put = function (relativeUrl, data, callback) {
    var opts = {
        method: 'PUT',
        url: relativeUrl.indexOf('http') === 0 ? relativeUrl : this.baseUrl +
            relativeUrl,
        headers: {
            'Authorization': this.authorizationHeader,
            'Content-Type': 'application/json',
            'User-Agent': 'Pipelines API',
            'Accept': '*/*'
        },
        json: data,
        timeout: 180000
    };
    var bag = {
        opts: opts,
        relativeUrl: relativeUrl,
        token: this.token,
        jFrogPipelinesSessionId: this.jFrogPipelinesSessionId
    };
    bag.who = util.format('common|artifactory|%s|PUT|url:%s|' +
        'jFrogPipelinesSessionId:%s', self.name, relativeUrl, this.jFrogPipelinesSessionId);
    logger.trace('Starting', bag.who);
    async.series([
        _performCall.bind(null, bag),
        _parseResponse.bind(null, bag)
    ], function () {
        logger.trace('Completed', bag.who);
        callback(bag.err, bag.parsedBody, bag.headerLinks, bag.res);
    });
};
Adapter.prototype.delete = function (relativeUrl, callback) {
    var opts = {
        method: 'DELETE',
        url: relativeUrl.indexOf('http') === 0 ? relativeUrl : this.baseUrl +
            relativeUrl,
        headers: {
            'Authorization': this.authorizationHeader,
            'Content-Type': 'application/json',
            'User-Agent': 'Pipelines API'
        },
        timeout: 180000
    };
    var bag = {
        opts: opts,
        relativeUrl: relativeUrl,
        token: this.token,
        jFrogPipelinesSessionId: this.jFrogPipelinesSessionId
    };
    bag.who = util.format('common|artifactory|%s|DELETE|url:%s|' +
        'jFrogPipelinesSessionId:%s', self.name, relativeUrl, this.jFrogPipelinesSessionId);
    logger.trace('Starting', bag.who);
    async.series([
        _performCall.bind(null, bag),
        _parseResponse.bind(null, bag)
    ], function () {
        logger.trace('Completed', bag.who);
        callback(bag.err, bag.parsedBody, bag.headerLinks, bag.res);
    });
};
function _performCall(bag, next) {
    var who = bag.who + '|' + _performCall.name;
    logger.trace('Inside', who);
    bag.startedAt = Date.now();
    var retryOpts = {
        times: bag.retry ? 5 : 1,
        interval: function (retryCount) {
            return Math.pow(2, retryCount) * 1000;
        }
    };
    async.retry(retryOpts, function (callback) {
        request(bag.opts, function (err, res, body) {
            var interval = Date.now() - bag.startedAt;
            logger.trace(who, 'Pipelines request ' + bag.opts.method + ' ' +
                bag.opts.url + ' took ' + interval +
                ' ms and returned HTTP status ' + (res && res.statusCode));
            if (res && res.statusCode > 299)
                err = err || res.statusCode;
            if (err)
                logger.trace(who, 'Artifactory returned status', err);
            bag.err = err;
            bag.res = res;
            bag.body = body;
            return callback(err);
        });
    }, function () {
        return next();
    });
}
function _parseResponse(bag, next) {
    var who = bag.who + '|' + _parseResponse.name;
    logger.trace('Inside', who);
    if (bag.res && bag.res.headers.link) {
        bag.headerLinks = parseLinks(bag.res.headers.link);
    }
    if (bag.body) {
        if (typeof bag.body === 'object') {
            bag.parsedBody = bag.body;
        }
        else if (bag.res && bag.res.headers &&
            _.isString(bag.res.headers['content-type']) &&
            bag.res.headers['content-type'].indexOf('text/plain') >= 0) {
            bag.parsedBody = bag.body;
        }
        else {
            bag.parsedBody = parseJsonBody(bag);
        }
    }
    next();
}
Adapter.prototype.aql = function (body, callback) {
    var relUrl = util.format('/api/search/aql');
    this.post(relUrl, { retry: true }, body, callback);
};
Adapter.prototype.deleteArtifactsByPath = function (path, callback) {
    var relUrl = util.format('/%s', path);
    this.delete(relUrl, callback);
};
Adapter.prototype.cleanFromTrash = function (path, callback) {
    var relUrl = util.format('/api/trash/clean/%s', path);
    this.delete(relUrl, callback);
};
Adapter.prototype.getUser = function (name, callback) {
    var relUrl = util.format('/api/users/%s', name);
    this.get(relUrl, {}, callback);
};
Adapter.prototype.getBlockUsages = function (callback) {
    this.get('/api/subscription/blockedUsages', {}, callback);
};
//# sourceMappingURL=Adapter.js.map