"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RunAdjacency = void 0;
const metadata_1 = require("@jfrog-pipelines/metadata");
const utility_1 = require("@jfrog-pipelines/utility");
const underscore_1 = require("underscore");
class RunAdjacency {
    constructor(dataAdapter) {
        this.dataAdapter = dataAdapter;
        this.runAdjacencyList = [];
        this.runAdjacencyByRunId = {};
    }
    async populate() {
        this.runAdjacencyList = await this.getRunAdjacencyList();
        this.runAdjacencyByRunId = (0, utility_1.groupBy)(this.runAdjacencyList, 'runId');
    }
    getRunAdjacencyStepletIds() {
        const stepletIds = (0, utility_1.uniq)((0, utility_1.map)(this.runAdjacencyList, 'stepletId'));
        return stepletIds;
    }
    getRunAdjacencyStepIds() {
        const stepIds = (0, utility_1.uniq)((0, utility_1.map)(this.runAdjacencyList, 'stepId'));
        return stepIds;
    }
    getRunAdjacencyRunIds() {
        const runIds = (0, utility_1.keys)(this.runAdjacencyByRunId);
        return runIds;
    }
    getRunAdjacencyByRunId(runId) {
        return this.runAdjacencyByRunId[runId];
    }
    getAffinityGroups(runId) {
        const runAdjacencies = this.getRunAdjacencyByRunId(runId);
        const listByAffinityGroup = (0, utility_1.groupBy)(runAdjacencies, 'affinityGroup');
        const affinityGroups = (0, utility_1.keys)(listByAffinityGroup);
        return { affinityGroups, listByAffinityGroup };
    }
    async getRunAdjacencyList() {
        const runAdjacencyList = await this.dataAdapter.getTriggerableSteps();
        return this.filterByDependencyModePipelineIds(runAdjacencyList);
    }
    async filterByDependencyModePipelineIds(runAdjacencyList) {
        const parentCompletePipelineIdsSet = new Set();
        const childCompletePipelineIdsSet = new Set();
        const parentSuccessPipelineIdsSet = new Set();
        for (const runAdj of runAdjacencyList) {
            this.populateDependencyPipelines(runAdj, parentCompletePipelineIdsSet, parentSuccessPipelineIdsSet, childCompletePipelineIdsSet);
        }
        const allPipelineIds = new Set([
            ...parentCompletePipelineIdsSet.values(),
            ...childCompletePipelineIdsSet.values(),
            ...parentSuccessPipelineIdsSet.values()
        ]);
        if (allPipelineIds.size === 0) {
            return runAdjacencyList;
        }
        const incompleteStatusCodes = [metadata_1.CREATING, metadata_1.READY, metadata_1.WAITING, metadata_1.PROCESSING, metadata_1.TIMINGOUT, metadata_1.CANCELLING];
        const incompleteRuns = await this.dataAdapter.getRuns([...allPipelineIds.values()], `include=id,pipelineId&statusCodes=${incompleteStatusCodes.join(',')}&sortBy=id&sortOrder=DESC`);
        const incompleteRunsByPipelineId = (0, utility_1.groupBy)(incompleteRuns, 'pipelineId');
        return this.checkAndFilterRunAdjacencies(allPipelineIds, incompleteRunsByPipelineId, parentSuccessPipelineIdsSet, runAdjacencyList);
    }
    populateDependencyPipelines(runAdj, parentCompletePipelineIdsSet, parentSuccessPipelineIdsSet, childCompletePipelineIdsSet) {
        if (runAdj.dependentEntityType === 3) {
            parentCompletePipelineIdsSet.add(runAdj.dependentEntityId);
        }
        if (runAdj.dependentEntityType === 4) {
            parentSuccessPipelineIdsSet.add(runAdj.dependentEntityId);
        }
        if (runAdj.dependentEntityType === 5) {
            childCompletePipelineIdsSet.add(runAdj.dependentEntityId);
        }
    }
    checkAndFilterRunAdjacencies(allPipelineIds, incompleteRunsByPipelineId, parentSuccessPipelineIdsSet, runAdjacencyList) {
        const checkForParentSuccess = new Set();
        const dependencyModePipelinesUnresolved = new Set();
        for (const pipelineId of allPipelineIds) {
            const incompleteRuns = incompleteRunsByPipelineId[pipelineId];
            if (!(0, underscore_1.isEmpty)(incompleteRuns)) {
                dependencyModePipelinesUnresolved.add(pipelineId);
            }
            else {
                if (parentSuccessPipelineIdsSet.has(pipelineId)) {
                    checkForParentSuccess.add(pipelineId);
                }
            }
        }
        return this.removeRunAdjacencies(checkForParentSuccess, dependencyModePipelinesUnresolved, runAdjacencyList);
    }
    async removeRunAdjacencies(checkForParentSuccess, dependencyModePipelinesUnresolved, runAdjacencyList) {
        if (checkForParentSuccess.size > 0) {
            const pipelineIds = [...checkForParentSuccess.values()];
            await this.checkForParentSuccess(pipelineIds, dependencyModePipelinesUnresolved);
        }
        if (dependencyModePipelinesUnresolved.size === 0) {
            return runAdjacencyList;
        }
        const unresolvedPipelines = this.getPipelinesWaitingForDependencyMode(dependencyModePipelinesUnresolved, runAdjacencyList);
        return this.filterUnresolvedDependencyModePipelines(unresolvedPipelines, runAdjacencyList);
    }
    async checkForParentSuccess(pipelineIds, dependencyModePipelinesUnresolved) {
        const includeParameters = `id,latestSuccessfulRunId,latestCompletedRunId`;
        const query = `include=${includeParameters}`;
        const pipelines = await this.dataAdapter.getPipelines(pipelineIds, query);
        const pipelinesById = (0, utility_1.keyBy)(pipelines, 'id');
        for (const pipelineId of pipelineIds) {
            const pipeline = pipelinesById[pipelineId];
            if (!(0, underscore_1.isEmpty)(pipeline) && pipeline.latestCompletedRunId !== pipeline.latestSuccessfulRunId) {
                dependencyModePipelinesUnresolved.add(pipelineId);
            }
        }
    }
    filterUnresolvedDependencyModePipelines(unresolvedPipelines, runAdjacencyList) {
        return runAdjacencyList.filter((runAdj) => {
            return !unresolvedPipelines.has(runAdj.pipelineId);
        });
    }
    getPipelinesWaitingForDependencyMode(pipelineIdSet, runAdjacencyList) {
        const dependencyModeEntityTypes = new Set([
            5,
            3,
            4
        ]);
        const waitingPipelines = new Set();
        for (const runAdj of runAdjacencyList) {
            const dependentEntityType = runAdj.dependentEntityType;
            const dependentEntityId = runAdj.dependentEntityId;
            if (!dependencyModeEntityTypes.has(dependentEntityType)) {
                continue;
            }
            if (pipelineIdSet.has(dependentEntityId)) {
                waitingPipelines.add(runAdj.pipelineId);
            }
        }
        return waitingPipelines;
    }
}
exports.RunAdjacency = RunAdjacency;
//# sourceMappingURL=RunAdjacency.js.map